//? cpmbasicinterfaces.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_BASIC_INTERFACES_H_
#define CPM_BASIC_INTERFACES_H_
/*

   Description: see cpminterfaces.h

*/

// The following makes sure that no client class will use Type in a way
// that would need copy constructor and assignment.
// Should be placed in the private section

#include <cpmbasictypes.h>
   // for Z in CPM_ORDER

#define CPM_INVAR(TypeName)\
   Type& operator = (Type const&);\
   TypeName(Type const&);

// example for usage
/*
   template <class X, class Y>
   class SUMFUO: public ...{
      typedef SUMFUO<X,Y> Type;
      CPM_INVAR(SUMFUO)
   public:
      ...
   };
*/

// order related stuff

#define CPM_ORDER_PLAIN\
   bool operator == ( Type const& x)const;\
   bool operator != (Type const& x)const;\
   bool operator < (Type const& x)const;\
   bool operator > (Type const& x)const;\
   bool operator <= (Type const& x)const;\
   bool operator >= (Type const& x)const;

// consistent generation of all six order-related operators from a
// single member functions. Notice that it would be useless to make com
// virtual since any useful re-definition probably differs in the
// type of the second argument too.

// see cpmnumers.h for explanations to 'com'

#define CPM_ORDER\
   CpmRoot::Z com(Type const&)const;\
   bool equalTo(Type const& x)const{ return com(x)==0;}\
   bool priorTo(Type const& x)const{ return com(x)==1;}\
   bool operator == ( Type const& x)const\
      { return com(x)==0;}\
   bool operator != (Type const& x)const\
      { return com(x)!=0;}\
   bool operator < (Type const& x)const\
      { return com(x)>0;}\
   bool operator > (Type const& x)const\
      { return com(x)<0;}\
   bool operator <= (Type const& x)const\
      { return com(x)>=0;}\
   bool operator >= (Type const& x)const\
      { return com(x)<=0;}

#endif
