//? cpmbasictypes.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_BASIC_TYPES_H_
#define CPM_BASIC_TYPES_H_
/*
   Description: Introduces mathematics style named
      aliases for basic integer number types.
*/
#include <cstddef> // for std::ptrdiff_t and std::size_t
#include <cpmdefinitions.h> // for CPM_LONG and CPM_MP

namespace CpmRoot{
   // Basic number-related C+- types and functions.

// This defines what types are to be used to represent
// C+-'s notion of integer numbers. Floating point types
// will be treated in cpmnumbers.h

// Integer types L, N, Z.
// Remember that N and Z are standard names for the sets of
// natural numbers and integer numbers in mathematics.

#if defined(CPM_LONG) || defined(CPM_MP) // take the largest possible data
// types
   typedef long int Z;
      // integer numbers
   typedef unsigned long int N;
      // natural numbers
      // provides cyclic definition of addition, without overflow.
#else // take normal int and unsigned
   typedef int Z;
      // integer numbers
   typedef unsigned int N;
      // natural numbers
      // Type N will be used only in implementation code,
      // where it is important that addition is a cyclic operation
      // which never creates overflow. As type of function
      // arguments I don't use N. The reason is the following:
      // Function arguments for which Z or N would be appropriate types
      // occur frequently. Often, in cases in which N seemed the natural
      // choice when defining a function in the first place, it has to
      // be discovered that an extension to Z could be useful. So one
      // would be tempted so switch (or oscillate!) between using Z and N.
      // This is to be avoided!
#endif // CPM_LONG || defined(CPM_MP)

   typedef unsigned char L;
      // 'L' for 'letter' represents a byte which after 'integral
      // promotion' is a value between 0 and 255.
      // Mainly needed for storing pixel values in image matrices.
      // Gets written to files as a number \in {0,1,...,255}
} // CpmRoot

#endif // CPM_BASIC_TYPES_H_
