//? cpmcompdef.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


// no double include guard needed since this occurs only as
// '#include <cpmcompdef.h>' in cpmdefinitions.h and the letter file
// is guarded against double inclusion.
/*************************************************************************
   cpmcompdef.h
   Description: In this file one defines how compiler options of the group
   -D influence the compilation directives defined in cpmdefinition.h
************************************************************************/
// ways to define R by compiler options of group -D
// one line in a makefile could for instance be
// defines = -D_CONSOLE -DCPM_RMP_C
// compilation than would result in an executable which uses a
// representation with 64 decimal places (actually the computed equivalent
// of binary places) for each number of type R.
// Since using multiple precision arithmetics conflicts with the Eigen
// library (which assumes standard float types) the usage of Eigen is
// disabled if CPM_MP is defined.
#if defined(CPM_RMP_A)
   #undef CPM_LONG
   #undef CPM_MP
   #define CPM_MP 16
#elif defined(CPM_RMP_B)
   #undef CPM_LONG
   #undef CPM_MP
   #define CPM_MP 32
#elif defined(CPM_RMP_C)
   #undef CPM_LONG
   #undef CPM_MP
   #define CPM_MP 64
#elif defined(CPM_RMP_D)
   #undef CPM_LONG
   #undef CPM_MP
   #define CPM_MP 128
#elif defined(CPM_RMP_E)
   #undef CPM_LONG
   #undef CPM_MP
   #define CPM_MP 256
#elif defined(CPM_RLONG)
   #undef CPM_MP
   #define CPM_LONG
#elif defined(CPM_R)
   #undef CPM_MP
   #undef CPM_LONG
#endif

// Asking for multiple precision is given priority over asking for
// fast linear algebra support provided by the Eigen library.
#if defined(CPM_MP)
   #undef CPM_USE_EIGEN
#endif

