//? cpmvr.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_VR_H_
#define CPM_VR_H_
/*

Purpose: Defining a vector template providing the full
functionality of the rich interface

*/
#include <cpmva.h>

//////////////////////////// class Vr<> //////////////////////////////////

namespace CpmArrays{

   using CpmRoot::R;
   using CpmRoot::Root;

template <class T>
   // We assume that T provides (explicitely or implicitely)
   // copy constructor, and assignement
   // or that T is a built-in type.

class Vr: public Va<T>{ // version of Va with a rich interface

public:
   typedef Vr<T> Type;
   typedef Va<T> Base;
   typedef T ScalarType; // was missing till 2001-10-12
      // detected by the gnu compiler

   explicit Vr(Z n=0):Va<T>(n){}
      // note that Vr(0) is defined and is different from Vr()

   explicit Vr(IvZ ivz):Va<T>(ivz){}
      // note that Vr(0) is defined and is different from Vr()

   Vr(Z n, T const& t):Va<T>(n,t){}
      // initializes all n components with t

   Vr( V<T> const& h):Va<T>(h){}
      // 'down-cast'-constructor

   Vr( Vo<T> const& h):Va<T>(h){}
      // 'down-cast'-constructor

   Vr(Va<T> const& h):Va<T>(h){}
      // 'down-cast'-constructor

   virtual V<T>* clone(void)const{ return new Vr<T>(*this);}

   CPM_TEST
   CPM_DESCRIPTORS
      // dis from Va
};

// member functions

template <class T>
Word Vr<T>::nameOf(void)const
{
   Word w1="Vr<";
   Word w2=Root<T>(T()).nameOf();
   return w1&w2&">";
}

template <class T>
Word Vr<T>::toWord(void)const
{
   if (Base::isVoid()) return "void list";
   Word res="( ";
   Z i;
   for (i=Base::b();i<Base::e();i++){
      Word wi="("&Root<Z>(i).toWord()&","&Root<T>((*this)[i]).toWord()&"),\
       ";
      res&=wi;
   }
   i=Base::e();
   Word we="("&Root<Z>(i).toWord()&","&Root<T>((*this)[i]).toWord()&")";
   res&=we;
   res&=" )";
   return res;
}

template <class T>
Z Vr<T>::hash(void)const
{
   Z res=0;
   for (Z i=Base::b(); i<=Base::e(); i++){
      res+=Root<T>((*this)[i]).hash();
   }
   return res;
}

template <class T>
Vr<T> Vr<T>::test(Z cpl)const
{
   Z it=0;
   it=Root<Z>(it).test(cpl);
   T xt;
   xt=Root<T>(xt).test(cpl);
   return Vr<T>(it,xt);
}

template <class T>
Vr<T> Vr<T>::ran(Z j)const
   // all randomized versions of *this thus have the same dimension as
   // *this
{
   Vr<T> res(Base::dom());
   T tempX;
   if (j==0){
      for (Z i=Base::b(); i<=Base::e(); i++){
         tempX=(*this)[i];
         res[i]=Root<T>(tempX).ran(j);
      }
   }
   else{
      Z j0=1000; // this is done to avoid an obvious correlation
         // among the coordinates of x.ran(i)
         // and x.ran(i+1)
      Z jIncr=Root<Z>(j0).ran(j);
      Z jc=j;
      for (Z i=Base::b(); i<=Base::e(); i++){
         tempX=(*this)[i];
         jc+=jIncr;
         res[i]=Root<T>(tempX).ran(jc);
      }
   }
   return res;
}

// We do not implement the iterated templates since they are now
// endowed with sufficient functionality already as Va, VVa, VVVa,
// and VVVVa

} // namespace

#endif
