//? cpmviewport.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmsystemdependencies.h>
   // This file is not included in any header file in order
   // to make only those translation units dependent on these
   // macros which really depend on them.
   // Important that this precedes all other includes.
#include <cpmviewport.h>
//#include <cpmtypes.h> // experiment, needed for cpmdebug

using CpmRoot::Z;
using CpmRoot::L;
using CpmRoot::R;
using CpmRoot::Word;
using CpmRoot::getByte;

using CpmSystem::Message;
//using CpmSystem::R_;

using CpmArrays::V;
using CpmArrays::LEAN;

using CpmGraphics::Rec;
using CpmGraphics::xy;
using CpmGraphics::xyxy;
using CpmGraphics::Viewport;
using CpmGraphics::rgb;
using CpmGraphics::ColRef;

using namespace std;

Z CpmGraphics::Viewport::heightStatusBar=16;
//////////////////////// struct xyxy ///////////////////////////////////
namespace {
Z infZ(Z i, Z j){ return i<=j ? i : j;}
Z supZ(Z i, Z j){ return i<=j ? j : i;}
R infR(R i, R j){ return i<=j ? i : j;}
R supR(R i, R j){ return i<=j ? j : i;}
   // to avoid need of namespace CpmRootX
}

xyxy xyxy::operator|(xyxy r)const
{
   Z rx1=infZ(x1,r.x1);
   Z ry1=infZ(y1,r.y1);
   Z rx2=supZ(x2,r.x2);
   Z ry2=supZ(y2,r.y2);
   return xyxy(rx1,ry1,rx2,ry2);
}

xyxy xyxy::operator&(xyxy r)const
{
   if (x2 < r.x1) return xyxy(-1,-1,-1,-1);
   if (r.x2 < x1) return xyxy(-1,-1,-1,-1);
      // if upper end of one is smaller than the lower
      // end of the other, then the projections of
      // *this and r to the x-axis are disjoint
   if (y2 < r.y1) return xyxy(-1,-1,-1,-1);
   if (r.y2 < y1) return xyxy(-1,-1,-1,-1);
      // same for x replaced by y
   Z rx1=supZ(x1,r.x1);
   Z ry1=supZ(y1,r.y1);
   Z rx2=infZ(x2,r.x2);
   Z ry2=infZ(y2,r.y2);
   return xyxy(rx1,ry1,rx2,ry2);
}

xyxy xyxy::scaleToFit(xyxy r)const
{
   Word loc("xyxy::scaleToFit(xyxy)");
   cpmassert(!r.isVoid()&&!isVoid(),loc);
   R w1=w(), h1=h(), w2=r.w(), h2=r.h();
   R sw=R(w1)/w2; // scale factor to adjust width
   R sh=R(h1)/h2; // scale factor to adjust height
   Z wN,hN; // width and hight of the result
   R s=infR(sw,sh);
   wN=cpmtoz(s*w2);
   hN=cpmtoz(s*h2);
   return (*this)&xyxy(cor(1),wN,hN);
 }

//////////////////////// class Rec ///////////////////////////////////////

Rec& Rec::operator+=(xy shift)
{
   rec_+=shift;
   rec_&=Viewport::win();
   return *this;
}
Rec& Rec::operator-=(xy shift)
{
   (rec_-=shift)&=Viewport::win(); return *this;
}
// the following implementation follows that of V<> in cpmv.h

Rec::~Rec(){ /*printf(" ~Rec() called ");*/ delete[] mem_;}

Rec::Rec(xyxy r): rec_(r&Viewport::win()),
   n_(rec_.w()*rec_.h()*3),mem_(new L[n_])
// together with the usage in Rec::vis, this is all we need from OpenGL
{
#if !defined(CPM_NOGRAPHICS)
   glFlush();
   Z x1=rec_.x1, y1=rec_.y1, w=rec_.w(), h=rec_.h(), wBytes=3*w;
   L* p=mem_;
   Z gly=Viewport::fullWin().h()-y1-h;
   Z j=h;
   while (j--){
     L* buff=new L[wBytes];
     glReadPixels(x1,gly++,w,1,GL_RGB,GL_UNSIGNED_BYTE,buff);
     Z ib=wBytes;
     L* q=buff;
     while (ib--) *p++=*q++;
     delete[] buff;
   }
#endif
}

Rec::Rec(xyxy r, L c): rec_(r&Viewport::win()),
   n_(rec_.w()*rec_.h()*3),mem_(new L[n_])
{
   L* p=mem_;
   Z j=n_;
   while (j--)*p++=c;
}
      // The pixels within the rectangle r are set equal to c
      // On display this is always a gray (black to white) frame.
      // Notice that colorful colors would be more difficult to
      // handle here

Rec& Rec::operator=(Rec const& r)
{
   if (this==&r) return *this;
   rec_=r.rec_;
   Z nMem=n_;
   n_=r.n_;
   if (n_!=nMem){ //  avoid superfluous delete-new action
      delete[] mem_;
      mem_=new L[n_];
   }
   L* q=r.mem_;
   L* it=mem_;
   Z i=n_;
   while (i--) *it++ = *q++;
   return *this;
}

Rec::Rec(Rec const& r):rec_(r.rec_),n_(r.n_)
{
   L* itp=r.mem_;
   L* q=new L[n_];
   L* it=q;
   Z i=n_;
   while (i--) *it++=*itp++;
   mem_=q;
}

Rec::Rec(V< V<ColRef> > const& bh)
{
   Z h1=bh.dim();
   Z w1=bh.fir().dim();
   xyxy r1(xy(0,0),w1,h1);
   rec_=r1&Viewport::win();
   Z height=rec_.h();
   Z width=rec_.w();
   n_=height*width*3;
   mem_=new L[n_];
   L* p=mem_;
   for (Z i=height-1;i>=0;i--){
      for (Z j=0;j<width;j++){
         ColRef val=bh[i][j];
         *p++=val[1];
         *p++=val[2];
         *p++=val[3];
       }
   }
}

V< V<ColRef> >  Rec::fold()const
{
   L* p=mem_;
   Z w1=w(),h1=h();
   V<V<ColRef> > res(h1,V<ColRef>(w1,LEAN),0);
   for (Z i=h1-1;i>=0;--i){
      for (Z j=0;j<w1;++j){
         L r=*p++;
         L g=*p++;
         L b=*p++;
         res[i][j]=ColRef(r,g,b);
      }
   }
   return res;
}

void Rec::vis()const
// screen representation of a linear block of pixel values.
// Basicly this is all we need from OpenGL
{
#if !defined(CPM_NOGRAPHICS)
//   Z mL=1;
//   Word loc("Rec::dis()");
//   CPM_MA
   Z h1=rec_.h();
   Z w1=rec_.w();
   glRasterPos2i(rec_.x1,rec_.y1+h1);
   glDrawPixels(w1,h1,GL_RGB,GL_UNSIGNED_BYTE,mem_);
   glFlush();
//   CPM_MZ
#endif
}

///////////////////// class ColRef /////////////////////////

Z ColRef::lowerLimit=0;
Z ColRef::upperLimit=255;
   // no surprises

Z ColRef::x(Z z)
{
   if (z>ColRef::upperLimit) return ColRef::upperLimit;
   else if (z<ColRef::lowerLimit) return ColRef::lowerLimit;
   else return z;
}

bool ColRef::prnOn(ostream& str)const
{
   cpmwt("ColRef");
   cpmp(r);
   cpmp(g);
   cpmp(b);
   return true;
}

bool ColRef::scanFrom(istream& str)
{
   cpms(r);
   cpms(g);
   cpms(b);
   return true;
}

Z ColRef::com(ColRef const& obj)const
{
   cpmc(r);
   cpmc(g);
   cpmc(b);
   return 0;
}

///////////////////// class rgb //////////////////////////////

Z CpmGraphics::rgb::mixingRule(Z z1, Z z2, R w1, R w2, R gamma)
{
   if (gamma==0)
      return ColRef::upperLimit;
   else if (gamma==1)
      return cpmtoz(w1*z1+w2*z2);
   else if (gamma==2)
      return cpmtoz(sqrt(w1*z1*z1+w2*z2*z2));
   else if (gamma>0 && gamma<10){
      R p1=w1*cpmpow(R(z1),gamma);
      R p2=w2*cpmpow(R(z2),gamma);
      R gammaInv=1./gamma;
      return cpmtoz(cpmpow(p1+p2,gammaInv));
   }
   else
      return (z1>=z2 ? z1 : z2);
}

rgb CpmGraphics::rgb::mix(rgb c1, rgb c2, R w1, R w2, R gamma)
{
   Z a=mixingRule(c1.r,c2.r,w1,w2,gamma);
   Z b=mixingRule(c1.g,c2.g,w1,w2,gamma);
   Z c=mixingRule(c1.b,c2.b,w1,w2,gamma);
   rgb res(a,b,c);
   res.normalize();
   return res;
}

///////////////////////// class Viewport /////////////////////////////

// common initializations of static variables

ColRef CpmGraphics::Viewport::colText=ColRef();
R CpmGraphics::Viewport::gamma=2.3;
// best guess from a Kodak tool GAMMA.TIF for my office monitor
// is 2.3, also for my TFT-Monitor this seems to be the best
xyxy CpmGraphics::Viewport::window;
xyxy CpmGraphics::Viewport::fullWindow;
Z CpmGraphics::Viewport::size=1;
Z CpmGraphics::Viewport::rank=1;

#if !defined(CPM_NOGRAPHICS)
using CpmGraphics::Font;
bool CpmGraphics::Viewport::initialized=false;
Z CpmGraphics::Viewport::applWindowWidth=0;
Z CpmGraphics::Viewport::applWindowHeight=0;
Z CpmGraphics::Viewport::fullWindowHeight=0;

namespace{

Z getX()
{
   GLint x4[4];
   glGetIntegerv(GL_CURRENT_RASTER_POSITION,x4);
   return x4[0];
}

} // namespace

CpmGraphics::Font::Font(Z type)
{
   if (type==1){
      name=GLUT_BITMAP_TIMES_ROMAN_10;
      h=10;
   }
   else if (type==2){
      name=GLUT_BITMAP_HELVETICA_10;
      h=10;
   }
   else if (type==3){
      name=GLUT_BITMAP_HELVETICA_12;
      h=12;
   }
   else if (type==4){
      name=GLUT_BITMAP_8_BY_13;
      h=13;
   }
   else if (type==5){
      name=GLUT_BITMAP_9_BY_15;
      h=15;
   }
   else if (type==6){ // GLUT_BITMAP_HELVETICA_18 is very poorly designed
      // or implemented and should not be used
      // name=GLUT_BITMAP_HELVETICA_18;
      // h=18;
      name=GLUT_BITMAP_9_BY_15;
      h=15;
   }
   else {
      name=GLUT_BITMAP_TIMES_ROMAN_24;
      h=24;
   }
   d=cpmrnd(h*0.2);
   mw=cpmrnd(h*0.6); // as in 9 by 15 font
      // only an approximation - needed since glutBitmapWidth(...)
      // couldn't be made working so far
}

CpmGraphics::Font::Font(R fh)
{
   if (fh<=10.0){
      name=GLUT_BITMAP_HELVETICA_10;
      h=10;
   }
   else if (fh<=11.0){
      name=GLUT_BITMAP_TIMES_ROMAN_10;
      h=10;
   }
   else if (fh<=12.5){
      name=GLUT_BITMAP_HELVETICA_12;
      h=12;
   }
   else if (fh<=14.0){
      name=GLUT_BITMAP_8_BY_13;
      h=13;
   }
   else if (fh<=16.5){
      name=GLUT_BITMAP_9_BY_15;
      h=15;
   }
   else if (fh<=21.0){ // see comment to Font(6)
   //   name=GLUT_BITMAP_HELVETICA_18;
    //  h=18;
      name=GLUT_BITMAP_9_BY_15; // experiment due to L'
      h=15;
   }
   else if (fh<=24.5) {
      name=GLUT_BITMAP_TIMES_ROMAN_24;
      h=24;
   }
   else{
      name=GLUT_BITMAP_TIMES_ROMAN_24;
      h=24;
   }
   d=cpmrnd(h*0.2);
   mw=cpmrnd(h*0.6); // as in 9 by 15 font
      // only an approximation - needed since glutBitmapWidth(...)
      // couldn't be made working so far
}

// basic Cpm graphics library should be independent from CpmSystem and
// thus not to have to include cpmsystem.h

xyxy CpmGraphics::Viewport::seg1;
xyxy CpmGraphics::Viewport::seg2;
xyxy CpmGraphics::Viewport::seg3;
xyxy CpmGraphics::Viewport::seg4;

// background colors of the panes on the status bar
ColRef CpmGraphics::Viewport::cs1(L(255),L(255),L(50));
ColRef CpmGraphics::Viewport::cs2(L(210),L(255),L(110));
ColRef CpmGraphics::Viewport::cs3(L(180),L(180),L(190));
ColRef CpmGraphics::Viewport::cs4(L(165),L(165),L(255));

void Viewport::fill(xyxy r, ColRef cr)
{
   glColor3f(cr.glR(),cr.glG(),cr.glB());
   glRecti(r.x1,r.y1,r.x2+1,r.y2+1);
   glFlush();
}

Viewport::Viewport(Z w, Z h, const Word& title)
{
      if (initialized) return;
      Word loc("Viewport::Viewport(Y,Y,Word)");
      size=CpmMPI::Cpm_com.getSize();
      rank=CpmMPI::Cpm_com.getRank();
      ColRef crBac(L(230),L(255),L(180));
      applWindowWidth=w;
      fullWindowHeight=h;
      fullWindow=xyxy(xy(0,0),applWindowWidth,fullWindowHeight);
      applWindowHeight=fullWindowHeight-heightStatusBar;
      window=xyxy(xy(0,0),applWindowWidth,applWindowHeight);
      if (rank==1){
         Word mes="OpenGL window got for applWindowWidth = ";
         mes&=cpm(applWindowWidth);
         mes&=" , applWindowHeight= ";
         mes&=cpm(applWindowHeight);
         Word mes2="Platform is "&Platform;
         cout<<mes<<endl;
         cout<<mes2<<endl;
      }
      R cp1=Message::pane1(); // from cpmsystemdependencies
      R cp2=Message::pane2();
      R cp3=Message::pane3();
      R cp4=Message::pane4();
      R sumPanes=cp1+cp2+cp3+cp4;
      if (sumPanes==0){
         cp1=2;
         cp2=2;
         cp3=1;
         cp4=1;
         sumPanes=cp1+cp2+cp3+cp4;
      }
      R awsp=applWindowWidth/sumPanes;
      Z w1=cpmrnd(awsp*cp1);
      Z w2=cpmrnd(awsp*cp2);
      Z w3=cpmrnd(awsp*cp3);
      Z w4=applWindowWidth-w1-w2-w3;
      Z yL=applWindowHeight;
      Z xL=0;
      seg1=xyxy(xy(xL,yL),w1,heightStatusBar);
      xL+=w1;
      seg2=xyxy(xy(xL,yL),w2,heightStatusBar);
      xL+=w2;
      seg3=xyxy(xy(xL,yL),w3,heightStatusBar);
      xL+=w3;
      seg4=xyxy(xy(xL,yL),w4,heightStatusBar);
      initialized=true;
      xy p0;
      xyxy all(p0,applWindowWidth,applWindowHeight);
      fill(all,crBac);
      R fontHeight=14;
      Font font(fontHeight);
      Z tx=font.getWidth();
      Z ty=font.getHeight();
      ColRef crText(L(0),L(0),L(255));
      textLine(tx,ty,title,crText,fontHeight);
}

void Viewport::onStatusBar(const Word& w, Z i)
{
   static Z th=cpmrnd(0.75*heightStatusBar);
   R fH=12.0;
   static Font font(fH); // Helvetica 12
   if (!isInitialized() ) return;
   ColRef cr;
   xyxy seg;
   if (i==1){
      cr=cs1;
      seg=seg1;
   }
   else if (i==2){
      cr=cs2;
      seg=seg2;
   }
   else if (i==3){
      cr=cs3;
      seg=seg3;
   }
    else if (i==4){
      cr=cs4;
      seg=seg4;
   }
   fill(seg,cr);
   Z shift=2;
   placeWord(seg.x1+shift, seg.y1+th, w, font, seg.w()-shift);
 }

void Viewport::clrPan()
{
   fill(seg1,cs1);
   fill(seg2,cs2);
   fill(seg3,cs3);
   fill(seg4,cs4);
}

void Viewport::placeWord(Z tx, Z ty, const Word& w, const Font& f,
   Z wMax)
{
   glColor3f(colText.glR(),colText.glG(),colText.glB());
   void* FONT=f.getName();
   Z x=tx,y=ty;
   Z xLim=tx+wMax;
   Z xM=f.getWidth();
   xLim-=xM;
   Z n=w.dim();
   glRasterPos2i(x,y);
   for (Z i=1;i<=n;++i){
      glutBitmapCharacter(FONT,w[i]);
      x=getX();
      if (x>xLim) break;
   }
   glFlush();
}

xyxy Viewport::textLine(Z tx, Z ty, const Word& w,
      ColRef crText, R h)
      // writing w on a line previously filledwith crBac
{
   glColor3f(colText.glR(),colText.glG(),colText.glB());
   Font f(h);
   void* FONT=f.getName();
   Z x=tx,y=ty;
   Z xM=f.getWidth();
   Z yM=f.getHeight();
   Z decender=(Z)(0.25*yM+0.5);
   Z i,n=w.dim();
   Z more=2; // makes the background rectangle a bit larger than the text
   glColor3f(crText.glR(),crText.glG(),crText.glB());
   glRasterPos2i(x,y);
   for (i=1;i<=n;++i){
       glutBitmapCharacter(FONT,w[i]);
   }
   Z txF=getX();
   glFlush();
   return xyxy(tx-more,ty-yM+decender-more,txF+more,ty+decender+more);
 }

xyxy Viewport::addText(Z x, Z y, Word const& line,
   R fontHeight, Z fontNumber)
{
   fontNumber; // not used
   return textLine(x,y,line,colText,fontHeight);
}

#else
// e.g. if CPM_NOGRAPHICS is defined
// trivial implementation; since then graphics
// works through a RAM-based data structure Img24

bool CpmGraphics::Viewport::initialized=true;

Viewport::Viewport(Z w, Z h, const Word& title){}
void Viewport::onStatusBar(const Word& w, Z i){}

Z CpmGraphics::Viewport::applWindowWidth=10;
Z CpmGraphics::Viewport::applWindowHeight=10;
Z CpmGraphics::Viewport::fullWindowHeight=10;

void CpmGraphics::Viewport::placeWord(Z tx, Z ty, const Word& w,
   const Font& f, Z wMax){tx;ty;w;f;wMax;}

void CpmGraphics::Viewport::fill(xyxy r, ColRef cr){r;cr;}

xyxy Viewport::textLine(Z tx, Z ty, const Word& w,
      ColRef crText, R h)
{
   tx;ty;w;crText;h;
   return xyxy();
}

xyxy Viewport::addText(Z x, Z y, Word const& text, R size, Z fontNumber)
{
   x;y;text;size;fontNumber;
   return xyxy();
}

void Viewport::clrPan()
{}

#endif
