//? cpmword.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmtypes.h>
#include <stdio.h>
#include <cctype>
#include <iomanip>

using namespace CpmStd;
using CpmRoot::Word;
using CpmRoot::Z;
using CpmRoot::N;
using CpmRoot::R;
using CpmRoot::L;
using CpmRoot::toDouble;
using CpmRoot::skipLeadingWhitespace;
using CpmRoot::skipTrailingWhitespace;

bool Word::scnLin=true;
   // true is dangerous for the presenly implemented movie file
   // production. Don't change without careful analysis of the
   // consequences

Z Word::endOfWord=2; // was 1
   // 0 nothing, 1 blank, else endl
   // choice 2 is the most stable since a endl, in addition to '\n'
   // also puts flush

bool Word::prnOn(ostream& str)const
{
   if (endOfWord==0) str<<rep_;
   else if (endOfWord==1) str<<rep_<<" ";
   else              str<<rep_<<endl; // now the default case
   //return (str!=0);
   return !str ? false : true;
}

bool Word::scanFrom(istream& str)
{
   if (scnLin) CpmRoot::readLine(rep_,str);
   else        CpmRoot::read(rep_,str);
   //return (str!=0);
   return !str ? false : true;
}

Z Word::find(char c)const
{
   Z res=0;
   Z n=dim();
   for (Z i=0;i<n;++i){
      if (rep_[i]==c){
         res=i+1;
         break;
      }
   }
   return res;
}

Z Word::find(const Word& w)const
{
   string::size_type res=rep_.find(w.rep_);
   if (res==string::npos) return 0;
   else return 1+(Z)res;
}

Word Word::resize(Z newDim)const
{
   Z n1=dim();
   Z n2= newDim<0 ? 0 : newDim;
   char blank=' ';
   Word res(n2,blank);
   Z nMin = n1<=n2 ? n1 : n2;
   for (Z i=1;i<=nMin;++i) res[i]=(*this)[i];
   return res;
}

Word Word::slc_(char c, bool firstSep)
{
   Z d=dim();
   if (d==0) return *this;
   Word temp= firstSep ? *this : rev();
   Z ic=temp.find(c);
   if (ic==0){ Word res=*this; *this=""; return res;}
   if (d==1 && ic==1){ // Then *this consists of a single c
      *this="";
      return "";
   }
   Word wl(ic-1);
   Word wr(d-ic);
   Z i,j;
   for (i=1;i<ic;++i) wl[i]=temp[i];
   for (i=ic+1,j=1;i<=d;++i,++j) wr[j]=temp[i];
   if (firstSep){
      *this=wr;
      return wl;
   }
   else{
      *this=wl.rev();
      return wr.rev();
   }
}

Word Word::appBefExt(Word const& app)const
{
   Word dot(".");
   Z posDot=find(dot);
   if (posDot==0) return *this&app;
   else{
      Z nameL=posDot-1;
      Z extL=dim()-posDot+1;
      Word name=resize(nameL);
      Word ext=tail(extL);
      name&=app;
      name&=ext;
      return name;
   }
}

Word Word::baseName()const
{
   Word res;
   string::size_type i1=rep_.find_last_of("/");
   string::size_type i2=rep_.find_last_of("\\");
   string repx=rep_;
   if (i1==string::npos && i2==string::npos){
      ; // nothing to do
   }
   else if (i1==string::npos){ // then i2 is a valid string position that
      // holds a backslash
      repx.erase(0,i2+1);
   }
   else if (i2==string::npos){// then i2 is a valid string position that
      // holds a slash
      repx.erase(0,i1+1);
   }
   else{
      string::size_type imax=i1;
      if (i2>i1) imax=i2;
         // then right of imax there is np slash or backslash
      repx.erase(0,imax+1);
   }
   return Word(repx);
}

char Word::operator[](Z i)const
{
   return rep_.at(i-1); // checked access accorcing to the error handling
      // facilities of the standard library
}

char& Word::operator[](Z i)
{
   return rep_.at(i-1);// checked access accorcing to the error handling
      // facilities of the standard library
}

Word Word::cut(Z i)const
{
   string res=rep_;
   if (i>=0){
      res.erase(res.length()-i,i);
   }
   else{
      res.erase(0,-i);
   }
   return Word(res);
}

Word Word::tail(Z n)const
{
   Z d=dim();
   if (d<=n) return *this;
   Z diff=d-n; // notice diff>=1
   Word res(n);
   for (Z i=0;i<n;i++) res.rep_[i]=rep_[i+diff];
      // notice that for i==n-1, we have i+diff=n-1+d-n=d-1
      // so that we just stop with the last letter of rep_.
      // No essential efficiency compromises. Should be fast.
      // We do not use the substring functionality of the standard
      // library since my compiler does not understand Stroustrup's name
      // Basic_substring<char> even after using namespace std;
      // Also forming a new template class for this minor purpose
      // seems not be justified; also the floppy description of the
      // substring constructors on p.596 of BSC3 would require some
      // testing for being sure what the function do in case of
      // non-matching arguments.
   return res;
}

Word CpmRoot::operator &(const Word& w1, const Word& w2)
{
   return Word(w1.rep_+w2.rep_);
}

Word CpmRoot::operator +(const Word& w1, const Word& w2)
{
     return Word(w1.rep_+w2.rep_);
}

Word Word::write(bool x)
{
   return CpmRoot::toWord(x);
}

Word Word::write(string x)
{
   return Word(x);
}

#define CPM_SC\
   ostringstream ost;\
   if (length>0) ost<<std::setfill('0')<<std::setw(length)<<n;\
   else ost<<n;\
   return Word(ost.str())

Word Word::write(R n, Z length){CPM_SC;}
Word Word::write(Z n, Z length){CPM_SC;}
Word Word::write(N n, Z length){CPM_SC;}

#undef CPM_SC

R Word::toR()const
{
   stringstream sstr;
   sstr<<rep_;
   R res;
   sstr>>res;
   return res;
}

Z Word::toZ()const
{
   stringstream sstr;
   sstr<<rep_;
   Z res;
   sstr>>res;
   return res;
}

bool Word::toBool()const
{
   if (rep_=="true" || rep_=="TRUE" ||
      rep_=="wahr" || rep_=="1" || rep_=="W")
      return true;
   else
      return false;
}

namespace{
   const int fieldLength=32;
   // privat name to file
}

#if defined(_WINDOWS)
   #define CPM_SPF sprintf_s
#else
   #define CPM_SPF sprintf
#endif

#define CPM_SC\
   char p[fieldLength];\
   CPM_SPF(p,ff,x);\
   return Word(p)

Word CpmRoot::toWord(N x, const char* ff){CPM_SC;}
Word CpmRoot::toWord(Z x, const char* ff){CPM_SC;}
Word CpmRoot::toWord(R y, const char* ff){ double x=cpmtod(y); CPM_SC;}

#undef CPM_SPF
#undef CPM_SC

Word CpmRoot::operator &(const char* cp, const Word& w)
{ return Word(cp)&w;}
Word CpmRoot::operator &(const Word& w, const char* cp)
{ return w&Word(cp);}
Word CpmRoot::operator &(char cp[], const Word& w)
{ return Word(cp)&w;}
Word CpmRoot::operator &(const Word& w, char cp[])
{ return w&Word(cp);}
Word& Word::operator &=(const Word& w)
{ return *this=(*this)&w;}
Word& Word::operator &=(const char* cp)
{ return *this=(*this)&Word(cp);}

// for convenience same is allowed to be denoted +

Word CpmRoot::operator +(const char* cp, const Word& w)
{ return Word(cp)&w;}
Word CpmRoot::operator +(const Word& w, const char* cp)
{ return w&Word(cp);}
Word CpmRoot::operator +(char cp[], const Word& w)
{ return Word(cp)&w;}
Word CpmRoot::operator +(const Word& w, char cp[])
{ return w&Word(cp);}
Word& Word::operator +=(const Word& w)
{ return *this=(*this)&w;}
Word& Word::operator +=(const char* cp)
{ return *this=(*this)&Word(cp);}

ofstream& CpmRoot::to_ofstream(const Word& w)
{
    ofstream*  outptr=new ofstream(w.rep_.c_str());
    return *outptr;
}

ifstream& CpmRoot::to_ifstream(const Word& w)
{
    ifstream*  inptr=new ifstream(w.rep_.c_str());
    return *inptr;
}

bool Word::readable()const
{
   ifstream ifs(rep_.c_str());
   if (ifs) return true;
   else return false;
}

Word Word::firstWord()const
{
   istringstream ist(rep_);
   string st;
   ist>>st;
   return Word(st);
}

Word Word::firstWord_()
{
   istringstream ist(rep_);
   string st;
   ist>>st;
   rep_=string();
   char c;
   while (ist>>c) rep_+=c;
   return Word(st);
}

Z Word::com(const Word& w)const
{
   if (rep_<w.rep_) return 1;
   else if (rep_>w.rep_) return -1;
   else return 0;
}

Word Word::makeFileName()const
{
   string::const_iterator i;
   ostringstream res;
   for (i=rep_.begin(); i!=rep_.end(); ++i){
      if (isalnum(*i)) res<<*i;  else res<<'_';
   }
   return Word(res);
}

void Word::skipSpc_()
{
   string::const_iterator i;
   ostringstream res;
   for (i=rep_.begin(); i!=rep_.end(); ++i){
      if (!isspace(*i)) res<<*i;
   }
   rep_=res.str();
}

Word Word::skipChr(char c)const
{
   string::const_iterator i;
   ostringstream res;
   for (i=rep_.begin(); i!=rep_.end(); ++i){
      if (*i!=c) res<<*i;
   }
   return Word(res);
}

namespace{
   bool isOK(char c, bool extended)
   {
      if (!extended) return c!='.';
      else return c!='_'&& (islower(c)||isdigit(c));
   }
}

// we should remove any contiguous array of upper case characters
// (here '_' is considered an upper case character).

Word Word::remExt(bool extended)const
{
   string::const_iterator i;
   ostringstream res;
   for (i=rep_.begin(); i!=rep_.end(); ++i){
      if (isOK(*i,extended)) res<<*i;
      else break;
   }
   return Word(res);
}

Word Word::remApp()const
{
   Z d=dim();
   char l=(*this)[d];
   while(isupper(l)||l=='_'){
      d--;
      l=(*this)[d];
   }
   return resize(d);
}

void Word::skipSurSpc_()
{
   string line=skipLeadingWhitespace(rep_);
   rep_=skipTrailingWhitespace(line);
}

R Word::abs(void)const{ return CpmRoot::absT<string>(rep_);}
R Word::absSqr(void)const{ return CpmRoot::absSqrT<string>(rep_);}
Word Word::con(void)const{ return Word(CpmRoot::conT<string>(rep_));}
Word Word::inv(void)const{ return Word(CpmRoot::invT<string>(rep_));}
Word Word::rev(void)const{ return Word(CpmRoot::invT<string>(rep_));}

Word Word::test(Z complexity)const
{
   string s;
   string res=CpmRoot::testT<string>(s,complexity);
   return Word(res);
}

Word Word::ran(Z j)const
{
   string res=CpmRoot::ranT<string>(rep_,j);
         // rep_ was res till 2016-03-10
   return Word(res);
}

Z Word::hash()const
{
   return CpmRoot::hashT<string>(rep_);
}

R Word::dis(Word const& x)const
{
   return CpmRoot::disT<string>(rep_,x.rep_);
}

Word Word::net(Z i)const
{
   return Word(CpmRoot::netT<string>(rep_,i));
}

// functions CpmRoot::toStr

#define CPM_SC\
   ostringstream ost;\
   if (prc>0) ost.precision(prc);\
   ost<<x;\
   return ost.str()

string CpmRoot::toStr(R const& x, Z prc){CPM_SC;}
#if defined(CPM_ULM)
   string CpmRoot::toStr(rwrap const& x, Z prc){CPM_SC;}
#endif
#undef CPM_SC

#define CPM_SC ostringstream ost; ost<<x; return ost.str()
string CpmRoot::toStr(Z x){ CPM_SC; }
string CpmRoot::toStr(N x){ CPM_SC; }
string CpmRoot::toStr(L x){ CPM_SC; }
string CpmRoot::toStr(bool x){ CPM_SC; }
#undef CPM_SC

const char asciiList[]={
   '0','1','2','3','4','5','6','7','8','9',
   'a','b','c','d','e','f','g','h','i','j','k',
   'l','m','n','o','p','q','r','s','t','u','v',
   'w','x','y','z',
   'A','B','C','D','E','F','G','H','I','J','K',
   'L','M','N','O','P','Q','R','S','T','U','V',
   'W','X','Y','Z',
   '\x20','\x21','\x22','\x23','\x24','\x25','\x26','\x27',
   '\x28','\x29','\x2a','\x2b','\x2c','\x2d','\x2e','\x2f',
   '\x3a','\x3b','\x3c','\x3d','\x3e','\x3f','\x40',
   '\x5b','\x5c','\x5d','\x5e','\x5f','\x60',
   '\x7b','\x7c','\x7d','\x7e','\x7f'
};

Word Word::ascii()
{
   ostringstream ost;
   ost<<'0'<<'1'<<'2'<<'3'<<'4'<<'5'<<'6'<<'7'<<'8'<<'9'<<
   'a'<<'b'<<'c'<<'d'<<'e'<<'f'<<'g'<<'h'<<'i'<<'j'<<'k'<<
   'l'<<'m'<<'n'<<'o'<<'p'<<'q'<<'r'<<'s'<<'t'<<'u'<<'v'<<
   'w'<<'x'<<'y'<<'z'<<
   'A'<<'B'<<'C'<<'D'<<'E'<<'F'<<'G'<<'H'<<'I'<<'J'<<'K'<<
   'L'<<'M'<<'N'<<'O'<<'P'<<'Q'<<'R'<<'S'<<'T'<<'U'<<'V'<<
   'W'<<'X'<<'Y'<<'Z'<<
   '\x20'<<'\x21'<<'\x23'<<'\x24'<<'\x25'<<'\x26'<<
   '\x28'<<'\x29'<<'\x2a'<<'\x2b'<<'\x2c'<<'\x2d'<<'\x2e'<<'\x2f'<<
   '\x3a'<<'\x3b'<<'\x3c'<<'\x3d'<<'\x3e'<<'\x3f'<<'\x40'<<
   '\x5b'<<'\x5c'<<'\x5d'<<
   '\x7b'<<'\x7c'<<'\x7d'<<'\x7e';
   return Word(ost.str());
}
