//? cpmzinterval.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmzinterval.h>
#include <cpmtypes.h>

using CpmRoot::Z;
using CpmRoot::R;
using CpmRoot::Word;
using CpmRoot::randomR;
using CpmArrays::IvZ;
using CpmArrays::T2;
using namespace CpmStd;

/////////////////////////// class IvZ //////////////////////////////////

IvZ::IvZ(R a, R b)
   {
      Z i,j;
      if (a<=b) {i=cpmrnd(cpmfloor(a)); j=cpmrnd(cpmceil(b));}
      else {j=cpmrnd(cpmfloor(a)); i=cpmrnd(cpmceil(b));}
      *this=IvZ(i,j);
   }

Z IvZ::gap(IvZ const& iv)
{
   if (isVoid()) return 0;
   if (iv.isVoid()) return 0;
   if (r_<iv.l_) return iv.l_-r_;
   else if (iv.r_<l_) return l_-iv.r_;
   else return 0;
}

bool IvZ::contains(const IvZ& iv)const
{
   if (iv.isVoid()) return true;
      // the void set is subset of every set, even of the void set
   else{ // now iv is non-void
      if (isVoid()) return false;
         // a non-void set iv is never a subset of the void set
      else return l_<=iv.l_ && iv.r_<=r_;
   }
}

IvZ IvZ::operator|(IvZ const& iv)const
{
   if (iv.isVoid()) return *this;
      // the union with the void set is the original
   else{ // now iv is non-void
      if (isVoid()) return iv;
         // a non-void set iv is never a subset of the void set
      else { // now both *this and iv are non-void
         Z i1=inf();
         Z s1=sup();
         Z i2=iv.inf();
         Z s2=iv.sup();
         Z i=CpmRootX::inf<Z>(i1,i2);
         Z s=CpmRootX::sup<Z>(s1,s2);
         return IvZ(i,s);
      }
   }
}

IvZ IvZ::operator&(IvZ const& iv)const
{
   if (isVoid()||iv.isVoid()) return IvZ();
      // the section with the void set is the void set
   else{ // now both *this and iv are non-void
      Z s1=sup();
      Z i2=iv.inf();
      if (s1<i2) return IvZ();
      else{
         Z i1=inf();
         Z s2=iv.sup();
         if (i1>s2) return IvZ();
         else{ // now there is an overlap
             Z i=CpmRootX::sup<Z>(i1,i2);
             Z s=CpmRootX::inf<Z>(s1,s2);
             return IvZ(i,s);
         }
      }
   }
}

T2<IvZ> IvZ::setUnn(IvZ const& iv)const
{
   if (iv.isVoid()) return T2<IvZ>(IvZ(),*this);
   if (isVoid()) return T2<IvZ>(IvZ(),iv);
   if (contains(iv)) return T2<IvZ>(IvZ(),*this);
   if (iv.contains(*this)) return T2<IvZ>(IvZ(),iv);
   if (meet(iv).isVoid()){
      if (*this < iv) return T2<IvZ>(*this,iv);
      else return T2<IvZ>(iv,*this); // then iv<*this
   }
   return T2<IvZ>(IvZ(),join(iv));
}

T2<IvZ> IvZ::setDif(IvZ const& iv)const
{
   if (iv.isVoid()) return T2<IvZ>(*this,IvZ());
   if (isVoid()) return T2<IvZ>(IvZ(),IvZ());
   if (contains(iv)){
      IvZ ir=rigOf(iv.e());
      IvZ il=leftOf(iv.b());
      if (!ir.isVoid())
         return T2<IvZ>(il,ir);
      else
         return T2<IvZ>(ir,il);
   }
   if (iv.contains(*this)) return T2<IvZ>(IvZ(),IvZ());
   if (meet(iv).isVoid()){ return T2<IvZ>(IvZ(),*this);}
   if (iv<*this) return T2<IvZ>(IvZ(),rigOf(iv.e()));
   else return T2<IvZ>(IvZ(),leftOf(iv.b()));
}

Z IvZ::ranSel(Z j)const
{
   if (c_==0){
      cpmwarning("IvZ::ranSel(Z): instance is void, 0 returned");
      return 0;
   }
   else if (c_==1){
      return l_;
   }
   else{
      R r=randomR(j);
      Z res=l_+cpmtoz(r*c_);
      // (Z)(r*c_) takes values 0,1,...c_-1 with equal probabilities.
      // These are c_ values, which is OK.
      if (res>r_){
         cpmwarning("IvZ::ranSel(Z) proposed a value outside of *this, r_\
 returned");
         res=r_;
      }
      return res;
   }
}

// CPM_IO

bool IvZ::prnOn(ostream& str)const
{
   cpmwt("IvZ");
   cpmp(c_);
   cpmp(l_);
   return true;
}

bool IvZ::scanFrom(istream& str)
{
   cpms(c_);
   cpms(l_);
   ini_();
   return true;
}

// CPM_ORDER

Z IvZ::com(IvZ const& iv)const
{
   if (isVoid()){
      if (iv.isVoid()) return 0;
      else return 1;
   }
   if (l_<iv.l_) return 1;
   if (l_>iv.l_) return -1;
   if (r_<iv.r_) return 1;
   if (r_>iv.r_) return -1;
   return 0;
}

// CPM_DESCRIPTORS (not exactly since not virtual)

Word IvZ::toWord()const
{
   if (c_==0) return Word("IvZ()");
   ostringstream ost;
   ost<<"IvZ("<<l_<<","<<r_<<")";
   return Word(ost);
}

// CPM_TEST_X, functions already in CPM_TEST

IvZ IvZ::test(Z complexity)const{ return IvZ(complexity);}

Z IvZ::hash()const
{ return CpmRoot::hashT<Z>(l_)+CpmRoot::hashT<Z>(r_);}

namespace{
   R f(R x){ return x*2-1;}
   // values between -1 and +1
}

IvZ IvZ::ran(Z j)const
{
   R x=R(l_), y=R(r_), c=R(c_);
   R r1,r2,r3;
   if (j==0){
      r1=randomR();
      r2=randomR();
      r3=randomR();
   }
   else{
      Z j3=j*3;
      r1=randomR(j3);
      r2=randomR(j3+1);
      r3=randomR(j3+2);
   }
   if (r1<0.1 && r2>0.9) return IvZ();
      // here one gets 1 percent void sets on average
   x*=f(r1);
   y*=f(r2);
   c*=f(r3);
   return IvZ(x+c,y+c);
}

// CPM_TEST_X , functions not in CPM_TEST

R IvZ::dis(IvZ const& iv)const{ return (*this)==iv ? 0. : 1.;}
R IvZ::abs()const{ return R(c_);}
R IvZ::absSqr()const{ R c(c_); return c*c;}

// testing lattice idendities for functions meet and join

Z IvZ::testLatIdn(Z complexity)
   {
      IvZ o0;
      IvZ o1=o0.test(complexity);
      Z n=complexity;
      // Z n=cpmtoz(cpmsqrt(R(complexity)));
      // could be useful
      Z err=0;
      for (Z j=1;j<=n;++j){
         IvZ a=o1.ran();
         IvZ b=o1.ran();
         IvZ c=o1.ran();
         //cout<<a.toWord()<<endl;
         //cout<<b.toWord()<<endl;
         //cout<<c.toWord()<<endl;
         bool bb;
         bb=a.meet(b)==b.meet(a);
         if (!bb) err++;
         bb=a.join(b)==b.join(a);
         if (!bb) err++;
         bb=a.meet(b.meet(c))==(a.meet(b)).meet(c);
         if (!bb) err++;
         bb=a.join(b.join(c))==(a.join(b)).join(c);
         if (!bb) err++;
         bb=a.meet(a)==a;
         if (!bb) err++;
         bb=a.join(a)==a;
         if (!bb) err++;
         bb=a.meet(a.join(b))==a.join(a.meet(b));
         if (!bb) err++;
         bb=a.meet(a.join(b))==a;
         if (!bb) err++;
      }

      cout<<"IvZ::testLattice(Z), complexity="<<complexity<<
         ", err="<<err<<endl;
      return err;
   }
