//? cpmhistogr.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmhistogr.h>
#include <cpmgraph.h>
#include <cpmvectormore.h>

using CpmImaging::Histogram;
using CpmRoot::Z;
using CpmRoot::R;
using CpmRoot::Word;
using CpmLinAlg::R2;
using CpmArrays::R_Vector;
using CpmArrays::R_Matrix;
using CpmArrays::Vo;
using CpmFunctions::R_Func;
using CpmFunctions::F5;
using CpmGraphics::Frame;
using CpmGraphics::Graph;
using CpmGeo::Iv;

Histogram::Histogram(Z nPlanesI, Z nBinsI, R valMinI, R valMaxI,
   const Word& theme ):

   nPlanes(nPlanesI),
   range(valMinI,valMaxI),
   valMin(range[1]),
   valMax(range[2]),
   nBins(nBinsI),
   values(nPlanesI,nBinsI),
   topic(theme)
{
   cpmassert(nBins>0,"Histogram::Histogram(...)");
   binValues=range.centers(nBins);
   binWidth=range.abs()/nBins;
}

R2 Histogram::minMax(R min,R max)const
{
   Z i,h;
   Z l=nPlanes;
   R_Vector all(l);
   R_Vector bMin(l);
   R_Vector bMax(l);
   R_Vector minSect(l);
   R_Vector maxSect(l);
   R large=1e33;
   R minS=large;
   R maxS=-large;
   R2 minmax;
   R sum=0;

   for(i=1;i<=nBins;i++) for(h=1;h<=l;h++) all[h]+=values[h][i];
   for(h=1;h<=l;h++) bMin[h]=all[h]*(R(1)-min);
   for(h=1;h<=l;h++) bMax[h]=all[h]*(R(1)-max);

   for(h=1;h<=l;h++)
   {
      i=nBins;
      while(sum<bMin[h])
      {
         sum+=values[h][i];
         i--;
         minSect[h]=i;
      }
   }
   sum=0;
   for(h=1;h<=l;h++)
   {
      i=1;
      while(sum<bMax[h])
      {
         sum+=values[h][i];
         i++;
         maxSect[h]=i;
      }
   }
   for(h=1;h<=1;h++)
   {
      if(minSect[h]<minS) minS=minSect[h];
      if(maxSect[h]>maxS) maxS=maxSect[h];
   }
   minmax.x1=valMin+minS*binWidth;
   minmax.x2=valMin+maxS*binWidth;

   return minmax;
}

R2 Histogram::minMax(Z pl,R min,R max)const
{
   Z i;
   R all=0;
   R bMin;
   R bMax;
   R minS;
   R maxS;
   R2 minmax;
   R sum=0;

   for(i=1;i<=nBins;i++) all+=values[pl][i];
   bMin=all*(R(1)-min);
   bMax=all*(R(1)-max);
   i=nBins;
   while(sum<bMin){
      sum+=values[pl][i];
      i--;
      minS=i;
   }
   sum=0;
   i=1;
   while(sum<bMax)      {
      sum+=values[pl][i];
      i++;
      maxS=i;
   }
   minmax.x1=valMin+minS*binWidth;
   minmax.x2=valMin+maxS*binWidth;

   return minmax;
}

void Histogram::incrementValue(Z h,Z SectNum)
{
     values[h][SectNum]++;
}

/***********
void Histogram::show(Frame& fr, bool cumulative)const
{
   using namespace CpmImaging;
   const R xPosRel=0.6;
   R_Vector xV(binValues);
   Z i,j;
   R_Matrix valR(nPlanes,nBins);
      // since suitable graphical representation is available
      // for R_Vector and R_Matrix
   for (i=1;i<=nPlanes;i++)
      for (j=1;j<=nBins;j++)
         valR[i][j]=values[i][j];
   if (cumulative)
      for (i=1;i<=nPlanes;i++)
         valR[i]=CpmArrays::accumulate(valR[i]);
   V<Color> vc(nPlanes);
   if (nPlanes==1)
      vc[1]=RED;
   else if (nPlanes==2){
      vc[1]=RED;
      vc[2]=LIGHTGREEN;
   }
   else if (nPlanes==3){
      vc[1]=RED;
      vc[2]=LIGHTGREEN;
      vc[3]=LIGHTBLUE;
   }
   else vc=rainBow; // has only 8 colors, no problem since colors are
   // used
      // cyclically if there are less than nPlanes
   Graph g(fr);
   g.setColorActive(Color(Z(0),Z(0),Z(0)));
      // background
   g.setGridTextColor(Color(Z(200),Z(200),Z(200)));
      // grid text should be much brighter than the background
   g.show(xV,valR,vc,topic,xPosRel);
   fr=(Frame)g;
}
**********************/

void Histogram::mark(Frame& fr, bool cumulative)const
{
   using namespace CpmImaging;
   const R xPosRel=0.6;
   R_Vector xV(binValues);
   Z i,j;
   R_Matrix valR(nPlanes,nBins);
      // since suitable graphical representation is available
      // for R_Vector and R_Matrix
   for (i=1;i<=nPlanes;i++)
      for (j=1;j<=nBins;j++)
         valR[i][j]=values[i][j];
   if (cumulative)
      for (i=1;i<=nPlanes;i++)
         valR[i]=CpmArrays::accumulate(valR[i]);
   V<Color> vc(nPlanes);
   if (nPlanes==1)
      vc[1]=RED;
   else if (nPlanes==2){
      vc[1]=RED;
      vc[2]=LIGHTGREEN;
   }
   else if (nPlanes==3){
      vc[1]=RED;
      vc[2]=LIGHTGREEN;
      vc[3]=LIGHTBLUE;
   }
   else vc=rainBow; // has only 8 colors, no problem since colors are
   // used
      // cyclically if there are less than nPlanes
   Graph g(fr);
   g.setColorActive(Color(Z(0),Z(0),Z(0)));
      // background
   g.setGridTextColor(Color(Z(200),Z(200),Z(200)));
      // grid text should be much brighter than the background
   g.mark(xV,valR,vc,topic,xPosRel);
   fr=(Frame)g;
}

void Histogram::put(Z h, R value, Z Method)
{
   Z mL=4;
   static Word loc("Histogram::put(Z,R,Z");
   CPM_MA
   R x=R(nBins)*(value-valMin)/(valMax-valMin);
      // for value in the proper range, this varies from 0 to nBins
   if (value<valMin){
      if (Method==0) values[h][1]++;
   }
   else if (value>valMax){
      if (Method==0) values[h][nBins]++;
   }
   else{
      Z bin=range.qun(value,nBins,true);
         // Iv::qun(...) had to be modified 2015-01-19 inorder to
         // strictly forbid bin>nBins
      values[h][bin]++;
   }
   CPM_MZ
}

namespace{

R auxFunc(const R& x,
   const R& xL, const R& xU, const R& yL, const R& yU, const R& ex)
{
   if (x<=xL) return yL;
   else if (x>xU) return yU;
   else{
      R p=(x-xL)/(xU-xL);
      p=cpmpow(p,ex);
      return yL+p*(yU-yL);
   }
}

}

R_Func Histogram::getBrightValues(Z pl, R fracL, R fracU, R p)const
{
   R yL=0;
   R yU=255;
   R2 mm=minMax(pl,fracL,fracU);
   R xL=mm[1];
   R xU=mm[2];
   return F5<R,R,R,R,R,R,R>(xL,xU,yL,yU,p)(auxFunc);

}

Iv Histogram::firstPeak(Z pl, R frac, Z nL, Z nR)
{
   static Word loc("Histogram::firstPeak(Z pl, R frac, Z nL, Z nR)");
   Vo<Z> val=values[pl];
   Z n=val.dim();
   cpmassert(n==nBins,loc);
   Z valMax=val.sup();
   Z valOk=cpmrnd(R(valMax)*frac);
   Z i,i0=0;
   for (i=1;i<n;i++){
      if (val[i+1]<=val[i] && val[i]>=valOk){
         i0=i;
         break;
      }
   }
   R bhw=binWidth*0.5;
   Z iL=CpmRootX::sup(i0-nL,Z(1));
   Z iR=CpmRootX::inf(i0+nR,n);
   return Iv(binValues[iL]-bhw,binValues[iR]+bhw);
}

