//? cpmrecordhandler.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmrecordhandler.h>
#include <cpmsystem.h>
#include <cpmfile.h>
#include <cpmapplication.h>

using CpmRoot::Z;
using CpmRoot::R;
using CpmRoot::Word;
using CpmRootX::Z1;
using CpmRootX::R1;
using CpmRootX::B;
using CpmRootX::RecordHandler;
using CpmRootX::Record;
using CpmRootX::sup;
using CpmSystem::IFileStream;
using CpmSystem::glue;
using CpmSystem::Message;
using CpmArrays::V;
using namespace std;

Word CpmApplication::iniFile_("defaultinifile");

V<Word> CpmRootX::combine(V<Word> const& argsC, V<Word> const& argsF)
{
   if (!argsC.valInd(1)) return V<Word>();
   if (!argsF.valInd(1)) return argsC;
   V<Word> argsIni=argsF.prepend(argsC[1]);
   Z nMax=sup<Z>(argsIni.dim(),argsC.dim());
   V<Word> res(nMax);
   for (Z i=res.b();i<=res.e();++i){
      res[i]= argsC.valInd(i) ? argsC[i] : argsIni[i];
   }
   return res;
}

V<Word> CpmRootX::extComLine(int argc, char *argv[])
{
   Z mL=1;
   Word loc("CpmRootX::extComLine(int,*char[])");
   CPM_MA
   V<Word> args=CpmArrays::comLine(argc, argv);
   cpmassert(args.valInd(1),loc);
   Word prog=args[1].baseName(); // important, since args_[1] carries the
      // full path
      // prog may read vqm2MS, or vqm2_MS, or vqm2MINGW, or vqm2CPM_R,
      // or vqm2CPM_RLONG,or vqm2CPM_RMP for program vqm2.
      // So one should also remove the suffixes which
      // the user may have added to the executables name in order to
      // allow comparison of various compilers. For this to work it is
      // important that the program name proper does not contain uppercase
      // letters at its end. I once had a program name graphicsTest which
      // made me wonder where the request for a file graphics.ini came
      // from. Since the migration to Code::Blocks and Linux I allow
      // program names such as smallExperiments which is OK, since here
      // the isolated upper case letter is followed by a lower case one.
   Word iniFile=prog.remApp()&".ini";
   RecordHandler rch(iniFile);
   if (!rch.isVal()){
      cpmwarning(loc&
         ": RecordHandler not valid, commandline not extended");
      CPM_MZ
      return args;
      // Thus no action if file fn does not exist
   }
   CpmApplication::iniFile_=iniFile;
// Now we have a valid RecordHandler to get data from
   V<Word> argsIni=rch.comLine();
   V<Word> res=combine(args,argsIni);
   CPM_MZ
   return res;
}

namespace{
   Word makeIni(Word const& w)
   // appends the extension ".ini" if w is a filename without an extension
   {
      Word res(w);
      Word p(".");
      if (w.find(p)==0) res&=".ini";
      return res;
   }
}

void RecordHandler::ini_(V<Word> const& files, Word const& path,
                         Word const& callMes)
{
   Z mL=1;
   Word pathAct = path=="fromCon" ? Message::getInpDir() : path;
   ostringstream ost;
   ost<<"files="<<files<<" path="<<pathAct<<" callMessage="
   <<callMes<<endl;
   Word mes(ost);
   Word loc="RecordHandler::ini_("&mes&")";
      // probably long, so not use too often
   CPM_MA
      // it is important to see the arguments
   rank=CpmMPI::Cpm_com.getRank();
   size=CpmMPI::Cpm_com.getSize();
   if (rank==1){
      Z i,fd=files.dim();
      V<Word> ffs(fd);
      for (i=1;i<=fd;++i) ffs[i]=glue(pathAct,makeIni(files[i]));
      bool binary=true;
      i=1;
      while(!valid && i<=fd){ // B valid is data member,
            // thus initialized as false
         valid=rec.readFromFile(ffs[i],binary);
         filename=ffs[i];
         cpmdebug(filename);
         i++;
      }
      if (!valid){
         cpmwarning(" could not read any of the "&cpm(fd)&" files");
         if (size>1) cpmwarning("nothing sent to other processes");
      }
      else{
         if (size>1 || mL<=1){
            if (size>1) cpmmessage(mL,"Rank 1 to all other processes:");
            else cpmmessage(mL,"this is what we got:");
            ostringstream ost;
            rec.writeFormatted(ost,false);
            cpmmessage(mL,Word(ost));
         }
      }

   } // rank==1
   CpmMPI::Cpm_com.bcast(rec);
   if (rank!=1) {
      cpmmessage(mL,"Rank "&cpm(rank)&", got this from rank 1:");
      ostringstream ost;
      rec.writeFormatted(ost,false);
      cpmmessage(mL,Word(ost));
      valid=rec.isVal();
   }
   cpmmessage(": valid="&cpm(valid));
   CPM_MZ
}

///////////////////////// read ///////////////////////////////////////////

bool RecordHandler::read(const Word& section, const Word& key,
   Z& value, Z pedantic)const
{
   bool b=rec.get(section,key,value,pedantic);
      // creates a warning if does not find
      // Z changed only if get successful
   if (b==false && pedantic>0){
      ostringstream ost;
      ost<<string("RecordHandler::read(): not able to read a Z")<<
      string(" in section=")<<section.toStr()<<string(" at key=")<<
      key.toStr();
      // very similar to the warning from get. But good to have
      // the text here too since it comes to the status line
      if (pedantic==2) cpmerror(Word(ost.str()));
   }
   return b;
}

bool RecordHandler::read(const Word& section, const Word& key,
   Z1& value, Z pedantic)const
{
   Z val; // using val fot type conversion
   // protection of value against changes in unsuccssful
   // reading is given anyway
   bool b=rec.get(section,key,val,pedantic);
   if (b==false && pedantic>0){
      ostringstream ost;
      ost<<string("RecordHandler::read(): not able to read a Z1")<<
      string(" in section=")<<section.toStr()<<string(" at key=")<<
      key.toStr();
      if (pedantic==2) cpmerror(Word(ost.str()));
   }
   else value=Z1(val);
   return b;
}

bool RecordHandler::read(const Word& section, const Word& key,
   R1& value, Z pedantic)const
{
   R val;
   bool b=rec.get(section,key,val,pedantic);
   if (b==false && pedantic>0){
      ostringstream ost;
      ost<<string("RecordHandler::read(): not able to read a R1")<<
      string(" in section=")<<section.toStr()<<string(" at key=")<<
      key.toStr();
      if (pedantic==2) cpmerror(Word(ost.str()));
   }
   else value=R1(val);
   return b;
}

/******************
bool RecordHandler::read(const Word& section, const Word& key,
   Rh& value, Z pedantic)const
{
   R val;
   bool b=rec.get(section,key,val,pedantic);
   if (b==false && pedantic>0){
      ostringstream ost;
      ost<<string("RecordHandler::read(): not able to read a Rh")<<
      string(" in section=")<<section.toStr()<<string(" at key=")<<
      key.toStr();
      if (pedantic==2) cpmerror(Word(ost.str()));
   }
#if defined(CPM_MP)
   else value=CpmRoot::toRh(val);
#elif defined(CPM_ULM)
   else value=CpmRoot::toRh(val);
#else
   else value=Rh(val);
#endif
   return b;
}
***********************/

bool RecordHandler::read(const Word& section, const Word& key,
   R& value, Z pedantic)const
{
   bool b=rec.get(section,key,value,pedantic);
   if (b==false && pedantic>0){
      ostringstream ost;
      ost<<string("RecordHandler::read(): not able to read a R")<<
      string(" in section=")<<section.toStr()<<string(" at key=")<<
      key.toStr();
      if (pedantic==2) cpmerror(Word(ost.str()));
   }
   return b;
}

bool RecordHandler::read(const Word& section, const Word& key,
   Word& value, Z pedantic)const
{
   bool b=rec.get(section,key,value,pedantic);
   if (b==false && pedantic>0){
      ostringstream ost;
      ost<<string("RecordHandler::read(): not able to read a Word")<<
      string(" in section=")<<section.toStr()<<string(" at key=")<<
      key.toStr();
      if (pedantic==2) cpmerror(Word(ost.str()));
   }
   return b;
}

bool RecordHandler::read(const Word& section, const Word& key,
   bool& value, Z pedantic)const
{
   bool b=rec.get(section,key,value,pedantic);
   if (b==false && pedantic>0){
      ostringstream ost;
      ost<<string("RecordHandler::read(): not able to read a bool")<<
      string(" in section=")<<section.toStr()<<string(" at key=")<<
      key.toStr();
      if (pedantic==2) cpmerror(Word(ost.str()));
   }
   return b;
}

bool RecordHandler::read(const Word& section, const Word& key,
   B& value, Z pedantic)const
{
   bool val;
   bool b=rec.get(section,key,val,pedantic);
   if (b==false && pedantic>0){
      ostringstream ost;
      ost<<string("RecordHandler::read(): not able to read a B")<<
      string(" in section=")<<section.toStr()<<string(" at key=")<<
      key.toStr();
      if (pedantic==2) cpmerror(Word(ost.str()));
   }
   value=B(val);
   return b;
}

bool RecordHandler::read(const Word& section, const Word& key,
   V<bool>& value, Z pedantic)const
{
   bool b=rec.get(section,key,value,pedantic);
   if (b==false && pedantic>0){
      ostringstream ost;
      ost<<string("RecordHandler::read(): not able to read Bs")<<
      string(" in section=")<<section.toStr()<<string(" at key=")<<
      key.toStr();
      if (pedantic==2) cpmerror(Word(ost.str()));
   }
   return b;
}

bool RecordHandler::read(const Word& section, const Word& key,
   V<Z>& value, Z pedantic)const
{
   bool b=rec.get(section,key,value,pedantic);
   if (b==false && pedantic>0){
      ostringstream ost;
      ost<<string("RecordHandler::read(): not able to read Zs")<<
      string(" in section=")<<section.toStr()<<string(" at key=")<<
      key.toStr();
      if (pedantic==2) cpmerror(Word(ost.str()));
   }
   return b;
}

bool RecordHandler::read(const Word& section, const Word& key,
   V<R>& value, Z pedantic)const
{
   bool b=rec.get(section,key,value,pedantic);
   if (b==false && pedantic>0){
      ostringstream ost;
      ost<<string("RecordHandler::read(): not able to read Rs")<<
      string(" in section=")<<section.toStr()<<string(" at key=")<<
      key.toStr();
      if (pedantic==2) cpmerror(Word(ost.str()));
   }
   return b;
}

bool RecordHandler::read(const Word& section, const Word& key,
   V<Word>& value, Z pedantic)const
{
   bool b=rec.get(section,key,value,pedantic);
   if (b==false && pedantic>0){
      ostringstream ost;
      ost<<string("RecordHandler::read(): not able to read Ws")<<
      string(" in section=")<<section.toStr()<<string(" at key=")<<
      key.toStr();
      if (pedantic==2) cpmerror(Word(ost.str()));
   }
   return b;
}

///////////////////////// write ////////////////////////////////////////

bool RecordHandler::write(const Word& section, const Word& key,
   const Z& value)
{
   bool b=rec.set(section,key,value);
   if (b==false)
      cpmerror("RecordHandler::write(Z): failed at "&section&" "&key);
   return b;
}

bool RecordHandler::write(const Word& section, const Word& key,
   const Z1& value)
{
   bool b=rec.set(section,key,value);
   if (b==false)
      cpmerror("RecordHandler::write(Z1): failed at "&section&" "&key);
   return b;
}

bool RecordHandler::write(const Word& section, const Word& key,
   const R& value)
{
   bool b=rec.set(section,key,value);
   if (b==false)
      cpmerror("RecordHandler::write(R): failed at "&section&" "&key);
   return b;
}

bool RecordHandler::write(const Word& section, const Word& key,
   const R1& value)
{
   bool b=rec.set(section,key,value);
   if (b==false)
      cpmerror("RecordHandler::write(R1): failed at "&section&" "&key);
   return b;
}

bool RecordHandler::write(const Word& section, const Word& key,
   const Word& value)
{
   bool b=rec.set(section,key,value);
   if (b==false)
      cpmerror("RecordHandler::write(Word): failed at "&section&" "&key);
   return b;
}

bool RecordHandler::write(const Word& section, const Word& key,
   const bool& value)
{
   bool b=rec.set(section,key,value);
   if (b==false)
      cpmerror("RecordHandler::write(bool): failed at "&section&" "&key);
   return b;
}

bool RecordHandler::write(const Word& section, const Word& key,
   const B& value)
{
   bool bVal=value;
   bool b=rec.set(section,key,bVal);
   if (b==false)
      cpmerror("RecordHandler::write(B): failed at "&section&" "&key);
   return b;
}

bool RecordHandler::write(const Word& section, const Word& key,
   const V<bool>& value)
{
   bool b=rec.set(section,key,value);
   if (b==false)
      cpmerror("RecordHandler::write(V<bool>): failed at "&section&
         " "&key);
   return b;
}

bool RecordHandler::write(const Word& section, const Word& key,
   const V<Z>& value)
{
   bool b=rec.set(section,key,value);
   if (b==false)
      cpmerror("RecordHandler::write(V<Z>): failed at "&section&" "&key);
   return b;
}

bool RecordHandler::write(const Word& section, const Word& key,
   const V<R>& value)
{
   bool b=rec.set(section,key,value);
   if (b==false)
      cpmerror("RecordHandler::write(V<R>): failed at "&section&" "&key);
   return b;
}

bool RecordHandler::write(const Word& section, const Word& key,
   const V<Word>& value)
{
   bool b=rec.set(section,key,value);
   if (b==false)
      cpmerror("RecordHandler::write(V<Word>): failed at "
         &section&" "&key);
   return b;
}

bool RecordHandler::writeFormatted(ostream& str,bool writeComments)const
{
   return rec.writeFormatted(str,writeComments);
}

V<Word> RecordHandler::comLine()const
{
   Word sec1("command line"); // preferred style
   Word sec2("command_line"); // for compatibility
   Word key("args");
   V<Word> res;
   bool valid=false;
   Z ped=1; // only warn
   valid=read(sec1,key,res,ped);
   if (valid) return res;
   valid=read(sec2,key,res,ped);
   if (valid) return res;
   return V<Word>();
}
