//? cpmrfuncrec.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmrfuncrec.h>

   using namespace CpmRoot;
   using namespace CpmRootX;
   using namespace CpmArrays;
   using namespace CpmSystem;
   using namespace CpmFunctions;
   using namespace CpmGeo;

bool R_FuncRec::vadjust(void)const
{
   Z mL=3;
   Word loc="R_FuncRec::vadjust()";
   CPM_MA
   Word sec="R_FuncRec";

   Z ipolMet=1, ipolOrd=1;
   R polyFac=0,bernsteinFac=0,bernsteinX=0, bernsteinY=0,
      bezierFac=0,ipolFac=0,gaussFac=0,indFac=0;
   V<R> poly,bernstein,bezierX,bezierY,ipolX,ipolY,
      gaussX,gaussY,gaussSigma,indX,indY,indWidth;

// efficiency asks for ordering according to increasing type complexity
// and alphabetic order within a type (as is obeyed in the following)
// Reading gives only a warning for missing entries.
// type Z
   cpmrc(ipolMet);
   cpmrc(ipolOrd);

// type R
   cpmrc(bernsteinFac);
   cpmrc(bernsteinX);
   cpmrc(bernsteinY);
   cpmrc(bezierFac);
   cpmrc(gaussFac);
   cpmrc(indFac);
   cpmrc(ipolFac);
   cpmrc(polyFac);

// type V<R>
   cpmrc(bernstein);
   cpmrc(bezierX);
   cpmrc(bezierY);
   cpmrc(gaussSigma);
   cpmrc(gaussX);
   cpmrc(gaussY);
   cpmrc(indWidth);
   cpmrc(indX);
   cpmrc(indY);
   cpmrc(ipolX);
   cpmrc(ipolY);
   cpmrc(poly);

   Z i;
   if (polyFac !=0.){ // polynomial contribution
      V<R> pcHead{0.,1.};
         // The first two components have a special meaning!
         // No rescaling of the polynomial here since this
         // should be done for the whole function.
      poly=pcHead&poly;
      R_Vector rpoly(poly);
      R_Func fPoly=R_Func::polyFunc(rpoly)*polyFac;
      f_+=fPoly;
   }
   if (bernsteinFac !=0){
         R_Vector rbernstein(bernstein);
         R2 r2(bernsteinX,bernsteinY);
         f_+=R_Func::Bernstein(rbernstein,r2)*bernsteinFac;
   }
   if (bezierFac!=0){
      Z bX=bezierX.dim();
      Z bY=bezierY.dim();
      cpmassert(bX==bY,loc);
         // not needed for R_Func::Bezier(rbX,rbY) to work
         // but probably violated only unintentionally
      R_Vector rbX(bezierX);
      R_Vector rbY(bezierY);
      f_+=R_Func::Bezier(rbX,rbY)*bezierFac;
   }
   if (ipolFac!=0){
      IpolControl ic(ipolOrd+1,ipolMet);
         // this class does not expect the order but the number of
         // adjustable parameters, which always is higher by one
      f_+=R_Func(ipolX,ipolY,ic)*ipolFac;
   }
   if (gaussFac!=0){
      Z nX=gaussX.dim();
      Z nSigma=gaussSigma.dim();
      Z nY=gaussY.dim();
      cpmassert(nX==nY,loc);
      cpmassert(nX==nSigma,loc);
      Z nMin=CpmRootX::inf<Z>(nX,nY,nSigma);
      R_Func fGauss;
      bool unitHeight=true;
      for (i=1;i<=nMin;++i){
         fGauss+=R_Func::Gauss(gaussX[i],gaussSigma[i],unitHeight)
            *gaussY[i];
      }
      fGauss*=gaussFac;
      f_+=fGauss;
   }
   if (indFac!=0){
      Z nX=indX.dim();
      Z nWidth=indWidth.dim();
      Z nY=indY.dim();
      cpmassert(nX==nY,loc);
      cpmassert(nX==nWidth,loc);
      Z nMin=CpmRootX::inf<Z>(nX,nY,nWidth);
      R_Func fInd;
      for (i=1;i<=nMin;++i){
         R xi=indX[i];
         R hwi=indWidth[i]*0.5;
         Iv ii(xi-hwi,xi+hwi);
         fInd+=R_Func::charFunc(ii)*indY[i];
      }
      fInd*=indFac;
      f_+=fInd;
   }
   CPM_MZ
   return true;
}
