//? cpmcamera2cpp.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/////////////////// class Camera ///////////////////////////////////

namespace CpmCamera2{

using namespace CpmStd;
using namespace CpmColCam;
using CpmDim2::AxVec;
using CpmDim2::Vec;
using CpmDim2::Spc;
using CpmDim2::RefSys;
using CpmDim2::Group;
using CpmDim2::SubsetOfSpace;
using CpmDim2::Skeleton;
using CpmDim2::WireFrame;
using CpmDim2::Texture;
using CpmImaging::RED;

void Camera::complete(void)
// s,frac,e1,e2,poi given
{}

V<Vec> Camera::getVectors()const
{
   V<Vec> res(2);
   res[1]=h;
   res[2]=v;
   return res;
}

Camera::Camera(R imageHeight, R aspectRatio, const Angle& viewAngle,
   R frac_, const Angle& conv, const Frame& fr):
   color1(WHITE), color3(STEREOBLACK),
   h(1,0), v(0,1),poi(),gr(fr)
{
   viewAngle; // not used
   frac_;
   conv;
   R width=imageHeight*aspectRatio;
   gr.setXY(-R(0.5)*width,-R(0.5)*imageHeight,width,imageHeight);
   complete();
}

Camera::Camera(Graph const& g):
   color1(WHITE), color3(BLACK),
   gr(g), poi(), h(1,0), v(0,1)
{
   complete();
}

bool Camera::prnOn(ostream& out)const
{
   if (!CpmRoot::writeTitle("Camera",out)) return false;
   if (!color1.prnOn(out)) return false;
   if (!color3.prnOn(out)) return false;
   if (!poi.prnOn(out)) return false;
   if (!v.prnOn(out)) return false;
   if (!h.prnOn(out)) return false;
   if (!gr.prnOn(out)) return false;
   return true;
}

bool Camera::scanFrom(istream& in)
{
   if (!color1.scanFrom(in)) return false;
   if (!color3.scanFrom(in)) return false;
   if (!poi.scanFrom(in)) return false;
   if (!v.scanFrom(in)) return false;
   if (!h.scanFrom(in)) return false;
   if (!gr.scanFrom(in)) return false;
   complete();
   return true;
}

Camera& Camera::operator *=(const Group& g)
{
   poi*=g; h*=g; v*=g; complete();
      // using complete here assures constancy of geometry even in the
      // presence
      // of accumulating roundoff errors
   return *this;
}

void Camera::shift(const R& sx, const R& sy, const R& sz)
{
   sz;
   Vec tr=h*sx+v*sy;
   Group g(tr);
   operator *=(g);
//   gr.shiftAbs(sx,sy);
   // looks natural but seems not to do what was expected in my
   // present context (02-09-09) so I don't include it in order
   // not to create side effects
}

void Camera::tilt(const Angle& a)
{
   a;
}

void Camera::pan(const Angle& a)
{
   a;
}

void Camera::roll(const Angle& a)
{
   AxVec av(a);
   Group g(poi,av);
   operator *=(g);
}

void Camera::adjustFieldOfView( R enlarge, R sx, R sy, R sz)
{
   sz; // not used
   gr.scale(enlarge);
   R sxa=gr.getdX();
   R sya=gr.getdY();
   sxa*=sx;
   sya*=sy;
   shift(sxa,sya,0);
}

void Camera::scale(R enlarge)
{
   gr.scale(enlarge);
}

T2<C> Camera::capture(const Spc& p)const
{
   C res=captureMono(p);
   return T2<C>(res,res);
}

T2<X2<C,R> > Camera::capture(const Spc& p,
   const R& r)const
{
   C res=captureMono(p);
   typedef X2<C,R> prr2;
   prr2 px(res,r);
   return T2<prr2>(px,px);
}

X2<C,R> Camera::captureMono(const Spc& p, const R& r)const
{
   C pr=captureMono(p);
   return X2<C,R>(pr,r);
}

void Camera::mark(const Spc& x, R r)const
{
   C pp=captureMono(x);
   fillCircle(pp,r,getColor1());
}

void Camera::mark(const Spc& x, R r, const Color& c)const
{
   X2<C,R> prr=captureMono(x,r);
   fillCircle(prr.c1(),prr.c2(),c);
}

bool Camera::mark(const Spc& x, const Color& c)const
{
   const Z size=4;
   C px=captureMono(x);
   if (!c.visible()){
      gr.mark(px,size,color1);
   }
   else{
      gr.mark(px,size,c);
   }
   return gr.indFunc(R2(px[1],px[2]));
}

void Camera::mark(const Spc& x1, const Spc& x2, R rFoot)const
{
   C a=captureMono(x1);
   C b=captureMono(x2);
   gr.draw(R2(a[1],a[2]),R2(b[1],b[2]),color1);
   if (rFoot>0){
      fillCircle(a,rFoot,color1);
   }
}

void Camera::mark(const Spc& x1, const Spc& x2, const Color& c,
   R rFoot)const
   // shows a line from x1 to x2 in color c (as seen by camera
   // *this in 1-direction)
{
   if ( !c.visible() ){
      mark(x1,x2,rFoot);
   }
   else{
      C a=captureMono(x1);
      C b=captureMono(x2);
      gr.draw(R2(a[1],a[2]),R2(b[1],b[2]),c);
      if (rFoot>0) fillCircle(a,rFoot,c);
   }
}

void Camera::mark(const RefSys& rs, Z i)const
{
   if (cpmtrigger>0){
      cpmdebug(rs);
      cpmdebug(i);
      cpmtrigger=0;
   }
   const R step=2e-6;
   Spc p=rs.o();
   Vec sh;
   if (i==1) sh=rs.e1();
   else if (i==2) sh=rs.e2();
   else if (i==-1) sh=-rs.e1();
   else if (i==-2) sh=-rs.e2();
   else cpmerror("Camera::mark(RefSys,Z i), i out of range");
   sh*=step;
   Color c(Z(-1));
   while (isVisible(p)){p+=sh; mark(p,c);}
}

void Camera::dirTo_(Spc const& x, R const& r, R fitFrame,
   Angle tilt, Angle pan, RefSys const& rs)
{
   tilt;
   pan;
   RefSys rsOrig=getRefSys();
   Group g=rs/rsOrig;
   operator *=(g); // now the camera has standard orientation with
      //respect to system rs
   Vec sh=x-poi;
   operator *=(Group(sh));
   R rAct= r<=0 ? R(1) : r;
   R imageHeight=rAct*fitFrame*2;
   R aspectRatio=getdX()/getdY();
   R width=imageHeight*aspectRatio;
   gr.setXY(-R(0.5)*width,-R(0.5)*imageHeight,width,imageHeight);
   complete();
}

void Camera::show(const SubsetOfSpace& sos, const Texture& tx,
   bool bodyFix)const
{
    using CpmArrays::VV;
    using CpmLinAlg::Z2;
    using namespace CpmGraphics;

    Z nx=dim1();
    Z ny=dim2();
    R dx=getdX()/nx;
    R dy=getdY()/ny;

    R xStart=getxLeft()+dx*0.5;
    R yStart=getyUpper()-dy*0.5;
         // (x,y) pixel center of the first (top-left) film pixel
         // in object space
    R x=xStart;
    R y;
    Z i0=gr.a0();
    Z j0=gr.b0();
    R nxInv=1./nx;
    for (Z i=0;i<nx; i++){
       cpmprogress("capture",R(i)*nxInv,3);
       y=yStart;
       RefSys sosrs=sos.getRefSys();
       for (Z j=0; j<ny; j++){
          Spc sp=poi+h*x+v*y;
          bool b=sos.ind(sp);
          R br;
          if (bodyFix){
            br=b ?  tx.eval(sosrs,sp) : R(0);
          }
          else{
             br=b ?  tx(sp) : R(0);
          }
          Color c;
          if (br==0 ) c=color3; else c=br*color1;
          gr.setPel(i+i0,j+j0,c);
          y-=dy;
       }
       x+=dx;
    }
    gr.vis();
}

#include <cpmcameraxcpp.h>
   // common part of the implementation

}//namespace

