//? cpmdim.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_DIM_H_
#define CPM_DIM_H_
/*
Description: Common basis for 2D and 3D geometry
*/

#include <cpmlinalg.h>
#include <cpmangle.h>
#include <cpminterval.h>
#include <cpmvo.h>
#include <cpmx.h>
#include <cpmr2func.h>
#include <cpmfield.h>
#include <cpmvectorgraphics.h>

namespace CpmColCam{ // color camera
   // lean namespace for declarations in the range of geometry-related
   // classes
   using CpmRoot::Z;
   using CpmRoot::R;
   using CpmRoot::C;
   using CpmRoot::Word;

   using CpmGeo::Angle;
   using CpmGeo::Iv; // Interval
   using CpmGeo::DiGraph; // directed graph
   using CpmGeo::RAD;
   using CpmGeo::DEG;

   using CpmRootX::R1;
   using CpmRootX::Z1;

   using CpmLinAlg::Z2;
   using CpmLinAlg::Z3;
   using CpmLinAlg::R2;
   using CpmLinAlg::R3;
   using CpmLinAlg::RR2;
   using CpmLinAlg::RR3;

   using CpmFunctions::F;
   using CpmFunctions::F1;

   using CpmArrays::V;
   using CpmArrays::X2;
   using CpmArrays::X4;
   using CpmArrays::Vo;
}

namespace CpmDim2{

   using namespace CpmStd;
   using namespace CpmColCam;

   typedef R2 Rdim;
      // space of components of polar vectors
   typedef Z2 Zdim;
      // related to lattice points in space
   typedef RR2 RRdim;
      // space of corresponding matrices
   typedef R1 Rdim_;
      // hyperplane in Rdim 'R(dim-1)'
   const Z dimVec=2;
   const Z dimAxVec=1;
   #include <cpmdim2.h>
   #include <cpmdimx.h>
}
namespace CpmDim3{

   using namespace CpmStd;
   using namespace CpmColCam;

   typedef R3 Rdim;
      // space of components of axial vectors
   typedef Z3 Zdim;
      // related to lattice points in space
   typedef RR3 RRdim;
      // space of corresponding matrices
   typedef R2 Rdim_;
      // hyperplane in Rdim 'R(dim-1)'
   const Z dimVec=3;
   const Z dimAxVec=3;
   #include <cpmdim3.h>
   #include <cpmdimx.h>
}

#endif
