//? cpmoperator.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmoperator.h>

using CpmQuantumMechanics::QMBase;
using CpmQuantumMechanics::State;
using CpmQuantumMechanics::Operator;
using CpmQuantumMechanics::Observable;
using CpmQuantumMechanics::TrjCon;
using CpmQuantumMechanics::TrjB;
using CpmFunctions::F1;
using CpmFunctions::F2;
using CpmFunctions::R_Func;
using CpmFunctions::R2_Func;

//////////////////////// class Operator //////////////////////////

namespace{

   State xsrobs(X2<State,R> const& xsr, Observable const& ob)
   {
      return ob*xsr.c1();
   }

#define CPM_SC\
      static R hD=QMBase::get_hDirac();\
      static R hD2=hD*hD;\
      State psi=xsr.c1();\
      R t=xsr.c2();\
      Z d=psi.dim();\
      R dx=R(1)/R(d);\
      R hInv2=R(d)/QMBase::getSize();\
      hInv2*=hInv2;\
      R fac=-hD2*hInv2/(QMBase::getMass()*2);\
      State res(d);\
      R xi=dx*0.5 -0.5;

   State xsrrfs(X2<State,R> const& xsr, R_Func const& pot)
   {
      CPM_SC
      for (Z i=0;i<d;++i){
         C val=psi[i-1]-psi[i]*2+psi[i+1];
         res[i]=val*fac+pot(xi)*psi[i];
         xi+=dx;
      }
      return res;
   }

   State xsrrfs2(X2<State,R> const& xsr, R2_Func const& pot)
   {
      CPM_SC
      for (Z i=0;i<d;++i){
         C val=psi[i-1]-psi[i]*2+psi[i+1];
         res[i]=val*fac+pot(C(xi,t))*psi[i];
         xi+=dx;
      }
      return res;
   }

#undef CPM_SC

}// namespace

Operator::Operator(Observable const& ob):d_(ob.dim())
{
   f_=F1<XSR,Observable,State>(ob)(xsrobs);
}

namespace{
   State sinParHamFunc(X2<State,R> const& st, R2_Func const& pot,
      R const& lambda)
   { return st.c1().sinParHam(pot,st.c2(),lambda);}

   State freeParFunc(State const& st)
   { return st.freeParHam();}
}

Operator Operator::sinParHam(R2_Func const& pot,Z d, R lambda)
{
   return
   Operator(d,F2<XSR,R2_Func,R,State>(pot,lambda)(sinParHamFunc));
}

Operator Operator::freePar(Z d)
{
   return Operator(d,F<State,State>(freeParFunc));
}

Operator::Operator(Z d, F<State,State> const& f):d_(d)
{
   f_=F1<XSR,F<State,State>,State>(f)(conInTime);
}

Operator Operator::operator*(Operator const& op)const
{
   return
   Operator(d_,F2<XSR,Operator,Operator,State>(*this,op)(mul2));
}

Operator Operator::operator+(Operator const& op)const
{
   return
   Operator(d_,F2<XSR,Operator,Operator,State>(*this,op)(add2));
}

Operator Operator::operator-(Operator const& op)const
{
   return
   Operator(d_,F2<XSR,Operator,Operator,State>(*this,op)(dif2));
}

Operator Operator::operator*(C z)const
{
   return
   Operator(d_,F2<XSR,Operator,C,State>(*this,z)(mulC));
}

Operator Operator::operator*(R r)const
{
   return
   Operator(d_,F2<XSR,Operator,R,State>(*this,r)(mulR));
}

Operator& Operator::operator+=(Operator const& fop)
{
   return *this=operator+(fop);
}

Operator& Operator::operator-=(Operator const& fop)
{
   return *this=operator-(fop);
}

Operator& Operator::operator*=(C z)
{
   return *this=operator*(z);
}

Operator& Operator::operator*=(R r)
{
   return *this=operator*(r);
}

Observable Operator::toObservable(R t)const
{
   Observable res(d_);
   for (Z i=0;i<d_;++i){
      State vsi= actOn(State(d_,i),t);
  //   res.set(i,vsi);
      res.c().cui(i)=vsi.cv();
   }
   return res;
}

//////////////////////// class TrjTD ///////////////////////////////
/*
void TrjTD::adjustTD_()
{
   State opx=op_.actOn(x_,t_);
   enr_=(x_|opx);
   v_=opx*QMBase::get_ihDiracInv();
   R n1=x_.norm();
   n_1_=n1-1;
   R n2=v_.norm();
   dt_=QMBase::step(n1,n2);
   text_="Evolution by the Time Dependent Direct Midpoint Integrator";
}

C TrjTD::eTot()const
{
   return (x_|v_)*QMBase::get_ihDirac();
}

TrjTD::TrjTD(const Operator& a, const State& x, TrjCon const& tc):
TrjB(x,tc),op_(a)
{
   adjustTD_();
}

void TrjTD::step_(R h)
{
   static C fac=QMBase::get_ihDiracInv()*QMBase::get_ihDiracInv();

   if (autoStep()){ // due to time dependence of
      // the generator, the proposed time step will vary with time
      // and has to be computed for each step (rather cheaply)
      R n1=x_.norm();
      R n2=v_.norm();
      dt_=QMBase::step(n1,n2);
   }
   else dt_=h;

   R tau=0.5*dt_;
   R ti=t_+tau;
   State xi=x_+v_*tau;
   State yi=op_.actOn(xi,ti);
   State dv=op_.actOn(yi,ti)*(fac*dt_);
   x_+=(v_*dt_+dv*tau);
   v_+=dv;
   t_+=dt_;
   stepCon_+=1;
   norIfEnf_();
   enr_=eTot();
}

bool TrjTD::prnOn(ostream& str)const
{
   cpmwt("TrjTD");
   if (! TrjB::prnOn(str)) return false;
//   cpmp(op_);
   cpmp(v_);
   return true;
}

bool TrjTD::scanFrom(istream& str)
{
   if (! TrjB::scanFrom(str)) return false;
//   cpms(op_);
   cpms(v_);
   return true;
}
*/
