//? cpmqmx20180827.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*
   cpmqmx.cpp
   History: see cpmqmx.h

*/

#include <cpmqmx.h>
#include <cpmvectormore.h>
#include <cpmobservable.h>
#include <cpmdistribution.h>
#include <cpmframes.h>

using CpmQM::RunConx;
using CpmQM::DisConx;

using CpmQM::SpinNx;
using CpmQM::Iax;

using CpmArrays::V;
using CpmArrays::VV;
using CpmArrays::VVV;
using CpmArrays::VVVV;
using CpmArrays::R_Vector;
using CpmArrays::DisR;
using CpmLinAlg::CC2;
using CpmLinAlg::CCCC2;
using CpmImaging::Color;
using CpmImaging::WHITE;
using CpmImaging::BLACK;
using CpmImaging::LIGHTGREEN;
using CpmImaging::LIGHTCYAN;
using CpmImaging::LIGHTBLUE;
using CpmImaging::FAINTYELLOW;

/////////////////////////// class  /////////////////////////////////

V<C2> SpinNx::dot(V<C2> const& vx, Iax const& ia)
{
   Word loc("SpinNx::dot(V<C2>,R,R,Iax)");
   Z mL=2;
   CPM_MA
   Outside os=CYCLIC;
   Z np=vx.dim();
   V<C2> r(np);
   for (Z i=1;i<=np;++i){
      C2 ci=vx.cui(i);
      C2 ri=ia(ci) + ia(vx(i-1,os),ci)+ia(vx(i+1,os),ci);
      r.cui(i)=ri*C::Ibar;
   }
   CPM_MZ
   return r;
}

C SpinNx::hInt()const
{
   C res;
   Outside os=CYCLIC;
   for (Z i=1;i<=np_;++i){
      res+= (ia_.avr(vx_(i-1,os),vx_.cui(i))
             + ia_.avr(vx_(i+1,os),vx_.cui(i)));
   }
   return res*0.5;
}

R SpinNx::hInt_i(Z i)const
{
   Outside os=CYCLIC;
   C2 ci=vx_[i];
   C z=ia_.avr(vx_(i-1,os),ci) + ia_.avr(vx_(i+1,os),ci);
   return z[1]*0.5;
}

R SpinNx::eSep_i(Z i)const
{
   C2 ci=vx_[i];
   return (ci|ia_(ci))[1];
}

R SpinNx::eTot_i(Z i)const
{
   return eSep_i(i)+hInt_i(i);
}

C SpinNx::eSep()const
{
   C res;
   Z np=vx_.dim();
   for (Z i=1;i<=np;++i){
      C2 ci=vx_.cui(i);
      C mi=(ci|ia_(ci));
      res+=mi;
   }
   return res;
}

R3 SpinNx::sig(Z i)const
{
   C2 ci=vx_.cui(i);
   C v1=(ci|CC2::sig1*ci);
   C v2=(ci|CC2::sig2*ci);
   C v3=(ci|CC2::sig3*ci);
   return R3(v1[1],v2[1],v3[1]);
}

namespace{ // anonymous
   // color related stuff
   const R xg=-0.5;
   const R yg=0.5*sqrt(3.);
   const R2 er(1,0);
   const R2 eg(xg,yg);
   const R2 eb(xg,-yg);
      // er+eg+eb=0,
      // er.absVal()==eg.absVal()==eb.absVal()==1
   C transform(R3 const& r){
      R xr=(r[1]+1);
      R xg=(r[2]+1);
      R xb=(r[3]+1);
      R2 res=er*xr+eg*xg+eb*xb;
      return C(res[1],res[2]);
   }
}

V<C2> SpinNx::col()const
{
   Z np=vx_.dim();
   if (np==0) return V<C2>(0);
   V<C2> res(np);
   for (Z i=1;i<=np;++i){
      R3 r3=sig(i);
      C z=transform(r3);
      res[i]=C2(z,z);
   }
   return res;
}

void SpinNx::stepLean_()
// step of the DALF and the ADALF method
{
   Z i;
   R tau=rc_.dt*0.25; // stepGam_() will be applied twice
   R tau2=tau*2.; // stepGam_() will be applied twice

   V<C2> vv1; // needed only if rc_.adalf

   t_+=tau;

   for (i=1;i<=np_;++i) vx_.cui(i)+=vv_.cui(i)*tau;

   V<C2> vv=dot(vx_,ia_);
   for (i=1;i<=np_;++i) vv_.cui(i)+=(vv.cui(i)-vv_.cui(i))*2.;

   if (rc_.adalf) vv1=vv;

   for (i=1;i<=np_;++i) vx_.cui(i)+=vv_.cui(i)*tau2;
   t_+=tau2;

   vv=dot(vx_,ia_);
   for (i=1;i<=np_;++i) vv_.cui(i)+=(vv.cui(i)-vv_.cui(i))*2.;

   for (i=1;i<=np_;++i) vx_.cui(i)+=vv_.cui(i)*tau;

   if (rc_.enfNor){
      for (i=1;i<=np_;++i) vx_.cui(i).nor_();
    //  R dev=cpmabs(n()-1.);
    //  if (dev>1e-6) cout<<"dev="<<dev<<endl;
   }

   if (rc_.adalf){
      for (i=1;i<=np_;++i) vv_.cui(i)=(vv_.cui(i)+vv1.cui(i))*0.5;
   }

   t_+=tau; // now t_ got increased by 4*tau = rc_.dt

   /*R_Vector rad(np_);
   Graph g;
   g.setWrtOnShow(true);*/

   //for (i=1;i<=np_;++i) rad.cui(i)=vx_.cui(i).nor_()-1.;
 //  for (i=1;i<=np_;++i) vx_.cui(i).nor_();

   /*if (rad.abs()/np_>0.1){
      g.setGridText(cpm(sc_));
      g.show(rad);
      cpmwait(6.);
   }*/
}

/************************
void SpinNx::step_()
{
   Word loc("SpinNx::step_()");
   Z mL=3;
   CPM_MA
   stepGam_();
   stepGam_();
   if (rc_.eps>0.){
      R ns = n();
      R diff = cpmabs(ns-1.);
      if (diff > rc_.eps){
         R nInv = cpminv(R(1.)*ns);
       //  nInv=cpmpow(nInv,1./np_);
         nInv=cpmpow(nInv,1.*np_);
         for (Z i=1;i<=np_;++i){
            vx_.cui(i)*=nInv;
         }
         vv_=dot(vx_,ia_);
      }
      R ns1=n();
      cpmassert(ns1<1e-9,"SpinNx::step_()");
   }
   CPM_MZ
}
***************************************/
/*************************
void SpinN::step_()
{
//   cout<<"SpinN::step started"<<endl;
   const R facRed=0.75;
   const R facInc=1.1;
   const R epsLow=0.1;
   V<C2> xMem=vx_;
   R tau=dt_*0.5;
   stepGam_(tau);
   stepGam_(tau);
   if (eps_>0){ // should not be active in the normal case ???
      // seems to be needed nevertheless since otherwise even harmless
      // configurations explode after having evolved rather regularly over
      // a long time. Together with time step adjustment this is much more
      // acceptable than before.
      // Conserved reversibility for eps_=1e-6 for the helicopter image
      // and seemed to fail reversibility for eps_=1e-4. In this case
      // the initial image was nevertheless recovered, but already in the
      // first third of the run.
      Z i;
      R sup=0;
      for (i=vx_.b();i<=vx_.e();++i){
         R ni=vx_[i].abs();
         R si=cpmabs(ni-1);
         if (si>sup) sup=si;
      }
      if (sup>eps_){ // do the step again with smaller dt
         t_-=dt_;
         dt_*=facRed;
         vx_=xMem;
         for (i=vx_.b();i<=vx_.e();++i) vx_[i].nor_();
         dot_();
         tau=dt_*0.5;
         stepGam_(tau);
         stepGam_(tau);
      }
      else if (sup<epsLow*eps_){ dt_*=facInc; }
      else ;
   } // eps_>0
   sc_++;
//   cout<<"SpinN::step done"<<endl;

}
*************************************/

void SpinNx::step_()
{
   Word loc("SpinNx::step_()");
   Z mL=3;
   CPM_MA
   if (rc_.eps<=0.){
         stepLean_();
   }
   else{
      const R facRed=0.75;
      const R facInc=1.1;
      const R epsLow=0.5;
      V<C2> xMem=vx_;
      Z iter=0;
label:
      stepLean_();
      // now we test whether the step was sound or whether we have to
      // redo it.
      Z i;
      R sup=0.;
      for (i=vx_.b();i<=vx_.e();++i){ // this is vx_ after stepGam2_
         R ni=vx_.cui(i).abs();
         R si=cpmabs(ni-1.);
         if (si>sup) sup=si;
      }
      cout<<"sup="<<sup<<endl;
      if (sup>rc_.eps){ // do the step again with smaller dt
     //    cout<<"iter="<<iter<<" sup="<<sup<<endl;
         iter++;
         t_-=rc_.dt;
         rc_.dt*=facRed;
         vx_=xMem;
         for (i=vx_.b();i<=vx_.e();++i) vx_.cui(i).nor_();
         goto label;
      }
      else if (sup<epsLow*rc_.eps){ rc_.dt*=facInc; }
      else ;
  //    cout<<"dt="<<rc_.dt<<endl;
   }
   sc_++;
   CPM_MZ
}

namespace{

void lineMark_1(Graph& g, V<C2> const& vx, DisConx const& dc)
{
   using CpmImaging::Color;
   static Z jShift=1;
   Z gw=g.w();
   Z gh=g.h();
   Z np=vx.dim();
   Z pix=gh/np;
   Z mult_i=pix;
   Z mult_j=pix;
   Z js=pix,is=pix;
   Z i,j;
   Z d_=np*mult_i;
   jShift+=js;

   Z iState=1;
   Z iCounter=0;
   Z iEff=is;
   Color c;

   for (i=1;i<=d_;i++){
      if (iCounter==0){
         C val;
         if (dc.colMet==1){ // Bloch angles based coloring
            val=vx.cui(iState).blc();
         }
         else if (dc.colMet==2){
            val=vx.cui(iState)[1];
         }
         else if (dc.colMet==3){
            val=vx.cui(iState)[2];
         }
         else if (dc.colMet==4){
            val=vx.cui(iState)[1]*vx.cui(iState)[2];
         }
         else val=C(1.);
         c=Color(val);
      }
      iCounter++;
      if (iCounter==mult_i){
         iCounter=0; // triggers new evaluation of c in the next loop
         iState++;
      }
      for (j=0;j<mult_j;j++){
         g.put(Z2(iEff,(j+jShift)%gw),c);
      }
         iEff++;
   }
}

} // namespace

void SpinNx::mark(Graph& g, DisConx const& dc)const
{
   lineMark_1(g,vx_,dc);
//   lineMark(g,dc);
}

void SpinNx::run_(Graph& g, DisConx const& dc)
{
   using CpmArrays::R_Matrix;
   using CpmArrays::R_Vector;
   using std::vector;
   Word loc("SpinN::run_(Graph,DisConx)");
   Z mL=1;
   CPM_MA
   g.setSclMem_(false);
   R n0=n();
   R n0Inv=cpminv(n0);
   C e0=eTot();
   R sig30=sig3();
   R2 blc0=blcAng();
   R theta0=blc0[1];
   R phi0=blc0[2];
   R eTotR=e0[1];
   R eInt=hInt()[1];
   cpmvalues2("eTot,eInt",eTotR,eInt,4);
   R e0Inv=cpminv(e0.abs());
   R e01=e0[1];
   R e02=e0[2];
   //cpmvalue("dtRel",dt_*dt0Inv_,3);

   Z nSteps=rc_.nSteps;
   R tTot=R(nSteps)*rc_.dt;
   R tTotInv=R(1)/tTot;
   R dt0Inv=cpminv(rc_.dt);
   R tWaitView=rc_.tWaitView;
   //R tWaitDocu=rc_.tWaitDocu;
   //R tWaitBoltzmann=rc_.tWaitDocu2;
   //R tWaitDocu3=rc_.tWaitDocu3;
   R tWaitDocuNormError=rc_.tWaitDocuNormError;
   R tWaitDocuEnergyError=rc_.tWaitDocuEnergyError;
   R tWaitRes=rc_.tWaitRes;
   CpmAlgorithms::PeriodicAlarm view(R(rc_.viewPeriod)*rc_.dt);
   CpmAlgorithms::PeriodicAlarm docu(R(rc_.docuPeriod)*rc_.dt);
   if (rc_.tWaitIni>0){
      mark(g,dc);
      g.setText("initial state");
      g.vis();
      cpmwait(rc_.tWaitIni,2);
   }
   g.clr_();
   vector<R> tv(1,0.), nv(1,0.), e1v(1,0.), e2v(1,0.),
      thetav(1,theta0), phiv(1,phi0), sig3v(1,0);
   Z np=vx_.dim();
   bool dtVar=getEps()>0;
   R ti=0;
   while (ti<tTot){
      step_();
      ti=t();
      if (dtVar){
         cpmvalue("dtRel",rc_.dt*dt0Inv,3);
      }
      cpmprogress("run",tTotInv*ti);
      if (docu(ti)){
         R ni=n();
         R sig3i=sig3();
         C ei=eTot();
         R2 blc=blcAng();
         R thetai=blc[1];
         R phii=blc[2];
         tv.push_back(ti);
         nv.push_back((ni-n0)*n0Inv);
         e1v.push_back((ei[1]-e01)*e0Inv);
         e2v.push_back((ei[2]-e02)*e0Inv);
         thetav.push_back(thetai);
         phiv.push_back(phii);
         sig3v.push_back(sig3i-sig30);
      }
      if (view(ti)||ti>=tTot){
         mark(g,dc);
         if (ti>=tTot) g.vis(true);
         else g.vis(dc.wrtOnShow);
         cpmwait(rc_.tWaitView);
      }
   }
   //R fac=10;
   //R facInv=R(1)/fac;
   R_Vector tR(tv);
   R_Vector nR(nv);
   R_Vector e1R(e1v);
   R_Vector e2R(e2v);
   R_Vector thetaR(thetav);
   R_Vector phiR(phiv);
   R_Vector sig3R(sig3v);
   if (tWaitDocuNormError>0){
      R_Matrix doc1(2);
      doc1[1]=tR;
      doc1[2]=nR;
      Graph g1; // it is neither important nor always useful
      // if diagnostic graphics is of the same size as
      // the evolving image. Hence we use a new frame here.
      g1.clr_();
    //  g1.setColorActive(LIGHTBLUE);
      g1.setText("SpinNx: evolution of norm error");
      g1.show(doc1);
      cpmwait(tWaitDocuNormError,2);
   }
   if (tWaitDocuEnergyError>0){
      R_Matrix doc2(3);
      doc2[1]=tR;
      doc2[2]=e1R;
      doc2[3]=e2R;
      Graph g2;
      g2.clr_();
      g2.setColorActive(FAINTYELLOW);
      g2.setText("SpinNx: evolution of energy errors");
      g2.show(doc2);
      cpmwait(tWaitDocuEnergyError,2);
   }
   if (tWaitRes>0){
      R_Matrix doc3(4);
      doc3[1]=tR;
      doc3[2]=thetaR;
      doc3[3]=phiR;
      doc3[4]=sig3R;
      Graph g3;
      g3.clr_();
    //  g3.setColorActive(LIGHTGREEN);
      g3.setText("SpinNx: evolution of mean Bloch angles and sig3");
      g3.show(doc3);
      cpmwait(tWaitRes,2);
   }
   CPM_MZ
}

/****************************************
void SpinNx::run_(Graph& g, DisConx const& dc)
{
   using CpmArrays::R_Matrix;
   using CpmArrays::R_Vector;
   using std::vector;
   Word loc("SpinNx::run_(Graph, DisConx)");
   Z mL=1;
   CPM_MA
   g.setSclMem_(false);
   R n0=n();
   R n0Inv=cpminv(n0);
   C e0=eTot();
   R sig30=sig3();
   R2 blc0=blcAng();
   R theta0=blc0[1];
   R phi0=blc0[2];
   R eTotR=e0[1];
   R eInt=hInt()[1];
   cpmvalue("dt",rc_.dt,3);
   cpmvalues2("eTot,eInt",eTotR,eInt,4);
   R e0Inv=cpminv(e0.abs());
   R e01=e0[1];
   R e02=e0[2];
   Z nSteps=rc_.nSteps;
   R nInv=1./nSteps;
   R tWaitView=rc_.tWaitView;
   R tWaitDocu=rc_.tWaitDocu;
   R tWaitBoltzmann=rc_.tWaitDocu2;
   R tWaitDocu3=rc_.tWaitDocu3;
   R tWaitRes=rc_.tWaitRes;
   CpmAlgorithms::CyclicAlarm view(rc_.viewPeriod);
   CpmAlgorithms::CyclicAlarm docu(rc_.docuPeriod);
   if (rc_.tWaitIni>0){
      mark(g,dc);
      g.setText("initial state");
      g.vis();
      cpmwait(rc_.tWaitIni,2);
   }
   g.clr_();
   vector<R> tv(1,0.), nv(1,0.), e1v(1,0.), e2v(1,0.),
      thetav(1,theta0), phiv(1,phi0), sig3v(1,0),
      temv, misfitv, t2v, magv, energies;
   Z np=vx_.dim();
   for (Z i=1;i<=nSteps;++i){
      step_();
      cpmprogress("run",nInv*i);
      if (docu()){
         cpmvalue("dt",rc_.dt,3);
         R ti=t();
         R ni=n();
         R sig3i=sig3();
         C ei=eTot();
         R2 blc=blcAng();
         R thetai=blc[1];
         R phii=blc[2];
         tv.push_back(ti);
         nv.push_back((ni-n0)*n0Inv);
         e1v.push_back((ei[1]-e01)*e0Inv);
         e2v.push_back((ei[2]-e02)*e0Inv);
         thetav.push_back(thetai);
         phiv.push_back(phii);
         sig3v.push_back(sig3i-sig30);
         if (tWaitBoltzmann>=0){ // for ==0 no display
            V<R> enrDis(np);
            for (Z k=1;k<=np;++k){
               R ek=eTot_i(k);
               enrDis[k]=ek;
               energies.push_back(ek);
            }
            DisR disr(enrDis);
            R2 r2=disr.tBoltzmann(tWaitBoltzmann,"histogramm of magnetic\
             energies");
            R misFit=r2[2];
            const R misFitLimit=0.3;
            if (misFit<misFitLimit){
               t2v.push_back(ti);
               temv.push_back(r2[1]);
               misfitv.push_back(misFit);
               R mag=sig3();
               magv.push_back(mag);
            }
         }
      }
      if (view()||i==nSteps){
         mark(g,dc);
         if (i==nSteps){
            g.vis(true);
            cpmwait(rc_.tWaitView,2);
         }
         else g.vis(dc.wrtOnShow);
      }
   }
   if (tWaitDocu>0){
      R_Matrix doc(4);
      doc[1]=R_Vector(tv);
      doc[2]=R_Vector(nv);
      doc[3]=R_Vector(e1v);
      doc[4]=R_Vector(e2v);
      g.clr_();
      g.setAutoScl_(true);
      g.setWriteOnShow(true);
      g.setText("SpinNx: evolution of norm and energy errors");
      g.show(doc);
      cpmwait(tWaitDocu,2);
   }
   if (tWaitRes>0){
      R_Matrix doc(4);
      doc[1]=R_Vector(tv);
      doc[2]=R_Vector(thetav);
      doc[3]=R_Vector(phiv);
      doc[4]=R_Vector(sig3v);
      g.clr_();
      g.setAutoScl_(true);
      g.setWriteOnShow(true);
      g.setText("SpinNx: evolution of mean Bloch angles and sig3");
      g.show(doc);
      cpmwait(tWaitRes,2);
   }
   if (tWaitDocu3>0 && tWaitBoltzmann>=0){
       g.setText("SpinNx: distribution of particle energies");
       DisR disEnergies(energies,64);
       disEnergies.show(g,true);

      R_Matrix doc(4);
      doc[1]=R_Vector(t2v);
      doc[2]=R_Vector(temv);
      doc[3]=R_Vector(misfitv);
      doc[4]=R_Vector(magv);
      g.clr_();
      g.setAutoScl_(true);
      g.setText("SpinNx: evolution temperature, misfit, and magnetization"\
      )\
      ;
      g.show(doc);

      cpmwait(tWaitDocu3,2);
      CPM_MZ
   }
}

*****************************************/

void SpinNx::adj_dt_()
{
   const R eps1 = R(1.) + 1.e-5;
   const R fac = cpminv(2.);
   Z nTest=rc_.nIter;
   cpmvalue("nTest",nTest,3);
   SpinNx stt0=*this;
   R dt = rc_.dt;
   cpmvalue("dt",dt);
   cpmwait(2);
label:
   dt*=fac;
   cpmvalue("dt",dt);
   SpinNx temp=stt0; // start always from the same initial state
   temp.set_dt_(dt);
   for (Z i=1;i<=nTest;++i){
      temp.step_();
      if (temp.n() > eps1) goto label;
   }
   dt*=fac; // last safety
   set_dt_(dt);
   cpmvalue("dtFinal",rc_.dt,2);
}

