//? cpmset.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmset.h>
#include <cpmframe.h>
#include <cpmmultiimage.h>
#include <cpmalgorithms.h>
#include <algorithm>

using namespace CpmArrays;
using namespace CpmRoot;
using namespace CpmRootX;
using namespace CpmSystem;
using namespace CpmGraphics;
using namespace CpmImaging;

using namespace std;

char Set::a='{';
char Set::z='}';

void Set::unwrap(string& str)
{
   Z ns=(Z)str.length();
   if (ns<2) cpmerror("string too short for unwrapping");
   else if (ns==2) str="";
   else{
      str.erase(0,1);
      str.erase(str.length()-2,1);
   }
}

void Set::ini()
{
   Z mL=3;
   Word loc("Set::ini()");
   cpmmessage(mL,loc&" started");
   elm=vector<string>(0);
   n_=0;
   if (val.length()==2){
     // isValid=true;
      cpmmessage(mL,loc&" done, void set created");
      return;
   }
   else{
      string str=val;
      unwrap(str);
         // taking appart the enclosing braces
      Z sum=0;
      while (true){
         if (str.length()==0) break;
         string ele("");
         while (true){
            char v=str[0];
            str.erase(0,1);
            if (v==a){
               sum++;
               ele+=a;
            }
            else if (v==z){
               sum--;
               ele+=z;
               if (sum==0) break;
            }
            else{
               cpmerror(loc&": unexpected character found");
            }
         }
         elm.push_back(ele);
         n_++;
         cpmmessage(mL+1,"reading element "&cpmwrite(n_)&" done");
      }
   }
   cpmmessage(mL,loc&" done");
}

void Set::iniFromElm()
{
   Z mL=3;
   Word loc("Set::iniFromElm()");
   cpmmessage(mL,loc&" started");
   n_=(Z)elm.size();
   val=string(1,a);
   for (Z i=0;i<n_;i++) val+=elm[i];
   val+=string(1,z);
   cpmmessage(mL,loc&" done");
}

Set::Set()
{
   val=string(1,a)+string(1,z);
   ini();
}

Set::Set(const V<Set>& vs):elm(vs.dim())
{
   Z mL=3;
   Word loc("Set(const V<Set>&)");
   cpmmessage(mL,loc&" started");
   n_=(Z)elm.size();
   if (n_==0){
      *this=Set();
      cpmmessage(mL,loc&" done, void set made");

   }
   else if (n_==1){
      cpmmessage(mL,loc&" done, one-element set made");
      *this=vs[1].push();
   }
   else{
      for (Z i=0;i<n_;i++){
         string si=vs[i+1].rep(); // nothing to unwrap
         // the representations of the vs[i] have directly be
         // concatenated
         // to form elm of the set to generate
         elm[i]=si;
      }
      std::stable_sort(elm.begin(),elm.end());
      vector<string>::iterator newend;
      newend=std::unique(elm.begin(),elm.end());
         // uniqe works only as needed on ordered arrays
      elm.erase(newend,elm.end()); // not done automatically by unique!
      iniFromElm();
   }
   cpmmessage(mL,loc&" done");
}

Set::Set(Z nI, bool start1)// note that {{{}}}<{{}} in the order of sets
// whereas 3 > 2 for the corresponding numbers.
{
   Set res;
   if( nI>0){
      if (start1){
         for (Z i=nI;i>=1;--i){
            Set ni=Set::makeNumber(i);
            res=res.add(ni);
         }
      }
      else{
         for (Z i=nI-1;i>=0;--i){
            Set ni=Set::makeNumber(i);
            res=res.add(ni);
         }
      }
   }
   *this=res;
 }

Set Set::push()const
// add enclosing braces
{
   string valRes=string(1,a)+val+string(1,z);
   return Set(valRes);
}

V<Set> Set::pop()const
// remove enclosing braces
    // s.push().pop() leaves s unchanged. If applied to a set that has
    // not just one element, *this will be left unchanged with respect
    // to val and elm.
{
   if (n_!=1){
      cpmwarning("Set::pop() created unvalid object");
      return V<Set>(0);
   }
   else{ // thus n_==1
      string valRes=elm[0];
      return V<Set>(1,Set(valRes));
   }
}

// order operators dirctly from the string representation

bool Set::operator == ( const Set & x)const
{ return val==x.val;}

bool Set::operator != (const Set & x)const
{ return val!=x.val;}

bool Set::operator < (const Set & x)const
{ return val<x.val;}

bool Set::operator > (const Set & x)const
{ return val>x.val;}

bool Set::operator <= (const Set & x)const
{ return val<=x.val;}

bool Set::operator >= (const Set & x)const
{ return val>=x.val;}

// lets use Word::prnOn and Word::scanFrom since than
// my mechanism to support the message passing interface is
// active
bool Set::prnOn(ostream& str)const
{
   CpmRoot::writeTitle("Set",str);
   if (!Word(val).prnOn(str)) return false;
   return true;
}

bool Set::scanFrom(istream& str)
{
   Word w;
   if (!w.scanFrom(str)) return false;
   val=w.str();
   ini();
   return true;
}

V<Set> Set::operator()(void)const
{
   V<Set> res(n_);
   for (Z i=1;i<=n_;i++){
      res[i]=Set(elm[i-1]);
   }
   return res;
}

Set Set::operator |(const Set& s)const
      // forming the union of *this and s (corresponds to the
      // 'or' operation for indicator functions)
{
   Z ns=s.card();
   vector<string> elms=s.elm;
   vector<string> res(n_+ns);

   vector<string>::iterator limit;
   limit=std::set_union
      (elm.begin(),elm.end(),elms.begin(),elms.end(),res.begin());
   res.erase(limit,res.end());
   return Set(res);
}

Set Set::operator &(const Set& s)const
      // forming the section of *this and s (corresponds to the
      // 'and' operation for indicator functions)
{
   Z ns=s.card();
   vector<string> elms=s.elm;
   Z nRes=(n_<=ns ? n_ : ns);
   vector<string> res(nRes);

   vector<string>::iterator limit;
   limit=std::set_intersection
      (elm.begin(),elm.end(),elms.begin(),elms.end(),res.begin());
   res.erase(limit,res.end());
   return Set(res);
}

Set Set::operator -(const Set& s)const
{
   Z ns=s.card();
   vector<string> elms=s.elm;
   Z nRes=n_;
   vector<string> res(nRes);
   vector<string>::iterator limit;
   limit=std::set_difference
      (elm.begin(),elm.end(),elms.begin(),elms.end(),res.begin());
   res.erase(limit,res.end());
   return Set(res);
}

Set Set::operator ^(const Set& s)const
{
   Z ns=s.card();
   vector<string> elms=s.elm;
   Z nRes=n_+ns;
   vector<string> res(nRes);
   vector<string>::iterator limit;
   limit=std::set_symmetric_difference
      (elm.begin(),elm.end(),elms.begin(),elms.end(),res.begin());
   res.erase(limit,res.end());
   return Set(res);
}

Set Set::operator[](Z i)const
{
   if (i<1) cpmerror("Set::operator[](Z i): i="&cpmwrite(i));
   if (i>n_) cpmerror
      ("Set::operator[](Z i): i="&cpmwrite(i)&", n="&cpmwrite(n_));
   return Set(elm[i-1]);
}

Set Set::operator()(Z i)const
{
   if (i<1) i=1;
   if (i>n_) i=n_;
   return Set(elm[i-1]);
}

bool Set::hasElem(const Set& s)const
{
   string ss=s.rep();
   for (Z i=0;i<n_;i++){
      if (ss==elm[i]) return true;
   }
   return false;
}

Z Set::locate(const Set& s)const
      // returns 0 if s is not an element of *this
      // and its index if it is an element of *this
{
   string ss=s.rep();
   for (Z i=0;i<n_;i++){
      if (ss==elm[i]) return i+1;
   }
   return 0;
}

bool Set::isSubSetOf(const Set& s)const
      // returns true iff *this is a subset of s
{
   return n_==((*this)&s).card();
}

bool Set::isSuperSetOf(const Set& s)const
      // returns true iff s is a subset of *this
{
   return n_==(s&(*this)).card();
}

//Set Set::orderedPair(const Set& y)const
Set Set::wkp(const Set& y)const
{
   Set c1=push();
   V<Set> vc2{*this,y};
   Set c2(vc2);
   V<Set> vRes{c1,c2};
   return Set(vRes);
}

V<Set> Set::getTwoFactors()const
{
   Z mL=3;
   Word loc("Set::getTwoFactors()");
   cpmmessage(mL,loc&" started");
   if (n_==1){
      V<Set> x=pop();
      if (x.dim()==1){
         Set x1=x[1];
         V<Set> y=x1.pop();
         if (y.dim()==1){
            Set y1=y[1];
            return V<Set>{y1,y1};
         }
         else return V<Set>(0);
      }
      else return V<Set>(0);
   }
   else if (n_==2){
      Set c1=first();
      Set c2=last();
      Z n1=c1.card();
      Z n2=c2.card();
      Z n3=n1+n2;
      if (n3!=3){
         cpmwarning(loc&" no valid result due to n3!=3");
         return V<Set>(0);
      }
      else{
         Set x=c1&c2;
         Set y;
         if (n1==1) y=c2-x; else y=c1-x;
         return V<Set>{x.pop()[1],y.pop()[1]};
      }
   }
   else{
      return V<Set>(0);
   }
   cpmmessage(mL,loc&" done");
}

Set Set::operator *(const Set& y)const
{
   Z mL=2;
   Word loc("Set::operator *(const Set& y)");
   cpmmessage(mL,loc&" started");
   Z ny=y.card();
   Z nRes=n_*ny;
   Z i,j;
   V<Set> vs=operator()();
   V<Set> vsy=y();
   V<Set> vRes(nRes);
   Z k=0;
   for (i=1;i<=n_;i++){
      for (j=1;j<=ny;j++){
         k++;
         vRes[k]=vs[i].wkp(vsy[j]);
      }
   }
   cpmmessage(mL,loc&
      " done, except of statement return Set(vRes), vRes.dim()="
      &cpmwrite(nRes));
   return Set(vRes);
}

Set Set::tupel(const V<Set>& vs)
{
   Z m=vs.dim();
   if (m==0) return Set();
   else if (m==1) return vs[1];
   else{
      Set res(vs[1]);
      for (Z i=2;i<=m;i++){
         res=res.wkp(vs[i]);
      }
      return res;
   }
}

V<Set> Set::getFactors()const
{
   Z mL=1;
   Word loc("Set::getFactors()");
   cpmmessage(mL,loc&" started");
   vector<Set> temp; // for fast appending at the end
   Set toFactorize=*this;
   bool go=true;
   Z nFactorsGot=0;
   while(go){
      V<Set> factors=toFactorize.getTwoFactors();
      if (factors.dim() !=2){
         go=false;
      }
      else{
         temp.push_back(factors[2]);
         toFactorize=factors[1];
         nFactorsGot++;
      }
   }
   if (nFactorsGot==0){
      cpmmessage(mL,loc&" done, trivial array returned");
      return V<Set>{toFactorize}; // if no split into two factor was
      // found,
         // the term is one factor himself
   }
   else{
      temp.push_back(toFactorize);
      Z i,m=(Z)temp.size();
      V<Set> res(m);
      for (i=1;i<=m;i++) res[i]=temp[m-i];
      cpmmessage(mL,loc&
         " done, array of dimension "&cpmwrite(m)&" returned");
      return res;
   }
}

Set Set::uni(const V<Set>& vs)
{
   Z i,m=vs.dim();
   if (m==0) return Set();
   else if (m==1) return vs[1];
   else{
      Set res=vs[1];
      for (i=1;i<=m;i++) res=res|vs[i];
      return res;
   }
}

Set Set::sec(const V<Set>& vs)
{
   Z i,m=vs.dim();
   if (m==0) return Set();
   else if (m==1) return vs[1];
   else{
      Set res=vs[1];
      for (i=1;i<=m;i++) res=res&vs[i];
      return res;
   }
}

Set Set::car(const V<Set>& vs)
{
   Z i,m=vs.dim();
   if (m==0) return Set();
   else if (m==1) return vs[1];
   else{
      Set res=vs[1];
      for (i=1;i<=m;i++) res=res*vs[i];
      return res;
   }
}

Set Set::subSet(const V<bool>& vb)const
{
   if (n_==0) return Set();
   else{
      V<Set> vs=operator()();
         // vs.dim()==n per def
      Z i,m=vb.dim();
      m=(m<=n_ ? m : n_); // inf(m,n)
      vector<Set> temp;
      for (i=1;i<=m;i++){
         if (vb[i]==true) temp.push_back(vs[i]);
            // vb[i] and vs[i] defined by definition of m
      }
      m=(Z)temp.size();
      if (m==0){
         return Set();
      }
      else{
         V<Set> vres(m);
         for (i=1;i<=m;i++) vres[i]=temp[i-1];
         return Set(vres);
      }
   }
}

Set Set::powerSet()const
      // set of all subsets of *this
{
   Z mL=2;
   Word loc("Set::powerSet()");
   cpmmessage(mL,loc&" started: n="&cpmwrite(n_));
   Z nBytes=sizeof(mL);
   Z nMax=8*nBytes;
   cpmassert(n_<=nMax,loc);
   Z m=Z1(2).pow(n_);
   V<Set> vres(m);
   for (Z i=0;i<m;i++){
      V<bool> vb=CpmAlgorithms::bitsFromZ(i,n_);
      vres[i+1]=subSet(vb);
   }
   cpmmessage(mL,
      "last possible message from "&loc&": collecting subsets done,"
      &" output may take time");
   return Set(vres);
}

bool Set::isRelation()const
{
   V<Set> vs=operator()();
   Z i,d;
   for (i=1;i<=n_;i++){
      d=vs[i].getTwoFactors().dim();
      if (d!=2) return false;
   }
   return true;
}

V<Set> Set::decompose()const
{
   Z maxFac=10;
   V<Set> res(maxFac);
   Z i,d;
   V<Set> vs=operator()();
   for (i=1;i<=n_;i++){
      d=vs[i].getFactors().dim();
      res[d]=res[d].add(vs[i]);
   }
   Z iNeeded=0;
   for (i=1;i<=maxFac;i++){
      Z di=res[i].card();
      if (di>0) iNeeded=i;
   }
   return res.resize(iNeeded);
}

V<Set>  Set::extractRelation()const
{
   V<Set> res(3);
   Z i,d;
   V<Set> vs=operator()();
   for (i=1;i<=n_;i++){
      V<Set> vsigf=vs[i].getTwoFactors();
      d=vsigf.dim();
      if (d==2){
         res[1]=res[1].add(vs[i]);
         res[2]=res[2].add(vsigf[1]);
         res[3]=res[3].add(vsigf[2]);
      }
   }
   if (res[1].isVoid()) return V<Set>(0);
   else return res;
}

bool Set::isFunction()const
      // returns true iff all elements of *this are wkp's
      // and if there apear no two different wkp's which agree
      // in their first component
{
   Word loc("Set::isFunction()");
   Z mL=1;
   CPM_MA
   V<Set> vs=operator()();
   for (Z i=1;i<=n_;i++){
      Z d=vs[i].getTwoFactors().dim();
      if (d!=2){
         CPM_MZ
         return false;
      }
   }
   CPM_MZ
   return true;
}

V<Set> Set::getValues(Set const& x)const
{
   V<Set> vs=operator()();
   V<Set> y;
   Z i,d;
   for (i=1;i<=n_;i++){
      V<Set> f=vs[i].getTwoFactors();
      cpmassert(f.dim()==2,"Set::getValues(Set)");
      if (f[1]==x) y&=f[2];
   }
   return y;
}

Set Set::getFunctionValue(Set const& x)const
{
   Word loc("Set::getFunctionValue(Set)");
   Z mL=3;
   CPM_MA
   V<Set> vs=operator()();
   for (Z i=1;i<=n_;i++){
      V<Set> f=vs[i].getTwoFactors();
      cpmassert(f.dim()==2,"Set::getFunctionValues(Set)");
      if (f[1]==x){
         CPM_MZ
         return f[2];
      }
   }
   cpmwarning(loc,Word("no function value found for x = ")&x.toWord());
   CPM_MZ
   return Set();
}

bool Set::isGroup()const
{
   Word loc("Set::isGroup()");
   Z mL=1;
   CPM_MA
   V<Set> vs=getFactors();
   if (vs.dim()!=4) return false;
   Set g=vs[1];
   Set mult=vs[2];
   Set inv=vs[3];
   Set e=vs[4];
   cout<<"loc="<<-loc<<endl;
   cout<<"g="<<g<<endl;
   cout<<"mult="<<mult<<endl;
   cout<<"inv="<<inv<<endl;
   cout<<"e="<<e<<endl;
   cpmcerr<<"g="<<g<<endl;
   cpmcerr<<"mult="<<mult<<endl;
   cpmcerr<<"inv="<<inv<<endl;
   cpmcerr<<"e="<<e<<endl;
   if (!mult.isFunction()){
      cout<<"mult is not a function"<<endl;
      return false;
   }
   if (!inv.isFunction()){
      cout<<"inv is not a function"<<endl;
      return false;
   }
   if (!g.hasElem(e)){
      cout<<"e is not element of the the group"<<endl;
      return false;
   }
   Z n=g.dim();
   cout<<"n="<<n<<endl;
   for (Z i=1;i<=n;++i){ // testing whether e is neutral
      Set gi=g[i];
      Set egi=mult.getFunctionValue(e.wkp(gi));
      if (egi!=gi){
         cout<<" e is not a neutral element"<<endl;
         return false;
      }
   }
   cout<<"first loop done"<<endl;
   for (Z i=1;i<=n;++i){ // testing whether inv is an inverse
      Set gi=g[i];
      Set giInv=inv.getFunctionValue(gi);
      Set gigiInv=mult.getFunctionValue(gi.wkp(giInv));
      if (gigiInv!=e){
         cout<<"inv fails to be the inverse function for i="<<i<<endl;
         cpmwarning(" inverse of "&gi.toWord()&" erroneously given as "&
         giInv.toWord());
         return false;
      }
   }
   cout<<"second loop done"<<endl;
   for (Z i=1;i<=n;++i){ // testing whether mult is associative
      Set gi=g[i];
      for (Z j=1;j<=n;++j){
         Set gj=g[j];
         Set gigj=mult.getFunctionValue(gi.wkp(gj));
         for (Z k=1;k<=n;++k){
            Set gk=g[k];
            Set gjgk=mult.getFunctionValue(gj.wkp(gk));
            Set gigj_gk=mult.getFunctionValue(gigj.wkp(gk));
            Set gi_gjgk=mult.getFunctionValue(gi.wkp(gjgk));
            if (gigj_gk!=gi_gjgk){
               cout<<"mult fails to be associative"<<endl;
               return false;
            }
         }
      }
   }
   cout<<"third loop done"<<endl;
   CPM_MZ
   return true;
}

Set Set::makeNumber(Z n)
{
   Z m = (n<0 ? -n : n); //m =|n|
   Set res; // void set
   if (m==0) return res;
   Z p=m; // p>0
   while (p>0){ res=res.push(); p--;}
      // thus for n=1, push is called once, OK
   return res;
}

Z Set::number()const
{
   Word loc("Set::number()");
   Z res=0;
   if (n_==0) return res; // the void set corresponds
      // to Number 0
   Set mem=*this;
   while (mem.n_==1){ mem=mem.pop()[1]; res++;}
      // thus for n=1, push is called once, OK
   return res;
}

Set Set::makeNumericFunction(Z n, const F<Z,Z>& f)
{
   Set res;
   for (Z i=1;i<=n;++i){
      res.add(Set::makeNumber(i).wkp(Set::makeNumber(f(i))));
   }
   return res;
}

void modn(Z&i, Z n)
{
   while(i<0) i+=n;
   while(i>=n) i-=n;
}

Set Set::makeCyclicGroup(Z n)
{
   Word loc("Set::makeCyclicGroup(Z)");
   Z mL=1;
   CPM_MA
   Set g(n,false); //{n-1, n-2, ...,0} n elements
   cpmassert(g.dim()==n,loc);
   Set e=g[n];
   Set inv;
   for (Z i=1;i<=n;i++){
      Set gi=g[i];
      Z zi=-gi.number();
      modn(zi,n);
      Set gk=makeNumber(zi);
      cpmassert(g.hasElem(gk),loc);
      Z ci1=inv.card();
      inv=inv.add(gi.wkp(gk));
      Z ci2=inv.card();
      if (ci1==ci2){
            cout<<" multiplicity error in inv "<<endl;
      }
   }
   cout<<"inv="<<inv<<endl;
   cpmassert(inv.card()==n,loc);
   for (Z i=1;i<=n;++i){
      cout<<"inv["<<i<<"]= "<<inv[i]<<endl;
   }
   Set mult;
   for (Z i=1;i<=n;i++){
      Set gi=g[i];
      Z zi=gi.number();
      for (Z j=1;j<=n;j++){
         Set gj=g[j];
         Z zj=gj.number();
         Z zk=zi+zj;
         modn(zk,n);
         Set gk=makeNumber(zk);
         cpmassert(g.hasElem(gk),loc);
         Z cm1=mult.card();
         mult=mult.add(gi.wkp(gj).wkp(gk));
         Z cm2=mult.card();
         if (cm1==cm2){
            cout<<" multiplicity error in mult "<<endl;
         }
      }
   }
   cout<<"mult="<<mult<<endl;
   cpmassert(mult.card()==n*n,loc);
   V<Set> list{g,mult,inv,e};
   CPM_MZ
   return Set::tupel(list);
}

Z ft1(Z const& i){ return i*i;}
bool Set::test1(Z n)
{
   Z err=0;
   for (Z i=0;i<=n;++i){
      Set ni=makeNumber(i);
      Z zi=ni.number();
      if (zi!=i) err++;
   }
   if (err!=0) cout<<"err = "<<err<<endl;
   bool b0=(err==0);
   F<Z,Z> f(ft1);
   Set a=Set::makeNumericFunction(n,f);
   bool b1=a.isFunction();
   Set g=Set::makeCyclicGroup(n);
   bool b2=g.isGroup();
   return b0&&b1&&b2;
}

R_Vector Set::grpRep()const
   //: graphical representation
{
   Z n=val.size();
   R_Vector res(n);
   res[1]=(val[0]==a ? 1. : -1);
   for (Z i=2; i<=n; ++i){
      res[i]=res[i-1]+(val[i-1]==a ? 1. : -1);
   }
   return res;
}

Img24 Set::imgRep(R aspRat, R prob, Z mult)const
{
   Z n=val.size();
   R probh=prob*0.5;
   R nr(n);
   Z n2=cpmtoz(cpmsqrt(nr/aspRat)+0.5); // height
   Z n1=cpmtoz(n2*aspRat+0.5); // width
   while (n1*n2<n) n2++; // guaranty for sufficient number of pixels
   Z z0=0,z255=255;
   rgb c0(z0,z0,z255);
   Img24 im(n1,n2,c0);
   Z p=-1;
   rgb c1(z0,z255,z0); // green, suggests open brace
   rgb c2(z255,z0,z0); // red, suggests closed brace
   for (Z j=0;j<n2;++j){ // changes the line
      for (Z i=0;i<n1;++i){ // varies along a line
         p++;
         if (p<n){
            rgb ca=(val[p]==a ? c1 : c2);
            if (prob>0.){
               R r=randomR();
               if (r>1.-probh) ca=c1;
               if (r<=probh) ca=c2;
            }
            im.putPel(i,j,ca); // indexes follow the x,y scheme,
               // that is i varies allong the line
         }
         else goto Label;
      }
   }
Label:
   if (mult==1) return im;
   ImgR res(im);
   res=res.interpolate(mult*n2,mult*n1,0);
      // ImgR has matrix understanding of indexes
   return res.toImg24();
}

bool Set::test2(Z n)
{
   Set g=Set::makeCyclicGroup(n);
   Img24 im=g.imgRep(1.5,0.0,1);
   // these are the settings which for n=10 gave my
   // 'Graphics of the month' for February 2013
   im.vis();
   im.write("cycX"&cpm(n)&"AsSet");
   cpmwait(30,2);
   return true;
}


