//? cpmsttd1.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*
   cpmsttd1.cpp
   History: Started 2016-05-18

*/

#include <cpmfft.h>
#include <cpmgraph.h>
#include <cpmframes.h>
#include <cpmvr.h>
#include <cpms.h>
#include <cpmrvector.h>
#include <cpmsttd1.h>
#include <cpmmultiimage.h>
#include <cpmfile.h>

using CpmQuantumMechanics::QMBase;
using CpmQuantumMechanics::Bio;
using CpmQuantumMechanics::SttD1;
using CpmQuantumMechanics::ObsD1;
using CpmQuantumMechanics::SpcObjD1;
using CpmQuantumMechanics::Cs;
using CpmQuantumMechanics::Vc;
using CpmLinAlg::R2;
using CpmGraphics::Frame;
using CpmGraphics::Frames;
using CpmGraphics::Graph;
using CpmGraphics::Z2;
using CpmGeo::Iv;
using CpmImaging::Color;
using CpmImaging::MultiImage;
using CpmFunctions::F;
using CpmFunctions::F1;
using CpmFunctions::R_Func;
using CpmSystem::OFile;
using namespace std;
using namespace CpmRoot;
using namespace CpmRootX;
using namespace CpmArrays;

////////////////////////////// class SttD1 //////////////////////////////

//CC2 SttD1::unit_=CpmLinAlg::CC2::sig0;
//CC2 SttD1::beta_=CpmLinAlg::CC2::sig3;
//CC2 SttD1::alpha1_=CpmLinAlg::CC2::sig1;

CC2 SttD1::unit_=CpmLinAlg::CC2(C(1.),C(0.),C(0.),C(1.));
CC2 SttD1::beta_=CpmLinAlg::CC2(C(1.),C(0.),C(0.),C(-1.));
CC2 SttD1::alpha1_=CpmLinAlg::CC2(C(0.),C(1.),C(1.),C(0.));
   // Never initialize static data with static data as done in the
   // commented three lines! Whether this works or not depends on the
   // order of initialization and this is not fixed by the code.

SttD1::SttD1(V<C2> const& v, Iv const& s):
   val_(2*v.dim()),s_(s)
{
   Z d=v.dim(),i,j,k;
   for (i=1,j=0,k=d; i<=d; ++i,++j,++k){
      C2 cci=v[i];
      val_(j)=cci.x1.std();
      val_(k)=cci.x2.std();
   }
   adjust_();
}

SttD1::SttD1(Bio const& b ): s_(b.iv())
{
   val_=Vc::Constant(b.dd(),Cs(0.,0.));
   adjust_();
}

SttD1::SttD1(Z i, Z a, Bio const& b): b_(b)
{
   d_=b_.d();
   n_=b_.dd();
   s_=b_.iv();
   sup_=s_;
   val_=Vc::Constant(n_,Cs(0.,0.));
   while (i<1){ i+=d_;}
   while (i>d_){ i-=d_;}
   Z k = i-1;
   if (a==2) k+=d_;
   val_(k)= Cs(1.,0.);
}

SttD1::SttD1(C const& z, Bio const& b):
   s_(b.iv())
{
   val_=Vc::Constant(b.dd(),z.std());
   adjust_();
}

SttD1::SttD1(F<R,C2> const& f, Bio const& b)
{
   V<R> x=b.points();
   V<C2> y=x(f);
   *this=SttD1(y,b.iv());
}

SttD1::SttD1(Bio const& b, R const& xPeak, R const& sigma , R const& p,
   R const& a1, R const& a2, R const& phase1, R const& phase2,
   bool compact)
{
   V<R> x=b.points();
   V<R> y=x(R_Func::bell(xPeak,sigma,compact));
   R phi1=phase1*cpmdeg;
   R phi2=phase2*cpmdeg;
   C2 c2(C::expi(phi1)*a1,C::expi(phi2)*a2);
   Z d=b.d();
   V<C2> w(d);
   for (Z i=1;i<=d;i++) w[i] = c2*C::expi(x[i]*p)*y[i];
   SttD1 res(w,b.iv());
   R hw=sigma*4; // half width of support
   res.setSup_(Iv(xPeak-hw,xPeak+hw));
   *this=res;
}

V<SttD1> SttD1::ortBase()const
{
   Z mL=1;
   Word loc("SttD1::ortBase()");
   CPM_MA
   V<SttD1> res(n_);
   Vc wf=Vc::Constant(n_,Cs(0.));
   Z i,j;
   for (i=1,j=0; i<=n_; ++i,++j){
      Vc wfi=wf;
      wfi(j)=Cs(1.);
      res.cui(i)=SttD1(wfi,s_);
   }
   CPM_MZ
   return res;
}

SttD1 SttD1::operator-(void)const
{
   return SttD1(-val_,s_);
}

SttD1& SttD1::operator*=(C z)
{
   val_*=z.std();
   return *this;
}

SttD1& SttD1::operator+=(SttD1 const& psi)
{
   val_+= psi.val_;
   return *this;
}

SttD1& SttD1::operator-=(SttD1 const& psi)
{

   val_-= psi.val_;
   return *this;
}

SttD1 SttD1::operator*(C z)const
{
   return SttD1(val_*z.std(),s_);
}

SttD1 SttD1::operator +(SttD1 const& psi)const
{
   return SttD1(val_+psi.val_,s_);
}

SttD1 SttD1::operator -(SttD1 const& psi)const
{
   return SttD1(val_-psi.val_,s_);
}

SttD1& SttD1::operator*=(R r)
{
   val_*=Cs(r);
   return *this;
}

SttD1 SttD1::operator*(R r)const
{
   return SttD1(val_*Cs(r),s_);
}

SttD1 SttD1::con()const
{
   SttD1 res=SttD1(val_.conjugate(),s_);
   res.setSup_(sup_);
   return res;
}

SttD1 SttD1::p()const
{
   Z mL=1;
   Word loc("SttD1::p()");
   CPM_MA
   Z i,j;
   Vc res(n_);
   Cs fac(0.,-b_.gapInv()*0.5); // - added 2016-07-15
   for (i=1,j=1+d_;i<d_-1;++i,++j){ // +/-1 works this way
      res(i)=(val_(i+1)-val_(i-1))*fac;
      res(j)=(val_(j+1)-val_(j-1))*fac;
   }
   res(0)=(val_(1)-val_(d_-1))*fac;
   res(d_-1)=(val_(0)-val_(d_-2))*fac;
   res(d_) = (val_(d_+1)-val_(n_-1))*fac;
   res(n_-1)= (val_(d_)-val_(n_-2))*fac;
   CPM_MZ
   return SttD1(res,s_);
}

SttD1 SttD1::applyCC2(CC2 const& mat)const
{
   Cs m11=mat[1][1].std();
   Cs m12=mat[1][2].std();
   Cs m21=mat[2][1].std();
   Cs m22=mat[2][2].std();
   Z i,j;
   Vc res(n_);
   for (i=0,j=d_;i<d_;++i,++j){
      res(i)=m11*val_(i)+m12*val_(j);
      res(j)=m21*val_(i)+m22*val_(j);
   }
   SttD1 res2=SttD1(res,s_);
   res2.setSup_(sup_);
   return res2;
}

SttD1 SttD1::cc()const
{
   return con().applyCC2(alpha1_);
}

SttD1 SttD1::tr()const
// sig3 found by experimentation
{
   return con().applyCC2(beta_);
}

SttD1 SttD1::sr()const
// space reversal
// Without the factor beta_ we already have the correct commutation
// relations with position and momentum but not with the hamiltonian.
{
   SttD1 res=*this;
   res.spaceReversal_();
   return res.applyCC2(beta_);
}

R SttD1::ortTest(V<SttD1> const& vs)
{
   R err=0.;
   for (Z i=vs.b();i<=vs.e();++i){
      SttD1 si=vs[i];
      for (Z j=vs.b();j<=vs.e();++j){
         err+=((si|vs.cui(j))- C(krn(i,j))).abs();
      }
   }
   return err;
}

SttD1 SttD1::operator *(R_Func const& f)const
{
   Z i,j,k;
   V<R> x=b_.points();
   Vc res(n_);
   for (i=1,j=0,k=d_;i<=d_;++i,++j,++k){
      R fi=f(x[i]);
      res(j)=val_(j)*fi;
      res(k)=val_(k)*fi;
   }
   return SttD1(res,s_);
}

C SttD1::operator |(SttD1 const& psi)const
{
   return C(val_.dot(psi.val_));
}

namespace{
   R const redFactor=0.25;
   // used not only in valMat
}

R_Matrix SttD1::valMat(bool logScale)const
{
   Z mL=3;
   Word loc("SttD1::valMat()");
   CPM_MA
   R_Func f=R_Func::log10;
   if (!sup_.contains(s_)){ // only then the support should be indicated
      R_Matrix y(6,d_);
      R_Vector x=pos();
      R vs=valSup2();
      R v0=valInf2();
      if (logScale){
         vs=f(vs); v0=f(v0);
      }
      for (Z i=1;i<=d_;i++){
         C2 c2=val(i);
         C s1=c2.x1, s2=c2.x2;
         R s11=s1[1], s12=s1[2], s21=s2[1], s22=s2[2];
         R q11=s11*s11,q12=s12*s12,q21=s21*s21,q22=s22*s22;
         R a2=q11+q12+q21+q22;
         if (logScale){ a2=f(a2); q11=f(q11); q12=f(q12); q21=f(q21);
            q22=f(q22);
         }
         y[1][i]=a2;
         y[2][i]=q11;
         y[3][i]=q12;
         y[4][i]=q21;
         y[5][i]=q22;
         y[6][i]=sup_.contains(x[i]) ? vs*redFactor : v0;
      }
      CPM_MZ
      return y;
   }
   else{
      R_Matrix y(5,d_);
      for (Z i=1;i<=d_;i++){
         C2 c2=val(i);
         C s1=c2.x1, s2=c2.x2;
         R s11=s1[1], s12=s1[2], s21=s2[1], s22=s2[2];
         R q11=s11*s11,q12=s12*s12,q21=s21*s21,q22=s22*s22;
         R a2=q11+q12+q21+q22;
         if (logScale){ a2=f(a2); q11=f(q11); q12=f(q12); q21=f(q21);
            q22=f(q22);
         }
         y[1][i]=a2;
         y[2][i]=q11;
         y[3][i]=q12;
         y[4][i]=q21;
         y[5][i]=q22;
      }
      CPM_MZ
      return y;
   }
}

void SttD1::mark(Graph& g, bool logScale )const
{
   Z mL=3;
   Word loc("SttD1::mark(Graph)");
   CPM_MA
   using namespace CpmImaging;
   X2<R_Vector,R_Matrix> xy=valVecMat(logScale); // valVecMat is defined in
      // cpmsttd1.h in terms of valMat()
   V<Color> cl{BLACK,GREEN,RED,LIGHTCYAN,LIGHTMAGENTA,LIGHTGRAY};
   if (sup_.contains(s_)) cl=cl.cut(1); // takes away the last component
   g.clear();
   g.markGL(xy.c1(),xy.c2(),cl);
   CPM_MZ
}

void SttD1::markArr(V<SttD1> const& vs, Graph& g)
{
   Z mL=3;
   Word loc("SttD1::markArr(V<SttD1>,Graph)");
   CPM_MA

   using namespace CpmImaging;
   using namespace CpmArrays;

   V<Color> cl(10);
   cl[1]=LIGHTGRAY;
   R fac=0.6;
   Z nL=2;
   V<R> list=Iv(0.,1.).div(nL);
   for (Z i=1;i<=nL;++i){
      Color c1=Color::hue(list[i]);
      Color c2=(c1*fac).addWhite(1.);
      Z j=2*i;
      cl[j]=c1;
      cl[j+1]=c2;
   }
// darker colors
   cl[6]=BLACK; // color for absolute value
   fac=0.75;
   cl[7]=cl[2]*fac;  // color for real part of v1
   cl[8]=cl[3]*fac;   // color for imaginary part v1
   cl[9]=cl[4]*fac;
   cl[10]=cl[5]*fac;
   Z i;
   SttD1 s1=vs.fir(); // making sure that the states match
   for (i=vs.b(); i<=vs.e(); ++i){
      cpmassert(s1.mat(vs[i]),loc);
   }
   R_Vector x=s1.valVec();
   R_Matrix y=s1.valMat();
   for (i=vs.b()+1; i<=vs.e(); ++i){
      y&=vs[i].valMat(); // appending the matrix part
   }
   g.clear();
   g.addText("SttD1. Real part: full color, Imaginary part: light color,\
 Abs: black");
   g.markGL(x,y,cl);
   CPM_MZ
}

void SttD1::mark2(Graph& g, SttD1 const& sc, R const& scale)const
      // Shows *this together with a comparison state
      // allows to boost the difference between the two by a factor.
{
   SttD1 s1(*this);
   SttD1 s2(sc);
   SttD1 sm=(s1+s2)*0.5;
   R sh=scale*0.5;

   SttD1 sDiff=(s1-s2)*sh;
   SttD1 v1=sm+sDiff;
   SttD1 v2=sm-sDiff;
   SttD1::markArr(V<SttD1>{v1,v2},g);
}

namespace{
   R convertC2toR(C2 const& c2){ return c2.abs2();}
}

R SttD1::valSup2()const
{
   V<C2> vc=val();
   F<C2,R> fc(convertC2toR);
   V<R> vr=vc(fc);
   R_Vector amp(vr);
   return amp.sup();
}

R SttD1::valInf2()const
{
   V<C2> vc=val();
   F<C2,R> fc(convertC2toR);
   V<R> vr=vc(fc);
   R_Vector amp(vr);
   return amp.inf();
}

void SttD1::gp(ostream& out, bool logScale)const
{
   Z mL=1;
   Word loc("SttD1::gp(Word,bool)");
   CPM_MA
   R_Func f=R_Func::log10;
   V<C2> vc=val();
   R_Vector x=pos();

   if (sup_.contains(s_)){ // no additional indicator function
      for (Z i=1;i<=d_;i++){
         C2 c2=vc[i];
         R  xi=x[i];
         R q0=c2.abs2();
         C s1=c2.x1, s2=c2.x2;
         R r11=s1[1], r12=s1[2], r21=s2[1], r22=s2[2];
         R q11=r11*r11, q12=r12*r12, q21=r21*r21, q22=r22*r22;
         if (logScale){
            q0=f(q0); q11=f(q11); q12=f(q12); q21=f(q21); q22=f(q22);
         }
         out<<endl<<" "<<xi<<" "<<q0<<" "<<q11<<" "<<q12<<" "
         <<q21<<" "<<q22;
      }
   }
   else{
      R q0Sup=valSup2();
      R q0Inf=valInf2();
      if (logScale){
         q0Sup=f(q0Sup);
         q0Inf=f(q0Inf);
      }
      for (Z i=1;i<=d_;i++){
         C2 c2=vc[i];
         R  xi=x[i];
         R q0=c2.abs2();
         C s1=c2.x1, s2=c2.x2;
         R r= sup_.contains(xi) ? q0Sup*redFactor : q0Inf;
         R r11=s1[1], r12=s1[2], r21=s2[1], r22=s2[2];
         R q11=r11*r11, q12=r12*r12, q21=r21*r21, q22=r22*r22;
         if (logScale){
            q0=f(q0); q11=f(q11); q12=f(q12); q21=f(q21); q22=f(q22);
         }
         out<<endl<<" "<<xi<<" "<<q0<<" "<<q11<<" "<<q12<<" "
         <<q21<<" "<<q22<<" "<<r;
      }
   }
   CPM_MZ
}

V<C2> SttD1::val()const
{
   Z i,j,k;
   V<C2> res(d_);
   for (i=1,j=0,k=d_;i<=d_;++i,++j,++k)
      res.cui(i)=C2(C(val_(j)),C(val_(k)));
   return res;
}

SttD1 SttD1::ran(Z j)const
// For a zero-vector psi,the vector psi.ran() should not be a zero-vector
{
   Z dd=d();
   C z(1.,0.5);
   Vr<C2> vr(dd,C2(z,z));
   SttD1 res(vr.ran(j),s_);
   res.nor_();
   return res;
}

R SttD1::dis(SttD1 const& psi)const
{
   Z mL=3;
   Word loc("StateD::dis(StateD)");
   CPM_MA
   V<C2> v2=val(), v2psi=psi.val();
   V<R> x=pos(), xpsi=psi.pos();

   R_Vector d(3);
   d[1]=v2.dis(v2psi);
   d[2]=x.dis(xpsi);
   d[3]=s_.dis(psi.s_);
   R res=d.supAbs();
   CPM_MZ
   return res;
}

bool SttD1::prnOn(ostream& str)const
{
   cpmwat;
   cpmp(C_Vector(val_));
   cpmp(s_);
   return true;
}

bool SttD1::scanFrom(istream& str)
{
   C_Vector cv;
   cpms(cv);
   val_=cv.toEigen();
   cpms(s_);
   adjust_();
   return true;
}

bool SttD1::isVal()const
{
   for (Z i=0;i<n_;++i){
      if (!C(val_(i)).isVal()) return false;
   }
   return true;
}

//////////////////////////// class ObsD1 //////////////////////////////////

ObsD1::ObsD1(V<SttD1> const& can, F<SttD1,SttD1> const& f):
   val_(can.dim(),can.dim())
{
   Z mL=1;
   Word loc("ObsD1(V<SttD1>,F<SttD1,SttD1>)");
   CPM_MA
   Z d=can.dim(),i,j,k,l;
   for (i=1,j=0;i<=d;++i,++j){
      for (k=1,l=0;k<=d;++k,++l){
         val_(j,l)= (can[i]|f(can[k])).std();
      }
   }
   adjust_();
   CPM_MZ
}

ObsD1 ObsD1::applyCC2(Bio const& b, CC2 const& mat)
{
   Z mL=3;
   Word loc("ObsD1::applyCC2(Bio,CC2)");
   CPM_MA
   Z i,j,k,l;
   Z d=b.d(),n=b.dd();
   Cs m11=mat.x1.x1.std();
   Cs m12=mat.x1.x2.std();
   Cs m21=mat.x2.x1.std();
   Cs m22=mat.x2.x2.std();
   Mc val=Mc::Constant(n,n,Cs(0.));
   for (i=0,j=d;i<d;i++,j++){
      val(i,i)=m11;
      val(i,j)=m12;
      val(j,i)=m21;
      val(j,j)=m22;
   }
   CPM_MZ
   return ObsD1(val);
}

ObsD1::ObsD1(CC2 const& c, Bio const& b)
{
   *this=ObsD1::applyCC2(b,c);
}

ObsD1 ObsD1::x(Bio const& b)
{
   Z mL=3;
   Word loc("ObsD1::x(Bio)");
   CPM_MA
   V<R> vx=b.points();
   vx.b_(0); // now the valid indexes of vx start with 0
   Z i,j;
   Z d=b.d(),n=b.dd();
   Mc val=Mc::Constant(n,n,Cs(0.));
   for (i=0,j=d;i<d;i++,j++){
      Cs xi=vx[i];
      val(i,i)=xi;
      val(j,j)=xi;
   }
   CPM_MZ
   return ObsD1(val);
}

ObsD1 ObsD1::x(Bio const& b, R const& x0)
{
   Z mL=3;
   Word loc("ObsD1::x(Bio,R)");
   CPM_MA
   V<R> vx=b.points();
   vx.b_(0); // now the valid indexes of vx start with 0
   Z i,j;
   Z d=b.d(),n=b.dd();
   Mc val=Mc::Constant(n,n,Cs(0.));
   for (i=0,j=d;i<d;i++,j++){
      Cs xi=vx[i]-x0;
      val(i,i)=xi;
      val(j,j)=xi;
   }
   CPM_MZ
   return ObsD1(val);
}

ObsD1 ObsD1::x(Bio const& b, R_Vector const& f)
{
   Z mL=3;
   Word loc("ObsD1::x(Bio,R_Vector)");
   CPM_MA
   V<R> vx=b.points();
   vx.b_(0); // now the valid indexes of vx start with 0
   Z i,j;
   Z d=b.d(),n=b.dd();
   Mc val=Mc::Constant(n,n,Cs(0.));
   for (i=0,j=d;i<d;i++,j++){
      Cs fi=f[i+1];
      val(i,i)=fi;
      val(j,j)=fi;
   }
   CPM_MZ
   return ObsD1(val);
}

ObsD1 ObsD1::x(Bio const& b, R_Func const& f)
{
   Z mL=3;
   Word loc("ObsD1::x(Bio,R_Func)");
   CPM_MA
   V<R> vx=b.points();
   vx.b_(0); // now the valid indexes of vx start with 0
   Z i,j;
   Z d=b.d(),n=b.dd();
   Mc val=Mc::Constant(n,n,Cs(0.));
   for (i=0,j=d;i<d;i++,j++){
      Cs fi=f(vx[i]);
      val(i,i)=fi;
      val(j,j)=fi;
   }
   CPM_MZ
   return ObsD1(val);
}

ObsD1 ObsD1::tp(Bio const& b)
{
   Z mL=3;
   Word loc("ObsD1::tp(Bio)");
   CPM_MA
   Cs fac = Cs(0.,b.gapInv()*0.5);
   Z i,j,k,l;
   Z d=b.d(),n=b.dd();
   Mc val=Mc::Constant(n,n,Cs(0.));
   for (i=0,j=d;i<d-1;i++,j++){
      val(i,i+1)=fac;
      val(j,j+1)=fac;
   }
   val(d-1,0)=fac;
   val(n-1,d)=fac;
   CPM_MZ
   return ObsD1(val);
}

ObsD1 ObsD1::tn(Bio const& b)
{
   Z mL=3;
   Word loc("ObsD1::tn(Bio)");
   CPM_MA
   Cs fac = Cs(0.,b.gapInv()*0.5);
   Z i,j,k,l;
   Z d=b.d(),n=b.dd();
   Mc val=Mc::Constant(n,n,Cs(0.));
   for (i=1,j=d+1;i<d;i++,j++){
      val(i,i-1)=fac;
      val(j,j-1)=fac;
   }
   val(0,d-1)=fac;
   val(d,n-1)=fac;
   CPM_MZ
   return ObsD1(val);
}

ObsD1 ObsD1::p(Bio const& b)
{
   return ObsD1::tn(b)-ObsD1::tp(b);
      // with tp and tn interchanged (by error) the boost series loook
      // very similar but with rho and -rho interchanged. 2016-05-24.
}

ObsD1 ObsD1::h(Bio const& b, R const& m)
{
   CC2 alpha=SttD1::alpha1_;
   CC2 betam=SttD1::beta_*m;
   return ObsD1(alpha,b)*p(b) + ObsD1(betam,b);
}

ObsD1 ObsD1::h(Bio const& b, R const& m, R_Func const& pot)
{
   CC2 alpha=SttD1::alpha1_;
   CC2 betam=SttD1::beta_*m;
   return ObsD1(alpha,b)*p(b) + ObsD1(betam,b) + ObsD1::x(b,pot);
}

ObsD1 ObsD1::h(Bio const& b, R const& m, R_Vector const& pot)
{
   CC2 alpha=SttD1::alpha1_;
   CC2 betam=SttD1::beta_*m;
   return ObsD1(alpha,b)*p(b) + ObsD1(betam,b) + ObsD1::x(b,pot);
}

ObsD1 ObsD1::booster(Bio const& b, R const& m)
{
   ObsD1 h = ObsD1::h(b,m);
   ObsD1 x = ObsD1::x(b);
   return (x*h+h*x)*R(0.5);
}

///////////////////////////// class SpcObjD1 //////////////////////////////

SpcObjD1::SpcObjD1(const ObsD1& a, Bio const& bio, bool test):b_(bio),
spectrum_(b_.dd()), eigenvectors_(b_.dd())
{
   const Z mL=1;
   Word loc("SpcObjD1(ObsD1,Bio,bool)");
   CPM_MA
   d_=a.d();
   cpmassert(b_.d()==d_,loc);
   n_=2*d_;

   Z i,j,k,l;
   SelfAdjointEigenSolver<Mc> ses(a.val_);
   sp_=ses.eigenvalues();
   ev_=ses.eigenvectors();
   eva_=ev_.adjoint();
   for (i=1,k=0;i<=n_;++i,++k) spectrum_.cui(i)=C(sp_(k));
   for (i=1,k=0;i<=n_;++i,++k) eigenvectors_.cui(i)=
      SttD1(ev_.col(k),b_.iv());

   if (test){
      Mc diag=Mc::Constant(n_,n_,Cs(0.));
      for (i=0;i<n_;++i) diag(i,i)=Cs(sp_(i));
      R err = (a.val_ - ev_*diag*eva_).squaredNorm();
      cout<<"SpcD1 test, err = "<<err<<endl;
      R errOrtho=0.;
      for (i=1;i<=n_;i++) for (j=1;j<=n_;j++)
      errOrtho+=
         ((eigenvectors_[i]|eigenvectors_[j])-C(krn(i,j))).abs();
      cout<<"SpcD1 test, errOrtho = "<<errOrtho<<endl;
      R errSpectral=0.;
      for (i=1;i<=n_;i++)
      errSpectral+=
         ( a*eigenvectors_[i]- eigenvectors_[i]*spectrum_[i]).abs();
      cout<<"SpcD1 test, errSpectral = "<<errSpectral<<endl;
   }
   CPM_MZ
}

SttD1 SpcObjD1::operator*(SttD1 const& st)const
{
   Z mL=1;
   Word loc("SpcObjD1::operator*(SttD1)");
   CPM_MA
   SttD1 res(b_);
   cpmassert(st.mat(res),loc);
   for (Z i=1;i<=n_;i++){
      SttD1 evi=eigenvectors_.cui(i);
      C ci=(evi|st)*spectrum_.cui(i);
      res+=evi*ci;
   }
   CPM_MZ
   return res;
}

ObsD1 SpcObjD1::operator()(F<R,C> const& f)const
{
   Z mL=1;
   Word loc("SpcObjD1::operator()(F<R,C>)");
   CPM_MA
   Mc diag=Mc::Constant(n_,n_,Cs(0.));
   for (Z i=0;i<n_;++i) diag(i,i)=f(sp_(i)).std();
   ObsD1 res(ev_*diag*eva_);
   CPM_MZ
   return res;
}

ObsD1 SpcObjD1::map(R_Func const& f)const
{
   Z mL=1;
   Word loc("SpcObjD1::map(R_Func)");
   CPM_MA
   Mc diag=Mc::Constant(n_,n_,Cs(0.));
   for (Z i=0;i<n_;++i) diag(i,i)=Cs(f(sp_(i)));
   ObsD1 res(ev_*diag*eva_);
   CPM_MZ
   return res;
}

R_Vector SpcObjD1::sp()const
{
   R_Vector res(n_);
   Z i,j;
   for (i=1,j=0;i<=n_;++i,++j) res.cui(i)=sp_(j);
   return res;
}

SttD1 SpcObjD1::apply(F<R,C> const& f, SttD1 const& st)
{
   Z mL=1;
   Word loc("SpcObjD1::apply(F<R,C>,SttD1)");
   CPM_MA
   SttD1 res(b_);
   cpmassert(st.mat(res),loc);
   for (Z i=1;i<=n_;i++){
      C ci=(eigenvectors_.cui(i)|st)*f(spectrum_.cui(i).real());
      res+=eigenvectors_.cui(i)*ci;
   }
   CPM_MZ
   return res;
}

SttD1 SpcObjD1::apply(R_Func const& f, SttD1 const& st)
{
   Z mL=1;
   Word loc("SpcObjD1::apply(R_Func,SttD1)");
   CPM_MA
   SttD1 res(b_);
   cpmassert(st.mat(res),loc);
   for (Z i=1;i<=n_;i++){
      C ci=(eigenvectors_.cui(i)|st)*C(f(spectrum_.cui(i).real()));
      res+=eigenvectors_.cui(i)*ci;
   }
   CPM_MZ
   return res;
}
