//? cpmdynaux.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_DYNAUX_H_
#define CPM_DYNAUX_H_
/*
   Description:auxiliar classes for dynamical systems
*/
#include <cpmtypes.h>
#include <cpmrecord.h>
#include <cpminterval.h>
#include <cpmdimdef.h>

namespace CpmDynSys{

   using CpmRoot::R;
   using CpmRoot::Z;
   using CpmRoot::Word;
   using CpmRootX::R1;
   using CpmRootX::Z1;
   using CpmRootX::B;
   using CpmRootX::hugeNumber;
   using CpmRootX::Record;

   using CpmGeo::Iv;


struct StepControl{ // a data collection providing I/O and initialization
    // and an important method done()

   R1 t;
      // time to which the present state *this refers. I hesitated for
      // some time to hide this important physical quantity among the
      // more technical auxiliaries that follow. However the information
      // flow strongly suggests this placement of t

   R1 tLast;
      // time to which the present state *this refered at the end of
      // the last step.

   R1 dtLast;
      // t-tLast

   R1 tLimit;
      // in function stepAsy(R&) (which implements asynchronous time step
      // adjustment) no final time larger than this value will be
      // proposed or reached by any intermediate time step.

   B relaxFlag;
      // if relaxFlag is true, function step() implements a 'motion
      // neglecting inertia' by setting all particle velocities to zero
      // at the end of step(). Setting this parameter to 'true' also sets
      // all particle velocities to zero. Thus an equilibrium will
      // effectively approached without 'overshooting'. If, for instance,
      // two particles overlap due to either a violent collision or a
      // careless setting of initial positions, on can achieve a
      // non-overlapping state by doing a few steps for relaxFlag==true.
      // Other than with normal dynamics, one will not get an 'exploding
      // state' but one with all particles at rest. State changes for
      // relaxFlag==true are considered now (since 2000-2-3) as processes
      // in time, i.e. the function done() does the updating of t,
      // dtMean, dtLast just as if relaxFlag were false. Also this is not
      // always the most natural choice from physics it is very economic
      // since so the function run can used to organize and visualize also
      // state evolution under 'relaxing dynamics'.

   R1 dtMean;
      // a floating mean value of the dt-values used over the near past.
      // The precise meaning results from the code in function done().

   R1 disLast;
      // The quantity which measures the difference between a state
      // obtained by a single step and the succession of two half steps
      // that occurred in obtaining the present step is an important
      // quantity in analysing the performance of the timestep adjustment
      // algorithm.

   R1 disMean;
      //  a floating mean value of the dis-values used over the near past.

   void update();

   typedef StepControl Type;
   CPM_IO
   StepControl():tLimit(hugeNumber){}
      // meaningful initialization helps to keep the constructors of
      // ParDynSys<> simple
   void done(R dt_, R dis_);
      // increasing the time by dt_ achieved
   void doneTill(R t_, R dis_);
      // increasing the time to value t_ achieved
};

void adjust_dt(R&,Z);

} // namespace

#endif
