//? cpmextractorx.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*
   see cpmextractor.h
*/
   using namespace CpmStd;

   using CpmRoot::Word;
   using CpmRoot::R;
   using CpmRoot::Z;
   using CpmRoot::L;
   using CpmRootX::B;
   using CpmRootX::Record;

   using CpmArrays::VV;
   using CpmArrays::Vp;
   using CpmArrays::R_Vector;

   using CpmSystem::IFile;
   using CpmSystem::OFile;

   using CpmGraphics::ColRef;
   using CpmImaging::ShowControl;
   using CpmImaging::Color;
   using CpmImaging::WHITE;
   using CpmImaging::YELLOW;

   using CpmAlgorithms::CyclicAlarm;
   using CpmAlgorithms::NonCyclicAlarm;

   using CpmProperties::SysControl;

   using CPM_DIM::Spc;

   using CPM_CAMERA::Camera;

// needs to agree with code defining Sys2(RecordHandler)
   using CPM_PHYSICS::Lid;
   using CPM_PHYSICS::Pot;
   using CPM_PHYSICS::Fen;
   using CPM_PHYSICS::SphTank;
   using CPM_PHYSICS::HomFil;
   using CPM_PHYSICS::Drum;
  // using CPM_PHYSICS::Mill;
      // Exact types to which the external systems of Sys2 may belong.
      // These are class names as tokens; these tokens can't be
      // recovered from data of type Word (which can easily communicated
      // to the present class via reading from file). So there is a lack
      // (so far ?) in supporting polymorphism of classes that get
      // transpoted via files. This is only a question of elegance,
      // though.
//doc namespace CpmPala2_x{
////////////////////////////// class ParExt ///////////////////////
template <class CC,class CV,class PT,class EX,class Sys>
class ParExt{ // particle extractor
   // for reading in a uniform manner states of
   // particle arrays and external systems from
   // PaLa 'movie'-files

   V<Word> exTypes;
      // list of types of external sytems
   Vp<EX> ex;
      // polymorphic array of external systems
   Partition parti;
   B partiPresent;
   V<CC> vc;
   V<CV> vv;
   V<PT> vp;
   Z1 np,ne,count;
   B valid;
   IFile file;
      // here it is not sufficient to hold the file name.
      // During the existence of a ParExt instance
      // a stream object has to be active for reading
      // from file.
   OFile outFile;
      // text file for non-graphical output
   void initialize(const Word& fileName,  const Word& outFileName);

public:
   CPM_NAM(ParExt<...>)
   ParExt(){}
      // trivial, all data initialized by default, especially valid=false
      // No file gets opened

   ParExt(Word const& fileName, Word const& outFileName=Word())
      { initialize(fileName,outFileName);}
      // creates such a state that after a call of next()
      // the first data-frame of file is loaded into *this
      // and can be accessed through the access functions to
      // be defined soon. If outFileName!=Word() an output
      // file named outFileName gets created so that it is open
      // for writing on it.

   bool next(bool tolerant=false);
      // loading the data with the values from the next frame
      // and returning whether this was done successfully
      // tolerant==true is advisable in applications where
      // reading partially completed files necessarily occurs and
      // is handled properly. The the warnings which would otherwise
      // occur in reading a incomplete last image  will be suppressed
      // if tolerant==true

// accessing the frame data read from file
   V<PT>& get_vp(){ return vp;}
   V<PT> const& get_vp()const{ return vp;}

   V<CC>& get_vc(){ return vc;}
   V<CC> const& get_vc()const{ return vc;}

   Vp<EX>& get_ex(){ return ex;}
   Vp<EX> const& get_ex()const{ return ex;}

   Partition& get_parti(){ return parti;}
   const Partition& get_parti()const{ return parti;}

   bool isVal()const{ return valid;}
   Z get_np()const{ return np;}
   Z get_ne()const{ return ne;}
   Z getCount()const{ return count;}
      // returns the number of the frame, the data of which
      // are presently accessible
   void show(Camera& ca, ShowControl const& sc);
      // representing the configuration as a screen image under
      // a variety of controls, which are bundled in struct
      // ShowControl.
};

template <class CC,class CV,class PT,class EX, class Sys>
void ParExt<CC,CV,PT,EX,Sys>::initialize
   (Word const& fileName, Word const& outFileName)
{
   const Z mL=1;
   Word loc( "ParExt<>:initialize(" );
   loc&=fileName;
   loc&=", ";
   loc&=outFileName;
   loc&=")";
   CPM_MA
   vc=V<CC>(0); // the default constructor creates
      // arrays of dimension 1 instead of the more economic
      // ones of dimension 0
   vv=V<CV>(0);
   //Word fn=Message::ext(fileName);
   file=IFile(fileName);
   //file=IFile(fn);
   if (outFileName.dim()>0) outFile=OFile(outFileName);
   bool succ;
// partition
   succ=parti.scanFrom(file());
      // We write the partition only once at the beginning and thus
      // abandon the elaborate mechanism which allowed to continue
      // long moviefiles without the need to read through them
      // for getting the partition from the beginning
   if (!succ){
      Word mes=loc&" not able to read a partition";
      cpmerror(mes);
   }
   else{
      Word mes=loc&" was able to read partition";
      cpmmessage(mes);
      cpmcerr<<"parti="<<endl<<parti<<endl;
   }
// reading exTypes
   succ=exTypes.scanFrom(file());
   if (!succ){
      Word mes=loc&" not able to read exTypes";
      cpmerror(mes);
   }
   else{
      Word mes=loc&" was able to read exTypes";
      cpmmessage(mes);
      cpmcerr<<"exTypes="<<endl<<exTypes<<endl;
   }
// reading vc
   succ=vc.scanFrom(file());
   if (!succ){
      Word mes=loc&" not able to read vc";
      cpmerror(mes);
   }
   else{
      Word mes=loc&" was able to read vc";
      cpmmessage(mes);
   }
   np=vc.dim();
   ostringstream ost;
   ost<<endl<<"vc read for vc.dim()=np="<<np<<endl;
   cpmmessage(ost);
   if (np==0) cpmwarning("!!! no particles read from file !!!");
// initializing vp
   vp=V<PT>(np);
// initializing ex with proper types of derived classes
// needs to agree with code defining Sys2(RecordHandler)
   ne=exTypes.dim();
   ex=Vp<EX>(ne);
   for (Z i=1;i<=ne;++i){
      Word wi=exTypes[i];
      if (wi=="Lid") ex[i]=Lid();
      else if (wi=="Pot") ex[i]=Pot();
      else if (wi=="Fen") ex[i]=Fen();
      else if (wi=="SphTank") ex[i]=SphTank();
      else if (wi=="HomFil") ex[i]=HomFil();
      else if (wi=="Drum") ex[i]=Drum();
      else cpmerror("unknown type of external system found");
   }
   valid=true;
   CPM_MZ
}

template <class CC,class CV,class PT,class EX, class Sys>
bool ParExt<CC,CV,PT,EX,Sys>::next(bool tolerant)
{
   const Z mL=2;
   Word loc(nameOf()&"::next()");
   CPM_MA
   cpmassert(valid==true,loc);
   if (!valid){
      ostringstream ost;
      ost<<loc<<" can't do a step for unvalid state";
      cpmwarning(ost);
      cpmmessage(mL,loc&" done , !valid");
      return valid;
   }
   bool succ;
// reading vv
   succ=vv.scanFrom(file());
   if (!succ || vv.dim()!=np){
      valid=false;
      if (!tolerant){
         ostringstream ost;
         ost<<endl<<loc<<
         " reading vv failed, vv.dim()="<<vv.dim()<<" np="<<np;
         cpmwarning(ost);
      }
      cpmmessage(mL,loc&" done, !succ");
      return valid;
   }
// setting the particles defined by vc and vv
   cpmmessage(mL,loc&" number of particles is "&cpm(np));
   R npInv=(np==0 ? 0 : 1./np);
   Z countProblem=0;
   for (Z j=1;j<=np;j++){
      if (cpmverbose>mL) cpmprogress("read particles",R(j-1)*npInv);
      vp[j]=PT(vc[j],vv[j]);
      bool val=vp[j].isVal();
      if (!val){
         Z id=vp[j].get_id();
         ostringstream ost;
         ost<<loc<<" particle with id="<<id<<" is not valid";
         cpmwarning(ost);
         countProblem++;
      }
      Spc origin;
      val=(vp[j].get_x()==origin);
      if (val){
         Z id=vp[j].get_id();
         ostringstream ost;
         ost<<loc<<" particle with id="<<id<<" is placed at origin";
         cpmwarning(ost);
         countProblem++;
      }
   }
   cpmmessage(mL,"np-loop done");
   if (countProblem>0){
      ostringstream ost;
      ost<<"There were "<<countProblem<<" dubious particles"  ;
      cpmwarning(ost);
   }
// reading ex
   Z i;
   cpmmessage(mL,"need to read "&cpm(ne)&" external systems from file");
   for (i=1;i<=ne;++i){
      cpmmessage(mL,"ex("&cpm(i)&") loop entry started");
      Word namei=ex(i).nameOf();
      cpmmessage(mL,"class type is "&namei);
      succ=ex(i).scanFrom(file());
      if (!succ){
         ostringstream ost;
         ost<<endl<<loc<<" reading ex["<<i<<"] failed";
         cpmwarning(ost);
         valid=false;
         cpmwarning(loc&" done, !succ");
         return valid;
      }
      cpmmessage(mL,"ex("&cpm(i)&") loop entry done");
   }
   count+=1;
   if (cpmverbose>=mL && valid){
      ostringstream ost;
      ost<<"Frame "<<count<<" read";
      cpmmessage(ost);
   }
   if (!valid && !tolerant){
      ostringstream ost;
      ost<<loc<<" unable to read frame "<<count;
      cpmwarning(ost);
   }
   cpmmessage(mL,loc&" done OK");
   return valid;
}

template <class CC,class CV,class PT,class EX, class Sys>
void ParExt<CC,CV,PT,EX,Sys>::show(Camera& ca, ShowControl const& sc)
// needs class Sys only if sc.modality=="normalView"
// The other modalities need ComPar::showProperty(...)
// (=PT::showProperty(...)).
{
   const Z fieldLength=4;
      // formatting in writing the current image number
   Z mL=3;
   static Word loc("ParExt<>::show(Camera,ShowControl)");
   CPM_MA
   static Color colTon(YELLOW);
   static Color colColorize(WHITE);
   ostringstream ost;
   ost<<endl<<"np="<<np<<" count="<<count<<endl;
   cpmmessage(mL,Word(ost));
   Z offSet=sc.offSet;
   if (count<offSet) return;
   Z i,j,npLoc=np;
      // don't change np since this spoils the execution of next()
      // the next time. Changing vp does not hurt, since next() builds
      // it from scratch.
   bool waitDone=false, succ=false, cclear=false;
   Word text="frame "&cpmwrite(count);
   CpmArrays::VV<ColRef> imgDat;
// an image should be formed if viewing period or imaging period is met
// and the number of offset images is passed or if the one wanted image
// is reached
   Z lastFrameNumber=sc.nocyalImg.last();
   bool scb1=sc.cyalImg();
   bool scb2=sc.nocyalImg();
   bool cimg=scb1||scb2;
   bool cview=sc.cyalView();
   if (cview) cclear=sc.cyalClear();
   bool final=(lastFrameNumber==count);

   parti.autoFrame1(ca,sc.autoFrameFactor);

   if ( cimg || cview || final ){ // only than the movie frame
      // needs to become processed
      bool caclear=cclear || cimg ;
      if (final) caclear=false;

// restricting the set of particles according to position
      V<PT> vpLoc=vp;
      if (sc.boxSel){
         V<bool> sel(np);
         for (i=1;i<=np;i++){
            Spc xi=vp[i].get_x();
            bool sx=sc.ivx.contains(xi[1]);
            bool sy=sc.ivy.contains(xi[2]);
            bool sz=Spc::dimension()==2 ? true : sc.ivz.contains(xi[3]);
            sel[i]=sx&&sy&&sz;
         }
         vpLoc=vp.select(sel);
         npLoc=vp.dim();
         cpmmessage(mL,"# of particles after boxSel="&cpm(npLoc));
      }
// main program control as a if control structure
      if (sc.modality=="HistogramTonerVelocity"){
         cpmmessage(mL,loc&": HistogramTonerVelocity selected");
         R_Vector parameters(1);
         parameters[1]=SysControl::vCrit;
         PT::showProperty(sc.modality,vpLoc,ca,parameters);
      }
      else if (sc.modality=="HistogramCarrierVelocity"){
         R_Vector parameters(1);
         parameters[1]=SysControl::vCrit;
         cpmmessage(mL,loc&": HistogramCarrierVelocity selected");
         PT::showProperty(sc.modality,vpLoc,ca,parameters);
      }
      else if (sc.modality=="Histogram_xPosition"){
         cpmmessage(mL,loc&": Histogram_xPosition selected");
         R_Vector parameters(1);
         R xMax=parti.get_xLength()*0.5;
         parameters[1]=xMax;
         PT::showProperty(sc.modality,vpLoc,ca,parameters);
      }
      else if (sc.modality=="Histogram_yPosition"){
         cpmmessage(mL,loc&": Histogram_yPosition selected");
         R_Vector parameters(1);
         R yMax=parti.get_yLength()*0.5;
         parameters[1]=yMax;
         PT::showProperty(sc.modality,vpLoc,ca,parameters);
      }
      else if (sc.modality=="Histogram_zPosition"){
         cpmmessage(mL,loc&": Histogram_zPosition selected");
         R_Vector parameters(1);
         R zMax=parti.get_zLength()*0.5;
         parameters[1]=zMax;
         PT::showProperty(sc.modality,vpLoc,ca,parameters);
      }
      else if (sc.modality=="normalView"){
         cpmmessage(mL,loc&": normalView selected");
         Sys sys(vpLoc,ex,parti);
         Z nLoc=vpLoc.dim();
         for (i=1;i<=nLoc;++i){
            PT pi=sys.getPar(i);
         }
         if (SysControl::showMeanFil) sys.prepareStep();
         Z lcd=sc.listColorize.dim();
         if (lcd>0){
            cpmmessage(mL,loc&": "&cpm(lcd)&" particles colorized");
            for (j=1;j<=lcd;j++){
               Z sclc=sc.listColorize[j];
               sys.setShowColor(sclc,colColorize);
            }
         }
         Z ipd=sc.invisibleParticles.dim();
         if (ipd>0){
            cpmmessage(mL,loc&": "&cpm(ipd)&" particles made invisible");
            Color civ(CpmImaging::TRANSPARENT_);
            for (j=1;j<=ipd;++j){
               Z scip=sc.invisibleParticles[j];
               sys.setShowColor(scip,civ);
            }
         }
         if (caclear) ca.clear();
         Word textShow=(cpmverbose<1 ? Word() : text);
         sys.mark(ca,cview,textShow); //This does not handle the
        // text insert cleanly if sys.mark calls ScalarField::mark
         cpmmessage(text,2);
      }
      if (final){
         ca.write();
         cpmmessage("point final reached");
      }
      if (cimg){
         sc.countImg+=1;
         Word imf=sc.imgFile;
         imf=imf.cut(4); // cutting ".txt" at the end
         imf=imf.tail(6);
         imf="i_"&imf&"_";
         if (sc.modality=="normalView"&&SysControl::bitmapToShow==3){
            // here an image gets created that has no direct
            // relationship to ca
            CpmGraphics::Img24 img(imgDat);
            img.write(imf&cpm(count,fieldLength));
         }
         else{
            ca.write(imf&cpm(count,fieldLength));
            // this makes no screen action since it calls finally
            // Img24::write
         }
      }
      if (!waitDone) cpmwait(sc.tWait());
   }
   if (sc.maxNumberImages<=sc.countImg || final) valid=false;
   CPM_MZ
}
