//? cpmsys2apscpp.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*
   piece of code for use in different namespace-contexts
*/

using CpmGeo::Angle;
using CpmPhysics::gEarth_PHYS;
using CpmPhysics::gEarth;

using namespace CpmSystem;
using namespace CpmRoot;
using namespace CpmRootX;
using namespace CpmApplication;
using namespace CpmProperties;
using namespace CpmFunctions;
using namespace CpmGraphics;
using namespace CpmImaging;

namespace{

   Z fAux(R const& t, R const& tSpan, R2 const& vC, R2 const& aC,
      R2 const& etaC)
   {
      cpmassert(tSpan>0,"Sys2Appl, fAux");
      R tN=t/tSpan;
      R fac=R(1)-R_Func::hill()(tN);
         // fac starts with 0 and goes to 1
      R vI=vC.x1;
      R vF=vC.x2;

      R aI=aC.x1;
      R aF=aC.x2;

      R etaI=etaC.x1;
      R etaF=etaC.x2;

      R d_v=vF-vI;
      R d_a=aF-aI;
      R d_eta=etaF-etaI;

      SysControl::vCrit=  vI+d_v*fac;
      SysControl::aCrit=  aI+d_a*fac;
      SysControl::eta=  etaI+d_eta*fac;
      return 1; // not needed
   }

   const R rimPaLa=0.;
   const Color cPaLa(Z(150),Z(150),Z(255));
}

void Sys2Appl::doTheWork()
{
   Z mL=1;
   Word loc("Sys2Appl::doTheWork()");
   CPM_MA
   sizeAndRank();
   Z i;
   Frame fr0;
   Graph g(fr0);
   g.setWithOrigin(0);
   g.paint(cPaLa);
   g.vis();
   if (size!=1){ // thus action only for defined(CPM_USE_MPI)
      // here all the files prepared in the base class should
      // get 'ornamented'
      for (i=1;i<=nInFiles();++i){
         inFiles[i]=inFiles[i].appBefExt("_"&cpm(rank));
         cpmmessage("inFiles["&cpm(i)&"]="&inFiles[i]);
      }
      for (i=3;i<=nOutFiles();++i){ // first two will be
         // used by rank 1 only (movie-files)
         outFiles[i]=outFiles[i].appBefExt("_"&cpm(rank));
         cpmmessage("outFiles["&cpm(i)&"]="&outFiles[i]);
      }
   }

   Word sec="selection";
   rch.read(sec,"cpmdbg",CpmSystem::Message::debug);
   R runTimeSpan=1e6;
      // should never come to action
   Z continue_run=0;
   sec="syscontrol";
   SysControl::read(rch,sec);
      // Some of SysControl's data will be overwritten soon.
   Z nSteps1,filingPeriod1,docuPeriod1,viewingPeriod1,
     stoppingPeriod,fixedStep1;
   R dtStep1;
   V<R> vsCrit(2,SysControl::vCrit),asCrit(2,SysControl::aCrit),
        etas(2,SysControl::eta);
      // should get initialized since we should not expect to
      // read this from a ini-file in all cases
   bool readFromFile1,readFromFile2;
   bool ignore1,ignore2;
   F<R,Z> f1(Z(0)),f2(Z(0)); // these are functions R -> Z, r |-> 0
   sec="runcontrol1";
   rch.read(sec,"ignore",ignore1);
   if (!ignore1){
      rch.read(sec,"readFromFile",readFromFile1);
      rch.read(sec,"dtStep",dtStep1);
      rch.read(sec,"nSteps",nSteps1);
      rch.read(sec,"filingPeriod",filingPeriod1);
      rch.read(sec,"docuPeriod",docuPeriod1);
      rch.read(sec,"viewingPeriod",viewingPeriod1);
      rch.read(sec,"stoppingPeriod",stoppingPeriod);
      rch.read(sec,"fixedStep",fixedStep1);
      rch.read(sec,"vsCrit",vsCrit,1);
      rch.read(sec,"asCrit",asCrit,1);
      rch.read(sec,"etas",etas,1);

      Z dvs=vsCrit.dim();
      cpmassert(dvs>1,loc);
      Z das=asCrit.dim();
      cpmassert(das>1,loc);
      Z detas=etas.dim();
      cpmassert(detas>1,loc);

      SysControl::fixedStep=fixedStep1;
      R t1=dtStep1*nSteps1;

      f1=F4<R,R,R2,R2,R2,Z>(t1,R2(vsCrit[1],vsCrit[2]),
         R2(asCrit[1],asCrit[2]), R2(etas[1],etas[2]))(fAux);
      // this makes SysControl::vCrit, SysControl::aCrit, and
      // SysControl::eta following a smooth sigmoid transition between
      // the initial and final values inputted here
   }

   Z nSteps2,filingPeriod2,docuPeriod2,viewingPeriod2;

   R dtStep2;
   bool storeGraphAsImage=false;
   bool albuming=false;
   Z nOrderSteps=0;
   Z facOrder=1;
   Z nSubsteps=0;
   Z facSubsteps=2;
   R tWaitSub=10;
   bool logView=true;
   V<R> ivOrder{0.75,1.};
   T2<Word> t2w=Sys2::docMode(rch());
   Word documode=t2w.c2();

   sec="runcontrol2";
   rch.read(sec,"ignore",ignore2);
   if (!ignore2){
      rch.read(sec,"readFromFile",readFromFile2);
      rch.read(sec,"dtStep",dtStep2);
      rch.read(sec,"nSteps",nSteps2);
      rch.read(sec,"filingPeriod",filingPeriod2);
      rch.read(sec,"docuPeriod",docuPeriod2);
      rch.read(sec,"viewingPeriod",viewingPeriod2);
      Z ped=1;
      rch.read(sec,"storeGraphAsImage",storeGraphAsImage,ped);
      rch.read(sec,"nOrderSteps",nOrderSteps,ped);
      rch.read(sec,"nSubsteps",nSubsteps,ped);
      rch.read(sec,"facSubsteps",facSubsteps,ped);
      rch.read(sec,"tWaitSub",tWaitSub,ped);
      rch.read(sec,"logView",logView,ped);
      rch.read(sec,"albuming",albuming,ped);
   }

   Timer tm(runTimeSpan*3600);
      // runTimeSpan is in hours, and Timer() expects seconds

   using namespace CpmAlgorithms;

   sec="cameracontrol";
   Camera cam(rch,sec,fr0);
   Sys2 sys;
   bool earlyStop=false;
   Record rc=rch();

   if (continue_run==2){
      ignore1=true;
      ignore2=false;
      readFromFile2=true;
   }
   else if (continue_run==1){
      ignore1=false;
      readFromFile1=true;
      // inFiles[1] is the file that will be continued
      // and the name will be given by outFiles[1] and runName
   }
   else{;}

   if (!ignore1){ // a Sys2-state will be either read from a movie-file
      // or constructed from the data provided by a ini-file.
      // It will be subjected to a preparatory run (indexed 1)
      if (readFromFile1){
         cpmmessage("branch (!ignore1) and (readFromFile1) taken");
         sys=Sys2(inFiles[1],rc);
      }
      else{
         cpmmessage("branch (!ignore1) and (!readFromFile1) taken");
         sys=Sys2(rch); // system state from inifile
      }
      if (!(continue_run==1)){
         cpmmessage("branch !(continue_run==1) taken");
         sys.set_t(0);
      }

      OFileStream ofs1(outFiles[3]);
      earlyStop=sys.run(dtStep1,nSteps1,filingPeriod1,docuPeriod1,
         viewingPeriod1,stoppingPeriod,outFiles[1],cam,tm,f1,ofs1,albuming\
         );
      if (docuPeriod1>0){
            // then docuStep was idle
         X2<Z,Word> data1=sys.docuStep(ofs1);
         R_Matrix rShow1(data1.c1());
         IFileStream ifs1(outFiles[3]);
         rShow1.scanFromFreeFor(ifs1());
         Frame fr1;
         fr1.paint(cPaLa);
         Graph gr1(fr1,rimPaLa);
         gr1.setGridColor(Color(FAINTGRAY));
         gr1.setColorActive(Color(WHITE));
         gr1.show(rShow1,rainBow,data1.c2(),0.25);
         cpmwait(tWaitWorkScreen_,2);
         fr1.paint(cPaLa);
         fr1.vis();
      }
      if (earlyStop){
         ignore1=true;
         ignore2=true;
      }
   } // !ignore1

   if (!ignore2){
      if (readFromFile2){
         cpmmessage("branch (!ignore2) and (readFromFile2) taken");
         if (continue_run==2){
            Record rc0;
            sys= Sys2(inFiles[2],rc0);
               // partition data taken from movie file since
               // rc0.isVoid()==true
         }
         else sys= Sys2(inFiles[2],rc);
            // partition data taken from active ini-file via rc
      }
      else if (ignore1) {
         cpmmessage("branch (!ignore2) and (ignore1) taken");
         sys= Sys2(rch);
            // Then no Sys2-state was created earlier.
            // Neither by reading from file nor as a result of a
            // preparatory run.
      }
      else{
         cpmmessage("branch (!ignore2) and (else) taken");
         ;  // use sys as created earlier
      }

// preparing the main run

      if (!(continue_run==2)){
         cpmmessage("branch !(continue_run==2) taken");
         sys.set_t(0);
         sys.setVel();
      }

      SysControl::read(rch,"syscontrol");
      // Notice that the second run is fully controlled by "syscontrol".
      // This allows flexible testing of the relative influence of
      // certain interactions. On the other hand, one has to make sure
      // that experimental values have to be correctly re-set for
      // 'production runs' of the program.
      OFileStream ofs2(outFiles[4]);
      if (nOrderSteps>1){ // order analysing run
         sec="runcontrol2";
         Z repLog=0;
         R facOrder=1.4142; // facOrder was Z till 2010-08-27
         R repOrderD=0,repOrderE=0,repOrderL=0;
         bool makeMovie=true;
         V<R> ivOrder{0.,1.};
         cpmrhf(facOrder);
         cpmrhf(ivOrder);
         cpmrhf(repOrderD);
         cpmrhf(repOrderE);
         cpmrhf(repOrderL);
         cpmrhf(repLog);
         cpmrhf(makeMovie);
         cpmrhf(storeGraphAsImage);
         Iv ivOrd(ivOrder[1],ivOrder[2]);
         R_Vector orders=sys.anlOrd(dtStep2,nSteps2,nOrderSteps,facOrder,
            getRunName(),cam,ivOrd,repLog,repOrderD,repOrderE,repOrderL,
            makeMovie,storeGraphAsImage);
            // last two arguments added 2010-08-28
      }
      else if (nSubsteps>0){ // newest addition 2014-02-09
         // analyses how increasing of the number of steplets in a step
         // changes the trajectory
         T2<R_Vector> rv=sys.anlExc(dtStep2,nSteps2,nSubsteps,
                                facSubsteps, logView);
         Graph gr;
         gr.setWriteOnShow(true);
         gr.show(rv.c1(),rv.c2());
         cpmwait(tWaitSub,2);
         return;
      }
      else{ // normal run
         stoppingPeriod=0;
         F<R,Z> f2(Z(0));
         earlyStop=sys.run(dtStep2,nSteps2,filingPeriod2,docuPeriod2,
         viewingPeriod2,stoppingPeriod,outFiles[2],cam,tm,
         f2,ofs2,albuming); // this does the main run.
            // it may create movie-files and image series.
         if (documode=="all"){ // here many data are collected which
            // are not suitable for a final screen display (which for
            // 'production runs' are irrelevant anyway).
            cpmmessage(loc&": early exit due to documode == all");
            CPM_MZ
            return;
         }
         if (docuPeriod2>0){
            // then docuStep was not idle and we have to show some
            // screen graphics
            X2<Z,Word> data2=sys.docuStep(ofs2);
         // here this adds the data for the final state which is OK
            Z dr2=data2.c1();
            Word title2=data2.c2();

            R_Matrix rShow2(dr2);

            IFileStream ifs2(outFiles[4]);
            bool succ=false;
            if (documode=="trajectoriesX" || documode=="trajectoriesY"
               || documode=="trajectoriesXY" ){
               succ=rShow2.scanLineWise(ifs2());
            }
            else{
               succ=rShow2.scanFromFreeFor(ifs2());
            }
            cpmassert(succ==true,loc);
            Frame fr2;
            fr2.paint(cPaLa);
            Graph gr2(fr2,rimPaLa);
            gr2.setGridColor(Color(FAINTGRAY));
            gr2.setColorActive(Color(WHITE));

            bool brown=sys.getRc().getB("done","Brown");

            if (brown){ // referring to Brownian motion
               gr2.setXY(rShow2[1],rShow2[2],1);
               gr2.draw(rShow2[1],rShow2[2],Color(BLACK));
               gr2.setText(title2);
               gr2.vis();
            }
            else{
               cpmmessage("graphics ahead");
               gr2.setAutoScale(1);
               if (documode=="trajectoriesXY"){
                  gr2.mark(rShow2,rainBow,1.05,0);
                  gr2.setText(title2);
                  gr2.vis();
               }
               else if (documode=="trajectoriesX"){
                  gr2.mark(rShow2,rainBow,1.05,1);
                  gr2.setText(title2);
                  gr2.vis();
               }
               else if (documode=="trajectoriesY"){
                  gr2.mark(rShow2,rainBow,1.05,2);
                  gr2.setText(title2);
                  gr2.vis();
               }
               else{
                  gr2.show(rShow2,rainBow,title2,0.25);
               }
            }

            if (storeGraphAsImage){
               Word in="img_"&getRunName();
               gr2.write(in);
            }
         }
         if (earlyStop){
            ignore1=true;
            ignore2=true;
         }
      }
   } // !ignore2
   cpmwait(tWaitWorkScreen_,2);
   CPM_MZ
}
