//? cpmsys2cpp.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


// This file defines the constructors of CpmTop2::Sys2
// and thus the main functionality also of Sys2Appl

#include <cpmunits.h>
#include <cpmroot.h>
#include <cpmperm.h>
#include <cpms.h>
#include <cpmfile.h>
#include <cpmmacros.h>
#include <cpminifilebasapp.h>
#include <cpmdistribution.h>

namespace CPM_TOP{

using CPM_RIGID_BODY::PointPar;
using CPM_RIGID_BODY::PointParC;
using CPM_RIGID_BODY::PointParV;

using CPM_RIGID_BODY::ComPar;
using CPM_RIGID_BODY::ComParC;
using CPM_RIGID_BODY::ComParV;

using CPM_RIGID_BODY::RigidObject;
using CPM_CAMERA::Camera;

using CPM_DIM::Spc;
using CPM_DIM::Vec;
using CPM_DIM::AxVec;
using CPM_DIM::Group;
using CPM_DIM::RefSys;
using CPM_DIM::SubsetOfSpace;
using CPM_DIM::dimVec;
using CPM_DIM::Rdim;

using CPM_PHYSICS::Force;
using CPM_PHYSICS::ForTor;
using CPM_PHYSICS::VectorField;
using CPM_PHYSICS::ExtSys;
using CPM_PHYSICS::VecE;
using CPM_PHYSICS::VecBH;
using CPM_PHYSICS::VecG;
using CPM_PHYSICS::FieldSpace;

using CPM_PHYSICS::VibData;
using CPM_PHYSICS::PotData;
using CPM_PHYSICS::LidData;
using CPM_PHYSICS::FenData;
using CPM_PHYSICS::SphTankData;
using CPM_PHYSICS::HomFilData;
using CPM_PHYSICS::DrumData;

using CPM_PHYSICS::Pot;
using CPM_PHYSICS::Lid;
using CPM_PHYSICS::Fen;
using CPM_PHYSICS::HomFil;
using CPM_PHYSICS::Drum;

using CpmDynSys::StepControl;
using CpmDynSys::ParDynSys;

using CpmProperties::NuggetData;

using CpmGeo::Angle;
using CpmGeo::DEG;

using CpmSystem::OFile;
using CpmArrays::S;
using CpmArrays::X2;
using CpmArrays::T2;
using CpmArrays::X4;
using CpmArrays::DisR;

using CpmMPI::Cpm_com;
using CpmMPI::Com;
using CpmFunctions::F;
using CpmFunctions::F1;
using CpmFunctions::F2;
using CpmFunctions::F3;
using CpmFunctions::F4;
using CpmFunctions::F5;
using CpmApplication::dimApp;

using namespace CpmRoot;
using namespace CpmRootX;
using namespace CpmLinAlg;
using namespace CpmGraphics;
using namespace CpmAlgorithms;
using namespace CpmImaging;
using namespace CpmProperties;
using namespace CpmPhysics;

//////////////////////////////// class Sys2 /////////////////////

namespace{

// remember: typedef ParOfSpc<Sys2Dat> Partition;
   // partition with a simple rectangular box geometry.
Partition makePartition(const Record rc)
{
   Z mL=3;
   Word loc("makePartition(Record)");
   cpmmessage(mL,loc&" started");

   Word sec="partition data";
   R xL,xU,yL,yU,zL,zU,overlap1,overlap2,overlap3;
   Z nL1,nL2,nL3,nS1,nS2,nS3;
   bool isTorus1,isTorus2,isTorus3;
   rc.get(sec,"xL",xL);
   rc.get(sec,"xU",xU);
   rc.get(sec,"yL",yL);
   rc.get(sec,"yU",yU);
   rc.get(sec,"zL",zL);
   rc.get(sec,"zU",zU);
   rc.get(sec,"overlap1",overlap1);
   rc.get(sec,"overlap2",overlap2);
   rc.get(sec,"overlap3",overlap3);
   rc.get(sec,"nL1",nL1);
   rc.get(sec,"nL2",nL2);
   rc.get(sec,"nL3",nL3);
   rc.get(sec,"nS1",nS1);
   rc.get(sec,"nS2",nS2);
   rc.get(sec,"nS3",nS3);
   rc.get(sec,"isTorus1",isTorus1);
   rc.get(sec,"isTorus2",isTorus2);
   rc.get(sec,"isTorus3",isTorus3);

   Sys2Dat stdat(xL,xU,yL,yU,zL,zU);
   Partition partition(stdat,nL1,nL2,nL3,overlap1,overlap2,overlap3,
      nS1,nS2,nS3);

   if (isTorus1){
      Vec shift;
      shift[1]=xU-xL;
      Group g1(shift);
      partition.setTorus(1,g1);
   }
   if (isTorus2){
      Vec shift;
      shift[2]=yU-yL;
      Group g2(shift);
      partition.setTorus(2,g2);

   }
   if (isTorus3){
      if (dimVec==3){
         Vec shift;
         shift[3]=zU-zL;
         Group g3(shift);
         partition.setTorus(3,g3);
      }
      else{
         partition.setTorus(3,Group());
      }
   }
   cpmmessage(mL,loc&" done");
   return partition;
}

} // namespace

Sys2::Sys2(V<ParT> const& par_, Vp<ExtSys> const& ext_,
   const Partition& part_, Record const& rc, bool careful):
   Sys2Base(par_,ext_,part_),rc_(rc)
{
   Z mL=1;
   Word loc("Sys2(V<ComPar>&,...,Partition&)");
   cpmmessage(mL,loc&" started");
   const R fracIgnore=1e-2;
   const R disTol=2e-6;
   setTimesFromExtSys();
   Z nTot=par.dim();
   Z i;
   if (careful){
      V<bool> vb(nTot);
      SubsetOfSpace sX=partition.getX();
      R d1,f1,f2;
      for (i=1;i<=nTot;i++){
         Spc xi=par[i].get_x();
         sX.shiftExternalPointToSurface4(xi,d1,f1,f2);
         if (d1>0) par[i].set_x(xi);
         vb[i]=(d1<disTol);
      }
      selectParticles(vb);
      Z diff=nTot-get_np();
      if (diff!=0){
         Z diffOK=cpmtoz(fracIgnore*nTot);
         cpmwarning(loc&": "&cpm(diff)&" particles lost");
         cpmassert(diff<=diffOK,loc);
      }
      nTot=get_np();
   }
   for (i=1;i<=nTot;i++){
      par[i].cancel_id();
      par[i].set_id(i);
      Z oi=partition.ownerSimple(par[i].get_x());
      par[i].setOwner(oi);
   }
   nTot_=Z1(nTot);
   cpmmessage(mL, loc&" done, nTot="&cpm(nTot));
}

Sys2::Sys2(const ExtrSys2& extr):
Sys2Base(extr.get_vp(),extr.get_ex(),extr.get_parti())
{
   Z mL=1;
   Word loc("Sys2(Extractor)");
   cpmmessage(mL,loc&" started");
   setTimesFromExtSys();
   cpmmessage(mL, loc&" done, nTot="&cpm(nTot_));
}

X4<V<ParT>,Vp<ExtSys>,Partition,Z>
Sys2::readFromMovie(const Word& movieFile, R frac)
{
   Z mL=1;
   Word loc("Sys2::readFromMovie("&movieFile&",frac)");
   CPM_MA
   if (frac<0 || frac>1) cpmerror(loc&" frac out of range");
   ExtrSys2 extr(movieFile);
   Partition parti=extr.get_parti();
   bool tolerant=true;
   if (frac<1){
      while(extr.next(tolerant)){;} // should not create warnings
         // for reading incomplete frames at the end
         // reads all files in order to see how much of them
         // are there.
      R countAimR=extr.getCount();
      countAimR*=frac;
      Z1 countAim(countAimR);
      ExtrSys2 extr2(movieFile);
      while(extr2.next() && extr2.getCount()<countAim){;}
      cpmmessage(mL,loc&" done");
      return X4<V<ParT>,Vp<ExtSys>,Partition,Z>
         (extr2.get_vp(),extr2.get_ex(),parti,extr2.getCount());
   }
   else{
      V<ParT> vpOK;
      Vp<ExtSys> exOK;
       Z countOK=0;
      while(extr.next(tolerant)){ // should not create warnings
         // for reading incomplete frames at the end
         vpOK=extr.get_vp();
         exOK=extr.get_ex();
         countOK=extr.getCount();
         cpmmessage(mL,"frame #"&cpm(countOK)&" read");
      }
      CPM_MZ
      return X4<V<ParT>,Vp<ExtSys>,Partition,Z>(vpOK,exOK,parti,countOK);
   }
}

Sys2::Sys2(const Word& fn, const Record& rc, R frac)
{
   Z mL=1;
   Word loc("Sys2::Sys2(Word,Record,R)");
   cpmmessage(mL,"Sys2("&fn&", Record,R) started");
   if (frac<0 || frac>1) cpmerror(loc&": frac out of range");
   Partition parti;
   bool careful=!rc.isVoid();
   Z count=0;
   sizeAndRank();
   if (rank==1){
      X4<V<ParT>,Vp<ExtSys>,Partition,Z> x4=
         readFromMovie(fn,frac);
      parti=x4.c3();
      count=x4.c4();
      if (careful) parti=makePartition(rc);
         // override parti by data from rc if these are not trivial
      *this=Sys2(x4.c1(),x4.c2(),parti,rc,careful);
      sizeAndRank();
         // needed again since the previous function call changed
         // size and rank to the values carried by
         // Sys2(x4.c1(),x4.c2(),parti,rc,careful);
      Z nParti=parti.card();
      if (size!=1 && nParti!=size){
         ostringstream ost;
         ost<<"Sys2(RecordHandler,Word): number of processes does"
            <<endl;
         ost<<" match the input data for nL1,nL2,nL3 from argument rc"
            <<endl;
         ost<<"size="<<size<<endl;
         ost<<"nParti="<<nParti<<endl;
         cpmerror(Word(ost.str()));
      }
      cpmmessage(mL,"object built successfully by process of rank 1");
   }
   distributeSys2(parti);
   cpmmessage(mL,"Sys2("&fn&", Record,R) done, "&cpm(count)&
      " frames found");
}

Sys2::Sys2(void):
Sys2Base(V<ParT>(0),Vp<ExtSys>(0)){}

Sys2::Sys2(V<Word> const& typeList):
Sys2Base(V<ParT>(0),Vp<ExtSys>(typeList.dim()))
{
   Z mL=1;
   static Word loc("Sys2(V<Word>)");
   CPM_MA
   for (Z i=1;i<=ne;++i){
      Word wi=typeList[i];
      if (wi=="Lid") ext[i]=Lid();
      else if (wi=="Pot") ext[i]=Pot();
      else if (wi=="Fen") ext[i]=Fen();
      else if (wi=="SphTank") ext[i]=SphTank();
      else if (wi=="HomFil") ext[i]=HomFil();
      else if (wi=="Drum") ext[i]=Drum();
      else cpmerror(loc&": unknown type "&wi&
         " of external system found");
   }
   CPM_MZ
}

namespace{ // auxiliar functions to be used in
   // Sys2::Sys2(const RecordHandler& rch)
Spc createFunc(Z const& i, Rdim const& xL, Rdim const& xU)
// random points in a rectangular box
{
   Z j=3*i;
   Z shift=1;
   R r1=Iv(xL[1],xU[1]).ran(j);
   j+=shift;
   R r2=Iv(xL[2],xU[2]).ran(j);
   R r3=0;
   if (dimVec==3){
      j+=shift;
      r3=Iv(xL[3],xU[3]).ran(j);
   }
   Rdim res(r1,r2,r3);
   return Spc(res);
}
// The following (obviously problematic) version gives a strict
// correlation between the rotation angle and the position. I did
// not expect that this would be possible. Since it is not understood
// I leave the code here

/*
ParT createFuncPar(Z const& i, Rdim const& xL, Rdim const& xU,
   V<ParT> const& vp, R const& alpha)
// random particles from list vp in a rectangular box
{
   Z j=i;
   Z shift=137;
   R r1=Iv(xL[1],xU[1]).ran(j);
   j+=shift;
   R r2=Iv(xL[2],xU[2]).ran(j);
   R r3=0;
   if (dimVec==3){
      j+=shift;
      r3=Iv(xL[3],xU[3]).ran(j);
   }
   Rdim rd(r1,r2,r3);
   Spc p(rd);
   ParT res=vp.cyc(i);
   res.ranRot(alpha,i);
   res.set_x(p);
   return res;
}
*/

ParT createFuncPar(Z const& i, Rdim const& xL, Rdim const& xU,
   V<ParT> const& vp, V<Spc> const& vs, R const& alpha)
// particles from list vp randomly placed in a rectangular box
// at random rotational orientation
{
   Z mL=3;
   Word loc("createFuncPar(Z,Rdim ,Rdim,V<ParT>,V<Spc>,R)");
   CPM_MA
   cpmassert(i>0,loc);
   Z vsd=vs.dim();
   Spc p;
   if (i<=vsd){ // use the positions from the list first
      p=vs[i];
   }
   else{
      Z j=(i-vsd); // j starts with 1
      j*=dimVec; // to use new j's for the next i
      R r1=Iv(xL[1],xU[1]).ran(j++);
      R r2=Iv(xL[2],xU[2]).ran(j++);
      R r3= dimVec==3 ? Iv(xL[3],xU[3]).ran(j++) : R(0);
      Rdim rd(r1,r2,r3);
      p=Spc(rd);
      if (cpmverbose>mL) cpmdebug(p);
   }
// for selecting the particle and its orientation, the mechanism
// for creation of p is irrelevant
   ParT res=vp.cyc(i);
   res.ranRot(alpha,i);  // setting orientation
   res.set_x(p); // setting location
   CPM_MZ
   return res;
}

bool cond1FuncPar(ParT const& par, SphTank const& st, Drum const&  b1,
   Drum const& b2, R const& d)
{
   if (st.isActive()){
      R g=st.gap(par);
      if (g<=d) return false;
   }
   if (b1.isActive()){
      R g=b1.gap(par);
      if (g<=d) return false;
   }
   if (b2.isActive()){
      R g=b2.gap(par);
      if (g<=d) return false;
   }
   return true;
}

bool cond2FuncPar(X2<ParT,ParT> const& pp, R const& d)
{
   ParT p1=pp.c1();
   ParT p2=pp.c2();
   R g=p1.gap(p2);
   return g<=d ? false : true;
}

bool cond2Func(X2<Spc,Spc> const& p, R const& d)
// yields 'true' if the distance of the points is larger
// than d
{
   return (p.first()-p.second()).abs()>d;
}

bool restrictFunc(Spc const& p, R const& r)
// returns true for an acceptable position
{
   Spc ori;
   return (p-ori).abs()>r;
}

void mc(ParT& p, bool isOsm=true) // make colorful
// unifying the rules for particle display colors
{
   if (Spc::dimension()==2){
      R q=p.get_q();
      if (q>0) p.setCol(BLUE);
      else if (q<0) p.setCol(RED);
      else{
         if (!isOsm) p.setCol(YELLOW);
      }
   }
   else p.setCol(INVISIBLE);
}
}// namespace

T2<Word> Sys2::docMode(Record const& rc)
{
   Z mL=3;
   Word loc("Sys2::docMode(Record)");
   CPM_MA
   Word sec="topic under consideration";
   Word topic,documode;
      // Although documode come from varius locations in the
      // ini-file, there is a single variable associeted
      // with it !
   rc.get(sec,"topic",topic);
   sec=topic&" data";
   rc.get(sec,"documode",documode);
   CPM_MZ
   return T2<Word>(topic,documode);
}

//////// the most important function ///////////////////////////////////

Sys2::Sys2(RecordHandler const& rch)
// this is the function for getting data from pala2.ini
// Notice that this function will be called with rch being the defining
// data source for the whole application Sys2Appl. Not all of these
// data get used here.
{
   const Z mL=1;
   Word loc("Sys2(RecordHandler)");
   CPM_MA
   Partition parti;
   sizeAndRank();
   cpmmessage(mL,"size="&cpm(size)&" rank="&cpm(rank));
   V<Z> mirList;
   Z i,j;
   Z ped=1;
   bool active=true; // a reading variable
   if (rank==1){
      cpmmessage(mL,"rank == 1 path entered");
#if defined(CPM_POINT_PARTICLES)
      cpmmessage("CPM_POINT_PARTICLES defined, using monospherical\
       particles");
#else
      cpmmessage("CPM_POINT_PARTICLES not defined, using polyspherical\
       particles");
#endif
      Record rc=rch(); // rch from argument
      // we use the name rc since the macros cpmr and cpmrf anticipate
      // usage of this name.
      cpmmessage(mL,"rc defined");
      // notice that the native Record-member is rc_
         // the effect of writing to rc_ here would be destroyed by
         // the '*this='-statement. So we write on rc for the present.
         // rc_ can here be filled with situation dependent data
         // that may be used via getSec() in any context. Notice also
         // that there is a prize to be payed for the flexibility of
         // this data container. The compiler will not warn if we
         // try to read some data from rc which we didn't write to
         // it earlier.
      T2<Word> t2w = docMode(rc);
      Word topic,documode;
      topic=t2w.c1();
      documode=t2w.c2();
      cpmmessage(mL, "topic = "&topic&", documode= "&documode);

      Word sec="runcontrol2";
      cpmmessage(mL,"reached sec = runcontrol2");

      R dtStep;
      Z nSteps;
      cpmr(dtStep);
      cpmr(nSteps);
         // needs not to be there
      R tTot=R(nSteps)*dtStep;
      parti=makePartition(rc);
         // assumes that there is a section 'partition data' to read from
      Sys2Dat pDat=parti.getDat();
      Rdim psL=pDat.corL();
      Rdim psU=pDat.corU();
      R hPot=psU[2]-psL[2];
      Iv iv1=pDat.iv1();
      Iv iv2=pDat.iv2();
      Iv iv3=pDat.iv3();
// the fill-in space describes a volume where the particles get
// placed initially. It is assumed that this region is a box (brick)
// which is contained in the box of the partition.
      sec="fill-in space";
      cpmmessage(mL,"reached sec = fill-in space");
      R xL,xU,yL,yU,zL,zU,frac;
      Z nTot;
      cpmr(xL);
      cpmr(xU);
      cpmr(yL);
      cpmr(yU);
      cpmr(zL);
      cpmr(zU);
      cpmr(nTot);
      cpmr(frac);
      cpmassert(xL>=iv1.inf(),loc);
      cpmassert(xU<=iv1.sup(),loc);
      cpmassert(yL>=iv2.inf(),loc);
      cpmassert(yU<=iv2.sup(),loc);
      cpmassert(zL>=iv3.inf(),loc);
      cpmassert(zU<=iv3.sup(),loc);

      Rdim fsL(xL,yL,zL);
      Rdim fsU(xU,yU,zU);
         // fs stands for fill-in space
         // the box should be a subset of the area for which the
         // partition is defined (let us refer to this rectangular box
         // as the 'master box'.
      Rdim svd=psU-psL;
      R rPot=svd.abs()*0.5;
         // radius of a bounding sphere of the 'master box'
      RigidObject ro(rPot);
      ExtSys es(ro);

      sec="homogeneous fields";
      cpmmessage(mL,"reached sec = homogeneous fields");
      V<R> vE,vB,vG;
      rc.get(sec,"E",vE);
      rc.get(sec,"B",vB);
      rc.get(sec,"G",vG);
    //  cpmdebug(vB);

      VecE fE(vE);
      VecBH fBH(vB,"B"); // understand Tesla units
      VecG fG(vG);

      HomFilData hfd(fE,fBH,fG); // homogeneous field data
   //   cpmdebug(hfd);
      HomFil hf(es,hfd);

      sec="pot data";
      cpmmessage(mL,"reached sec = pot data");
      BodyProperties bpPot(rch,sec);
      active=true;
      cpmr(active);

      Body pb(es,bpPot);
      PotData pd(svd);
      bool sideWalls=true;
      cpmrf(sideWalls); // needs not to be there
      pd.setSideWalls(sideWalls);
// new vibration data
      VibData vd(rc,"pot vibration data");
      cpmmessage(mL,"reached sec = pot vibration data");
      Pot pot(pb,pd,vd);
         // construction is based on svd and thus symmetrical to the
         // origin. This may not always be wanted; so lets shift it in a
         // position given by psU and psL and not only by their
         // difference.
      pot.setAct(active);
      Spc cU(psU);
      Spc cL(psL);
      Spc cC=cL.cen(cU); // center = midpoint
      Spc ori;
      Vec shPot=cC-ori;
      pot*=Group(shPot); // how nicely geometry works !!
      V<ParT> vc;
      Vp<ExtSys> ext;
      // these two objects have still to be defined in order to make the
      // 'final' statement meaningful: *this=Sys2(vc,ext,parti,rch);
      // If topic = general, we need a Pot (which has vertical walls
      // and a bottom plane), and an upper Lid, and no other external
      // systems. Homogeneous fields are part of the Pot.

      sec="lid data";
      cpmmessage(mL,"reached sec = lid data");
      BodyProperties bpLid(rch,sec);
      Body lb(es,bpLid);
      LidData ldU(iv1,iv3,iv2.sup());
      VibData lvd(rc,"lid vibration data");
      Lid lidU(lb,ldU,lvd);
      cpmrhf(active);
      lidU.setAct(active);

      sec="particle data";
      cpmmessage(mL,"reached sec = particle data");
         // data for a single particle type
      Z stochasticParticles=1;
      R rotAng1(0);
      bool randomizeOrientation=false;
      cpmr(rotAng1);
      cpmr(stochasticParticles);
      cpmrf(randomizeOrientation);
      Angle ang1(rotAng1,DEG);
      AxVec ax1(ang1);
      Group g1(ax1);
      cpmmessage(mL,"array vpp of "&cpm(stochasticParticles)&
      " stochastic particles ahead");
      V<ParT> vpp(stochasticParticles);
      for (i=1;i<=stochasticParticles;++i){
         vpp[i]=ParT(rch,sec,i);
         vpp[i]*=g1;
            // for each new i, a new random selection
      }
      cpmmessage(mL,"array vpp of stochastic particles done");
      Z dimSpc=Spc::dimension();
      ParT pp=vpp[1];
      // the'master particle'
      R rP=pp.get_r();
      R mP=pp.get_m();
      cpmdata<<endl<<"The radius of the master particle is "<<rP<<endl;
      cpmdata<<endl<<"The mass of the master particle is "<<mP<<endl;
     // if (dimSpc==3) pp.setCol(INVISIBLE);
     // else pp.setCol(YELLOW);
      R dMin=rP*frac*2;
      V<R> omega,rot;
      cpmrf(omega);
      cpmrf(rot);
      cpmmessage(mL,"reached second topic controlled loop, dMin ="
         &cpm(dMin));
      if (topic=="spherical_tank"){
         bool active=true;

         Rdim fsd=fsU-fsL;
         R rTank=fsd.abs()*0.5;
         R facMag=0;
         //ped=2;
         sec="spherical_tank data";
         cpmr(facMag);
         cpmr(active);
         sec="particle data";
         rTank*=facMag;
         cpmdata<<endl<<"The radius of the spherical tank is "
            <<rTank<<endl;
         SphTankData sphtankdata(rTank);
         SphTank sphtank(pb,sphtankdata);
         sphtank.setAct(active);
         ext=Vp<ExtSys>(2);
         ext[1]=hf;
         ext[2]=sphtank;
         bool unipolarCharge=false;
         cpmr(unipolarCharge);
         F<Spc,bool> c1(true);
         F<Z,Spc> cr=F2<Z,Rdim,Rdim,Spc>(fsL,fsU)(createFunc);
         F<X2<Spc,Spc>,bool> c2=F1<X2<Spc,Spc>,R,bool>(dMin)(cond2Func);
         Place<Spc> place(cr,c1,c2);
         V<Spc> xp=place(nTot);
         Z nAll=xp.dim();
         sec="special data for a few particles";
         Z ns=0;
         cpmr(ns);
         V<Spc> vpos(ns);
         V<AxVec> vrot(ns);
         V<AxVec> vomg(ns);
         V<bool> vfix(ns);
         for (i=1;i<=ns;i++){
            V<R> vi;
            rc.get(sec,Word("p")&cpmwrite(i),vi);
            Spc pi(vi);
            vpos[i]=pi;
            V<R> roti;
            rc.get(sec,Word("rot")&cpmwrite(i),roti);
            AxVec ri(roti);
            vrot[i]=ri;
            V<R> omgi;
            rc.get(sec,Word("omg")&cpmwrite(i),omgi);
            AxVec oi(omgi);
            vomg[i]=oi;
            bool bi;
            rc.get(sec,Word("fix")&cpmwrite(i),bi);
            vfix[i]=bi;
         }
         nAll+=ns;
         vc=V<ParT>(nAll);
         for (i=1;i<=ns;++i){
            vc[i]=vpp.cyc(i);
            vc[i].set_x(vpos[i]);
            vc[i]*=Group(vpos[i],vrot[i]);
            vc[i].spin(vomg[i]);
          //  cpmdebug(vc[i].get_v());
            vc[i].lock_(vfix[i]);
         }
         for (i=ns+1,j=1;i<=nAll;i++,j++){
            vc[i]=vpp.cyc(i);
            vc[i].set_x(xp[j]);
            if (!unipolarCharge){
               if (i%2){
                  vc[i].invertCharge();
                  vc[i].removeMagDipole();
               }
            }
            mc(vc[i]);
         }
      }
      else if (topic=="mill"){
         // most functional and mature part, needs not to deal with
         // mills only. We have two drums, a hollow one and a massive
         // one. We also may have a spherical container. The containers
         // may be set to a passive state, so that they don't add to the
         // computational burden.
         cpmmessage(mL,"mill-loop entered");
         sec="mill data";
         R minimumGap=SysControl::singRange;
         R rotAngleRange=0;
         rc.get(sec,"minimumGap",minimumGap,ped);
         rc.get(sec,"rotAngleRange",rotAngleRange,ped);

         SphTank sphtank;
         Drum drum1,drum2;

         sec="spherical_tank data";
         R rRel=1;
         active=false;
         cpmrhf(rRel);
         cpmrhf(active);
         R rTank = rPot*rRel;
         rTank*=rRel;
         SphTankData sphtankdata(rTank);
         cpmdebug(rTank);
         sphtank=SphTank(pb,sphtankdata);
         sphtank.setAct(active);

         sec="drum1 data";
         active=true;
         ped=1;
         cpmrhf(active);
         DrumData drumData1(rch,sec,ped);
         VibData dvd1(rc,"drum1 vibration data");
         drum1=Drum(pb,drumData1,dvd1);
            // notice that the material data come from pb and thus
            // from section "pot data"
         drum1.setAct(active);
         sec="drum1 shift";
         Vec shift1;
         rc.get(sec,"shift1",shift1[1]);
         rc.get(sec,"shift2",shift1[2]);
         Group g1(shift1);
         drum1*=g1;

         sec="drum2 data";
         active=true;
         cpmrhf(active);
         DrumData drumData2(rch,sec,ped);
         VibData dvd2(rc,"drum2 vibration data");
         drum2=Drum(pb,drumData2,dvd2);
         drum2.setAct(active);

         sec="drum2 shift";
         Vec shift2;
         rc.get(sec,"shift1",shift2[1]);
         rc.get(sec,"shift2",shift2[2]);
         Group g2(shift2);
         drum2*=g2;

         Z nExt=1; // The homogeneous field is always there
         // We add only the active external systems.
         if (sphtank.isActive()) nExt++;
         if (drum1.isActive()) nExt++;
         if (drum2.isActive()) nExt++;
         if (pot.isActive()) nExt++;
         if (lidU.isActive()) nExt++;

         ext=Vp<ExtSys>(nExt);

         Z iAct=0;
         ext[++iAct]=hf;
         if (sphtank.isActive()) ext[++iAct]=sphtank;
         if (drum1.isActive()) ext[++iAct]=drum1;
         if (drum2.isActive()) ext[++iAct]=drum2;
         if (pot.isActive()) ext[++iAct]=pot;
         if (lidU.isActive()) ext[++iAct]=lidU;

         for (i=1;i<=nExt;++i){
            cpmdebug(ext(i).nameOf());
         }

         sec="particle position list";
         Z npl=0;
         rc.get(sec,"n",npl,ped);
         V<Spc> vs(npl);
         Word wpl("p");
         for (i=1;i<=npl;i++){
            V<R> vi;
            rc.get(sec,wpl&cpmwrite(i),vi,ped);
            Spc pi(vi);
            vs[i]=pi;
         }
         bool addMirrorImages=false;
         rc.get(sec,"addMirrorImages",addMirrorImages,ped);
         F<Z,ParT> cr=F5<Z,Rdim,Rdim,V<ParT>,V<Spc>,R,ParT>
            (fsL,fsU,vpp,vs,rotAngleRange)(createFuncPar);
               // uses the particle position list
         F<ParT,bool> c1=F4<ParT,SphTank,Drum,Drum,R,bool>
            (sphtank,drum1,drum2,minimumGap)(cond1FuncPar);
         F<X2<ParT,ParT>,bool> c2=F1<X2<ParT,ParT>,R,bool>
            (minimumGap)(cond2FuncPar);

         Place<ParT> place(cr,c1,c2);
         cpmmessage(mL,"placing of "&cpm(nTot)&" particles ahead");
         vc=place(nTot,100000);
         cpmmessage(mL,"placing of "&cpm(nTot)&" particles done");
         Z nAll=vc.dim();
         if (SysControl::coloredParticles){
            for (i=1;i<=nAll;i++) vc[i].setCol(RED);
         }
         RefSys master;
         Spc om=master.o();
         Vec vm=master.e1();
         if (addMirrorImages){
            mirList = V<Z>(nAll);
            V<ParT> vc2=vc;
            for (i=1;i<=nAll;i++){
               vc2[i].ref_(vm,om);
               if (SysControl::coloredParticles){
                  vc2[i].setCol(GREEN);
               }
               mirList[i]=i+nAll;
             }
             vc&=vc2;
             nAll=vc.dim();
         }
         cpmmessage(mL,"mill-loop left");
      }
      else if (topic=="general"){
         sec="general data";
         ped=1;
            // ped='pedantry'
            // we read data that are not present in older ini-files
         R fracVol=1;
         rc.get(sec,"fracVol",fracVol,ped);
         R dh=hPot*(fracVol-1);
         R vh=dh/tTot;
       // setting the velocity of the lid
         Vec vLid(0,vh,0);
         lidU.boost(vLid);

         ext=Vp<ExtSys>(3);
         ext[1]=hf;
         ext[2]=pot; // pot, by design, has a bottom -
            // however not as a separable entity; i.e. forces
            // on the side walls and on the bottom can not be
            // registered separately.
         ext[3]=lidU;
            // closing upper lid

         bool unipolarCharge=false;
         rc.get(sec,"unipolarCharge",unipolarCharge,ped);
         R xBrown=1;
         rc.get(sec,"xBrown",xBrown,ped);
         F<Spc,bool> c1(true);
         if (xBrown>1){ // then overlap may result, and we take counter-
            // measures.
            const R facFree=2.0;
            R rRestrict=facFree*rP*xBrown;
            c1=F1<Spc,R,bool>(rRestrict)(restrictFunc);
         }
         F<Z,Spc> cr=F2<Z,Rdim,Rdim,Spc>(fsL,fsU)(createFunc);

         F<X2<Spc,Spc>,bool> c2=F1<X2<Spc,Spc>,R,bool>(dMin)(cond2Func);
         Place<Spc> place(cr,c1,c2);
         V<Spc> xp=place(nTot);
         Z nTot1=xp.dim();
         Z nAll=(xBrown>1 ? nTot1+1 : nTot1);
         vc=V<ParT>(nAll);
         for (i=1;i<=nTot1;i++){
            vc[i]=vpp.cyc(i);
            vc[i].set_x(xp[i]);
            if (!unipolarCharge){
               if (i%2){
                  vc[i].invertCharge();
                  vc[i].removeMagDipole();
               }
            }
            mc(vc[i],false); // giving vc[i] a color
            if (randomizeOrientation) vc[i]*=g1.ran(vc[i].get_x(),i);
         }
         if (xBrown>1){
            rc.setB("done","Brown",true);
            Spc ori;
            vc[nAll]=vpp[1];
            vc[nAll].set_x(ori);
            vc[nAll]*=xBrown;
            vc[nAll].setCol(LIGHTBLUE);
         }
      }
      else if (topic=="osmosis"){
         sec="osmosis data";
         R expPrsVel=1.5;
            // to let both parts of the osmosic cell with
            // equal pressure one has to give the smaller particles
            // a larger velocity. We assume also in 2D the mass to scale
            // with radius^3 (constant density, 3-dimensional particle
            // geometry). We assume a law pressure ~ v^expPrsVel.
            // For ideal gases one has expPrsVel=2, for the gases where
            // the particle volume fills a substantial part of the space
            // expPrsVel=1.5 is a better value.
            // This is a floppy statement. Influence of particle volume on
            // pressure, volume, velocity(temperature) relation known from
            // the
            // van der Waals gas equation. (UM 2012-02-27)
         Z sof=1;
         rc.get(sec,"expPrsVel",expPrsVel,sof);
         sec="particle data";
         R facLarge;
         rc.get(sec,"facLarge",facLarge);
         ParT pp1=pp;
         pp1*=facLarge;
         pp1.invertCharge();
         if (dimSpc==2) pp1.setCol(LIGHTBLUE);
         // the larger particles is obtained by scaling
         if (dimSpc==2) pp.setCol(YELLOW);

         Z nPar=2;
         V<ParT> vcPrel(nPar);
         vcPrel[1]=pp1;
         vcPrel[2]=pp;

         R_Vector vcr(nPar);
         for (i=1;i<=nPar;++i) vcr[i]=vcPrel[i].get_r();
         R3 r3=vcr.infMeanSup();
         R rMin=r3[1];
         R rMean=r3[2];
         R rMax=r3[3];
         R fac=rMin/rMax;
         R exVel=R(3)/expPrsVel;
         R facVel=cpmpow(fac,exVel);
         sec="data";
         rc.setR(sec,"rMin",rMin);
         rc.setR(sec,"rMean",rMean);
         rc.setR(sec,"rMax",rMax);
         rc.setR(sec,"facVel",facVel);
         dMin=rMax*frac*2;
         // This is the distance between particles that is not allowed
         // to surpass.
         // Setting frac<1 allows to create configurations where
         // overlapping particles are allowed.
         // Now we design the fill-in box so that it fits the upper half
         // of the whole space automatically. We hold the same distance
         // to the semi-permeable membrane that the input data set for the
         // distance to the upper lid.
         R gapToLid=psU[2]-fsU[2];
         R hph=(psU[2]-psL[2])*0.5;
         fsL[2]=psU[2]-hph+gapToLid;
         F<Z,Spc> cr=F2<Z,Rdim,Rdim,Spc>(fsL,fsU)(createFunc);
         F<Spc,bool> c1(true);
         F<X2<Spc,Spc>,bool> c2=F1<X2<Spc,Spc>,R,bool>(dMin)(cond2Func);
         Place<Spc> place(cr,c1,c2);
         V<Spc> xp=place(nTot);
         Z nTot1=xp.dim();
         // there is no guarantee that nTot positions can be found
         // by this method
         Z nTot2=2*nTot1;
         vc=V<ParT>(nTot2);
         Vec shift(0,-hph);
         for (i=1;i<=nTot1;i++){
            vc[i]=vcPrel[1];
            vc[i].set_x(xp[i]);
            vc[i+nTot1]=vcPrel[2];
            vc[i+nTot1].set_x(xp[i]+shift);
         }
         for (i=1;i<=nTot2;++i) mc(vc[i]);
         sec="osmosis data";
         V<bool> lidFence;
         rc.get(sec,"lidFence",lidFence);
         Z nMembranes=0;
         Z dlf=lidFence.dim();
         cpmassert(dlf==2,loc);
         for (i=1;i<=dlf;++i) if (lidFence[i]) nMembranes++;
         Z nExt=3+nMembranes;
         ext=Vp<ExtSys>(nExt);
         Z iAct=0;
         ext[++iAct]=hf;
         ext[++iAct]=pot;
         ext[++iAct]=lidU;
         if (lidFence[1]){
            Lid lidM=lidU;
            lidM.setPinhole(rMean);
            lidM.makeBottom();
            Vec shift;
            shift[2]=-iv2.abs()*0.5;
            lidM*=Group(shift);
            ext[++iAct]=lidM;
         }
         if (lidFence[2]){
            sec="fence data";
            ContactProperties cpFen(rch,sec);
            BodyProperties bpFen(cpFen);
            Body fb(es,bpFen);
            R facPicket,relWidth;
            rc.get("fence data","facPicket",facPicket);
            rc.get("fence data","relWidth",relWidth);

            R rPicket=rMin*(R(1)-facPicket)+rMax*facPicket;
            R gap=relWidth*rMin*2;
            FenData fd(iv1,iv3,rPicket,gap);
          // the smallest particle should pass the fence
            Fen fen(fb,fd);
            ext[++iAct]=fen;
         }
      }
      *this=Sys2(vc,ext,parti,rc);
      //cpmdebug(vc);
      cpmmessage(mL,"*this = ... executed");
      setPer_(mirList);
      Z nParti=parti.card();
      sizeAndRank(); // called already but needed here
      if (size!=1 && nParti!=size){
         ostringstream ost;
         ost<<
            "Sys2(RecordHandler): number of processes does not"<<endl;
         ost<<
            " match the input data for nL1,nL2,nL3 in station.ini"<<endl;
         ost<<"size="<<size<<endl;
         ost<<"nParti="<<nParti<<endl;
         cpmerror(Word(ost.str()));
      }
   }
   distributeSys2(parti);
      // notice that by definition of setPartition(...), the
      // function argument gets used only for rank==1. For
      // syntactic reasons it has to be there for each rank.

   CPM_MZ
}
//////// end of the most important function ////////////////////////////
////////////////////////////////////////////////////////////////////////

void Sys2::mark(const Camera& ca, bool display, const Word& text,
                bool fileOnShow)const
// Shows the whole system even in parallel mode, when the particles
// get distributed over the processes.
// It seems to be important that all ranks have the same system to
// show since it seems to depend on casualities of program run which
// rank does the job of bringing the data to screen.
{
   const Z mL=3;
   Word loc("Sys2::mark(Camera,bool,Word)");
   CPM_MA
   sizeAndRank();
   if (size==1){
      ca.clear();
      Sys2Base::mark(ca,display,text,fileOnShow);
      CPM_MZ
      return;
   }
   V<ParT> par_p=par;
   cpmmessage(mL,"nTot_="&cpm(nTot_)&" np="&cpm(np));
   cpmassert(np<=nTot_,loc);
   V< V<ParT> > parAll(size, V<ParT>(0));
   Cpm_com.gather(par_p,parAll);
   V<ParT> vp(nTot_);
   if (rank==1){
      Z i,j;
      for (i=1;i<=size;i++){
         Z n=parAll[i].dim();
         for (j=1;j<=n;j++){
            ParT val=parAll[i][j];
            Z k=val.get_id();
            vp[k]=val;
         }
      }
   }
   Cpm_com.bcast(vp);
   Sys2 sysTot(vp,getExt(),getPartition(),getRc());
   ca.clear();
   sysTot.Sys2Base::mark(ca,display,text);
   CPM_MZ
}

X2<R,R> Sys2::totChr()const
{
   R qMinus=0, qPlus=0;
   Z np=get_np();
   for (Z i=1;i<=np;i++){
      R q=getPar(i).get_q();
      if (q<0) qMinus+=q;
      if (q>0) qPlus+=q;
   }
   return X2<R,R>(qMinus,qPlus);
}

namespace{
   Word frId("// frame number =");
}

bool Sys2::file(const Word& fn, const Timer& tm)const
// Even in parallel mode the whole system gets written to file.
// The whole system is put together from the parts seen by the
// individual processes via message passing.
// A simplified form of this 'putting together' is being used
// in function mark(...)
{
   Z mL=1;
   Word loc("Sys2::file("&fn&"),Timer)");
   CPM_MA
   sizeAndRank();
   if (size==1){
      file1(fn);
      cpmmessage(mL,loc&": done for size==1");
      return tm.timeOut();
   }
   static Word fnActive("this_is_not_a_filename");
      // it is assumed that fn is a reasonable file name and thus
      // different from this strange initialization value for
      // fnActive
   static OFile ofi;
   static Z count;
   Z i,j;
   if (fn!=fnActive){
      // mechanism for writing some data only once
      fnActive=fn;
      count=1;
      if (rank==1){
         ofi=OFile(fn);
// partition
         partition.prnOn(ofi());
// exTypes
         V<Word> exTypes(ne);
         for (i=1;i<=ne;++i){
            Word wi=getExt()(i).nameOf();
            exTypes[i]=wi;
         }
         exTypes.prnOn(ofi());
      }
      // now we need to gather from all ranks
      V<ParCT> vcc(np);
      if (cpmverbose>=mL)
         cpmmessage("nTot_="&cpm(nTot_)&
         " np="&cpm(np));
      cpmassert(np<=nTot_,loc);
      for (i=1;i<=np;i++) vcc[i]=par[i].getC();
      V< V<ParCT> > vccAll(size, V<ParCT>(0));
      Cpm_com.gather(vcc,vccAll);
      if (rank==1){
         V<ParCT> vc(nTot_);
         for (i=1;i<=size;i++){
            Z n=vccAll[i].dim();
            for (j=1;j<=n;j++){
               ParCT val=vccAll[i][j];
               Z k=val.get_id();
               vc[k]=val;
            }
         }
         for (i=1;i<=nTot_;i++){
            if (!vc[i].get_idValid())
               cpmwarning("vc["&cpm(i)&"] not initialized");
         }
         vc.prnOn(ofi());
      }
   }
// this gets done for each frame
   typedef X3<B,Z1,ParVT> xcz;
   V<xcz> vcv(np);
   for (i=1;i<=np;i++){
      Z idi=par[i].get_id();
      Spc xi=par[i].get_x();
      B valid(par[i].getOwner()==rank);
      vcv[i].c1()=valid;
      if (!valid) continue;
      vcv[i].c2()=idi;
      ParVT cvi=par[i].getV();
      cpmassert(cvi.get_x()!=Spc(),loc);
      vcv[i].c3()=cvi;
   }
   V< V<xcz> > vcvAll(size,V<xcz>(0));
   Cpm_com.gather(vcv,vcvAll);
   B earlyStop; // will be set by rank 1 and
      // broadcasted to all others ranks
   if (rank==1){
      V<B> testArray(nTot_);
      V<ParVT> vv(nTot_);
      for (i=1;i<=size;i++){
         Z ni=vcvAll[i].dim();
         for (j=1;j<=ni;j++){
            B valid=vcvAll[i][j].c1();
            if (!valid) continue;
            Z k=vcvAll[i][j].c2();
            ParVT val=vcvAll[i][j].c3();
            vv[k]=val;
            testArray[k]=true;
            cpmassert(val.get_x()!=Spc(),loc);
         }
      }
      for (i=1;i<=nTot_;i++){
         bool bi=testArray[i];
         if (!bi){
            ostringstream ost;
            ost<<endl<<"particle "<<i<<" could not be found";
            cpmwarning(Word(ost.str()));
         }
      }
      Word frNum=frId&cpm(count);
      count++;
      frNum.prnOn(ofi());
      B vvOK=ParT::isVal(vv);
      vv.prnOn(ofi());
      for (i=1;i<=ne;++i) getExt()(i).prnOn(ofi());
      earlyStop=B(tm.timeOut()).vel(vvOK.non());
   }
   Cpm_com.bcast(earlyStop);
   cpmmessage(mL,"Sys2::file("&fn&") done");

   return earlyStop;
}

void Sys2::file1(const Word& fn)const
// creating a file and writing to it sich that
// the extractor class in cpmextractorx.h can read it
// Notice the new order partition, exTypes, vc (all once)
// vv, ex (one per frame)
// This is the simple method to be used for size==1
// in function file
{
   Z mL=2;
   Word loc("Sys2::file1("&fn&")");
   CPM_MA
   static Word fnActive("this_is_not_a_filename");
      // it is assumed that fn!="this_is_not_a_filename"
   static OFile ofi;
   static Z count;
   Z i,id,oi;
   if (fn!=fnActive){ // do once
      ofi=OFile(fn);
      fnActive=fn;
      count=1;
// partition
      partition.prnOn(ofi());
// exTypes
      V<Word> exTypes(ne);
      for (i=1;i<=ne;++i){
         Word wi=getExt()(i).nameOf();
         exTypes[i]=wi;
      }
      exTypes.prnOn(ofi());
// vc
      V<ParCT> vc(nTot_);
      for (i=1;i<=np;i++){
         id=par[i].get_id();
         oi=par[i].getOwner();
         if (oi>0) vc[id]=par[i].getC();
      }
      cpmmessage(mL,"nTot_="&cpm(nTot_)&" np="&cpm(np));
      vc.prnOn(ofi());
   }
// do for each frame
   V<ParVT> vv(nTot_);
   for (i=1;i<=np;i++){
      id=par[i].get_id();
      oi=par[i].getOwner();
      if (oi>0) vv[id]=par[i].getV();
   }
   Word frNum=frId&cpm(count);
      // frId starts with '//', thus irrelevant for the
      // Extractor
   count++;
   frNum.prnOn(ofi());
   B vvOK=ParT::isVal(vv);
// print vv
   vv.prnOn(ofi());
// print ex
   for (i=1;i<=ne;++i) getExt()(i).prnOn(ofi());
   CPM_MZ
   cpmassert(vvOK,loc);
}

void Sys2::distributeSys2(const Partition& parti)
{
   const Z mL=1;
   Word loc("Sys2::distributeSys2(...)");
   CPM_MA
   sizeAndRank();
   if (size==1){
      cpmmessage(mL,loc&" trivially done, size==1");
      CPM_MZ
      return;
   }
   cpmmessage(mL,"size="&cpm(size)&" rank="&cpm(rank));
   V<Word> typeList;
   Z i;
   if (rank==1){
      typeList=V<Word>(ne);
      for (Z i=1;i<=ne;++i){
         Word ti=ext(i).nameOf(); // having [i] instead
            // of (i) would give the name of the common base class
         typeList[i]=ti;
         cpmmessage(loc&": type is "&ti);
      }
   }
   Cpm_com.bcast(typeList);
   if (rank==1){
      Z j,paca=parti.card();
      if (paca!=size){
         Word mes("Sys2::distributeSys2(const Partition&): ");
         mes&="partition does not match the number of processes";
         cpmwarning(mes);
         cpmmessage("partition card() found as"&cpm(paca));
      }
      partition=parti; // now *this has the correct partition

      cpmassert(np==nTot_,loc);
         // in the present state of initialization par is the array of all
         //
         // particles of the whole system (all processes together)
      V< V<bool> > vvb(size,V<bool>(np,false));
      for (j=1;j<=np;j++){
         Z oj=par[j].getOwner();
         vvb[oj][j]=true;
      }
      for (i=2;i<=size;i++){
         cpmmessage("sending Sys2 to process "&cpm(i)&
            " in preparation");
         Z tagi=Com::tagFromToStatic(1,i,size);
         Sys2 sta(getPar(),getExt(),getPartition(),getRc());
            // both the statement
            //    Sys2 sta(*this);
            // and the experimental replacement
            //    Sys2 sta=*this;
            // make the following statement change *this if
            // #define CPM_USECOUNT is set and g++ is used for
            // compilation. The use-count mechanism seems to get
            // stressed too much if this function is called
            // within a constructor (as the present function will
            // be used exclusively) where *this is to be considered
            // incomplete. So to be defensive we call here only
            // functions that are members of the base class.
         sta.selectParticles(vvb[i]);
            // this does not mean that nTot_ is changed to
            // the number of selected particles

         Z ne=sta.getExt().dim();
         ostringstream ost;
         for (Z ii=1;ii<=ne;++ii){
            ost<<sta.getExt()(ii).nameOf()<<" ";
         }
         cpmmessage(Word(ost.str()));

         cpmmessage(mL,"preparation done, sending started");
         sta.send(i,tagi);
            // sending to i the station with particles that are right
            // for i, special sending implementation for extra large
            // data structures
         cpmmessage("sending Sys2 to process "&cpm(i)&" done; "&
            cpm(sta.get_np())&" particles sent");
      }
      selectParticles(vvb[1]);
         // adjusting the particles of rank 1
      cpmmessage("rank 1 has "&cpm(get_np())&" particles");
   }
   else{
      // receiving from rank 1 what was sent to rank
      Sys2 staRec(typeList);
      staRec.rec(1,Com::tagFromToStatic(1,rank,size));
      *this=staRec;
         // Notice that despite the internal complexity of
         // instances of Sys2 this initialization of an instance
         // (even within a constructor) from sent data follows
         // a very simple syntax
      cpmmessage("we got "&cpm(get_np())&" particles");
   }
   B masterReady;
   if (rank==1) masterReady=true;
   Cpm_com.bcast(masterReady);
   while (!masterReady){;}
   CPM_MZ
}

bool Sys2::prnOn(ostream& str)const
{
   Z mL=2;
   Word loc("Sys2::prnOn(ostream)");
   CPM_MA
   cpmwt("Sys2");
   if(!Sys2Base::prnOn(str)){
      cpmwarning(loc&": printing of base failed");
      CPM_MZ
      return false;
   }
   cpmp(nTot_);
   cpmp(rc_);
   CPM_MZ
   return true;
}

bool Sys2::scanFrom(istream& str)
{
   Z mL=2;
   Word loc("Sys2::scanFrom(istream)");
   CPM_MA
   if (!Sys2Base::scanFrom(str)){
      cpmwarning(loc&": scanning of base failed");
      CPM_MZ
      return false;
   }
   cpms(nTot_);
   cpms(rc_);
   CPM_MZ
   return true;
}

namespace{

   R logS(R x){ return x<1e-12 ? -R(12) : cpmlog10(x);}
}

R Sys2::asym(bool log)const
{
   Z np=per_.dim();
   if (np==0) return 1.;

   RefSys master;
   Spc o=master.o();
   Vec mv=master.e1();
   R res=0;
   for (Z i=1;i<=np;++i){
      ParT pi=par[i];
      ParT pj=par[per_[i]];
      pj.ref_(-mv,o);
      R dij=pi.dis(pj);
      res+=dij;
   }
   if (log){
      if (res!=0) res = cpmlog10(res);
   }
   return res;
}

X2<Z,Word> Sys2::docuStep(OFileStream& ofs)const
{
   Word loc("Sys2::docuStep(OFileStream& ofs)");
   Z mL=3;
   static R eTotIni=0;
   static R eTotIniInv=0;
   CPM_MA
   Word title("idle");
   Z i,n=0,np=nTot_,ne=get_ne();
   T2<Word> gdm=getDocMode();
   Word topic=gdm.c1();
   Word documode=gdm.c2();
   if (topic=="osmosis" && documode=="normal"){
      R1 tx(get_t());
      Z sumsu=0;
      Z sumll=0;
      for (i=1;i<=np;++i){
         Z v1=parType(i);
         Z v2=parLoc(i);
         if (v1==1){
            if (v2==2) sumsu++;
         }
         else{ // v1=2
            if (v2==1) sumll++;
         }
      }
      tx.prnOn(ofs());
      R1(sumsu).prnOn(ofs());
      R1(sumll).prnOn(ofs());
      n=3;
      title=
         "x: time, y: # of small particles in upper box (black), # of";
      title&=" large part. in lower box (violet)";
   }
   else if (topic=="mill" && documode=="symmetry"){
      R tx=get_t();
      R r1=asym();
      ofs()<<tx<<"  "<<r1<<endl; // two numbers in a line for simple
         // gnuplot processing of the file
      n=2;
      title="x: time, y: asymmetry measure";
   }
   else if (documode=="force"){
      R1 tx(get_t());
      Z yDir=2;
      R mf1=getExt()(2).getForce()[yDir];
         // pot which consists of a bottom plane and side walls
         // This is the force which the pot feels from the particles
      R mf2=getExt()(3).getForce()[yDir];
         // upper lid.
         // This is the force which the lid feels fromthe particles
      Spc ori;
      FieldSpace fExt=getExt()(1).getField(ori);
         // the field is homogeneous and may be evaluated at any point
      R accG=(fExt.c3())[yDir];
      R fg=get_m()*accG;
      // This is the force which the particles feel from the field
      tx.prnOn(ofs());
      R1(mf1).prnOn(ofs());
      R1(mf2).prnOn(ofs());
      R1(fg).prnOn(ofs());
      n=4;
      title="x: time, y: force on bottom (black), ceiling (violet)";
      title&=" weight of particles (blue)";
   }
   else if (documode=="mean_force"){
      R1 tx(get_t());
      Z yDir=2;
      R mf1=-getExt()(2).getMeanForce()[yDir];
      R mf2=getExt()(3).getMeanForce()[yDir];
         // here the intention is to easily see whether the pressure
         // on the lid and the bottom are the same
      tx.prnOn(ofs());
      R1(mf1).prnOn(ofs());
      R1(mf2).prnOn(ofs());
      n=3;
      title="x: time, y: mean force on bottom (black), ceiling (violet)";
   }
   else if (documode=="energy"){
      R tx=get_t();
      R r1=kinEnr();
      FieldSpace fs=getExt()(1).getField(Spc());
      R r2=potEnr()+potElasticExt()+potHomFil(fs);
      R r3=r1+r2;
      R1(tx).prnOn(ofs());
      R1(r1).prnOn(ofs());
      R1(r2).prnOn(ofs());
      R1(r3).prnOn(ofs());
      n=4;
      title="x: time, y: E_kin (black), E_pot (violet), E (blue)";
   }
   else if (documode=="total_energy"){
      R tx=get_t();
      FieldSpace fs=getExt()(1).getField(Spc());
      R r3=kinEnr()+potEnr()+potElasticExt()+potHomFil(fs);
      if (!running_){
         eTotIni=r3;
         eTotIniInv=cpminv(cpmabs(r3));
         running_=B(true);
      }
      else{
         R r4=(r3-eTotIni)*eTotIniInv;
         cpmvalue("dEtotRel",r4,4);
         ofs()<<tx<<"  "<<r4<<endl; // two numbers in a line for simple
         // gnuplot processing of the file
      }
      n=2;
      title="x: time, y: relative change of total energy";
   }
   else if (documode=="momentum"){
      R tx=get_t();
      R r1=0;
      Z com=2;
      for (Z i=1;i<=np;++i){
         ParT pi=par[i];
         r1+=pi.get_m()*pi.get_v()[com];
      }
      R r2=getExt()(1).get_momentum()[com];
      R r3=getExt()(2).get_momentum()[com];
      R r4=r1+r2+r3;
      R1(tx).prnOn(ofs());
      R1(r1).prnOn(ofs());
      R1(r2).prnOn(ofs());
      R1(r3).prnOn(ofs());
      R1(r4).prnOn(ofs());
      n=5;
      title="x: time, y: vertical component of momentum of";
      title&=" particles(black), pot(violet), lid(blue), all(cyan)";
   }
   else if (documode=="angular_momentum"){
      R tx=get_t();
      Z com=3;
      R r1=getExt()(2).get_angularMomentum()[com];
      R1(tx).prnOn(ofs());
      R1(r1).prnOn(ofs());
      n=2;
      title="x: time, y: 3-component of angular momentum(violet)";
   }
   else if (documode=="torque"){
      R tx=get_t();
      Z com=3;
      R r2=getExt()(2).getTorque()[com];
      R r3=getExt()(3).getTorque()[com];
      R1(tx).prnOn(ofs());
      R1(r2).prnOn(ofs());
      R1(r3).prnOn(ofs());
      n=3;
      title=
         "x: time, y: 3-component of torque 2(black), torque 3(violet) ";
   }
   else if (documode=="equilibrium"){
      V<R> ek(np);
      for (i=1;i<=np;++i) ek[i]=par[i].kinEnr();
      DisR dek(ek);
      R tDisplay=rc_.getR("equilibrium control","tDisplay");
      Word title1=
         "x: kinetic energy, y: binned particle numbers(black),";
      title1&=" Boltzmann fit(violet)";
      R2 val=dek.tBoltzmann(tDisplay,title1);
      R tx=get_t();
      R r1=logS(val[1]);
      R r2=logS(val[2]);
      R1(tx).prnOn(ofs());
      R1(r1).prnOn(ofs());
      R1(r2).prnOn(ofs());
      n=3;
      title="x: time, y: log T(black), log of misfit(violet)";
   }
   else if (documode=="trajectory"){
      Spc ori;
      Z np=nTot_;
      Spc p=par[np].get_x();
      Vec x=p-ori;
      R r1=x[1];
      R r2=x[2];
      R1(r1).prnOn(ofs());
      R1(r2).prnOn(ofs());
      n=2;
      title="Path of the large particle (equal scale in x- and ";
      title&="y-direction). Start at center of screen.";
   }
   else if (documode=="trajectoriesX" ||
      documode=="trajectoriesY" || documode=="trajectoriesXY"){
      Spc ori;
      R tx=get_t();
      ofs()<<"\n"<<tx;
      for (i=par.b();i<=par.e();++i){
         Spc p=par[i].get_x();
         R r1=p[1];
         R r2=p[2];
         ofs()<<" "<<r1<<" "<<r2;
      }
      if (documode=="trajectoriesXY")
         title="2D-paths of the particles";
      else if (documode=="trajectoriesX")
         title="X-coordinates of the particles";
      else
         title="Y-coordinates of the particles";
   }
   else if (topic=="mill" && documode=="all"){ // total_energy,
      // asymmetry, trajectories
      R tx=get_t();
      R er=0;
      FieldSpace fs=getExt()(1).getField(Spc());
      R eTot=kinEnr()+potEnr()+potElasticExt()+potHomFil(fs);
      if (!running_){
         eTotIni=eTot;
         eTotIniInv=cpminv(cpmabs(eTotIni));
         running_=B(true);
      }
      else{
         er=(eTot-eTotIni)*eTotIniInv;
      }
      R asy=asym(); // logaritmic value
      ofs()<<"\n"<<tx<<" "<<er<<" "<<asy; // all values
      // beloning to tx are written on one line for easy gnuplot
      // processing
      for (i=par.b();i<=par.e();++i){
         Spc p=par[i].get_x();
         ofs()<<" "<<p[1]<<" "<<p[2]<<" "<<p[3];
      }
   }
   else if (documode=="idle"){
      cpmmessage(loc&": idle documode selected");
   }
   else{
      cpmwarning(loc&": unvalid documode");
   }
   CPM_MZ;
   return X2<Z,Word>(n,title);
}

Z Sys2::parType(Z i)const
   // returns a discrete type number of par[i]. For the osmosis
   // application 1 for the small particles and 2 for the large ones.
{
   Word topic=rc_.getW("topic under consideration","topic");
   R rMean=rc_.getR("data","rMean");
   if (topic=="osmosis"){
      R ri=par[i].get_r();
      return ri<=rMean ? 1 : 2;
   }
   else return 0;
}

Z Sys2::parLoc(Z i)const
  // returns a discrete type number of par[i]. For the osmosis
  // application
  // 1 for the small particles and 2 for the large ones
{
   Word topic=rc_.getW("topic under consideration","topic");
   if (topic=="osmosis"){
      Spc xi=par[i].get_x();
      R h=xi[2];
      return h<=0 ? 1 : 2;
   }
   else return 0;
}

void Sys2::boost(const Vec& v, R angleRange)
{
   Word topic=rc_.getW("topic under consideration","topic");

   if (topic=="osmosis"){ // let the smaller particles
      // get the same kinetic energy than the large ones
      R facVel=rc_.getR("data","facVel");
      Angle a(angleRange,DEG);
      Z j=0;
      for (Z i=1;i<=nTot_;++i){
         Vec vi=v;
         Z ti=parType(i);
         if (ti!=1) vi*=facVel;
         vi=vi.randomize(0,a,++j);
         par[i].boost(vi);
      }
   }
   else{
      Sys2Base::ranBoost(v,0,angleRange);
   }
}

void Sys2::setVel()
{
   Word sec("runcontrol2");
   V<R> v0=rc_.getRs(sec,"v0");
   R vAngleRange=rc_.getR(sec,"vAngleRange");
   Vec vIni(v0);
   boost(vIni,vAngleRange);
      // now each particle has its regular velocity
      // Finally we have to stop the Brownian large
      // particles if we are dealing with this case.
   sec="topic under consideration";
   Word topic=rc_.getW(sec,"topic");
   if (topic=="general"){
      sec="general data";
      R xBrown=rc_.getR(sec,"xBrown");
      if (xBrown>1){
         Z np=nTot_;
         par[np].stop();
      }
   }
}

} // namespace
