//? cpmdynaux.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmdynaux.h>

using namespace CpmRoot;

using CpmDynSys::StepControl;
using CpmRootX::Record;

void CpmDynSys::adjust_dt(R& dt, Z counter)
{
   const R incr1=2;
   const R incr2=cpmsqrt(incr1);
   const R incr1Inv=R(1)/incr1;
   const R incr2Inv=R(1)/incr2;

   if (counter<=1) dt*=incr1;
   else if (counter==2) dt*=incr2;
   else if (counter<=4) ;
   else if (counter<=8) dt*=incr2Inv;
   else  dt*=incr1Inv;
}

/////////////////// struct StepControl ////////////////////

bool StepControl::prnOn(ostream& str)const
{
   CpmRoot::writeTitle("StepControl",str);
   cpmp(t);
   cpmp(tLast);
   cpmp(dtLast);
   cpmp(tLimit);
   cpmp(relaxFlag);
   cpmp(dtMean);
   cpmp(disLast);
   cpmp(disMean);
   return true;
}

bool StepControl::scanFrom(istream& str)
{
   cpms(t);
   cpms(tLast);
   cpms(dtLast);
   cpms(tLimit);
   cpms(relaxFlag);
   cpms(dtMean);
   cpms(disLast);
   cpms(disMean);
   return true;
}

void StepControl::update()
{
   const R w=0.9; // the nearer to 1 this quantity is, the longer into
   // the past the dt-memory reaches
   dtMean=w*dtMean+(-w+1)*dtLast; // floating mean value
   disMean=w*disMean+(-w+1)*disLast;
}

void StepControl::done(R dt_, R dis_)
{
   tLast=t;
   t+=dt_;
   dtLast=dt_;
   disLast=dis_;
   update();
}

void StepControl::doneTill(R t_, R dis_)
{
   tLast=t;
   t=t_;
   dtLast=t.x1-tLast;
   disLast=dis_;
   update();
}
