//? pala.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/************************************************************************
   pala.cpp

   Author: Ulrich Mutze
   Description: Application framework for programs of
   project pala

   History: Started 2005-03-15
   2009-02-20 option to have only one of CPM_DIMDEF2 and CPM_DIMDEF3
      defined re-introduced.

************************************************************************/
#include <cpmrigbodydyn.h>
#include <cpmsys2aps.h>
#include <cpmsysviewer.h>
#include <cpmactionprinciple.h>
   // these include cpmapplication.h

using CpmRoot::Z;
using CpmRoot::Word;
using CpmSystem::glue;
using CpmSystem::Message;
using CpmArrays::V;
using CpmFunctions::F;
using CpmApplication::Task;
   // a new (2005-04-23) powerful tool for implementing
   // menu-like interfaces to programs
using CpmProperties::SysControl;
using namespace CpmStd;

namespace{

// creating and running tasks is here defined as functions
// of a common generic signature which allows to combine
// them into complex selection and call structures (menus)

Z pala1(Word const& dummy)
// Notice that, by convention, an object of type F<Word,Z>
// is needed as the second argument of a Task-constructor.
// To define such an object is the reason for defining pala1().
// So it has to have a Word-typed argument for syntactic reasons.
{
   dummy;
   cpmmessage("pala1 started");
   cpmstatus("pala1 started",4);
   SysControl::reset();
#if defined(CPM_DIMDEF3)
   CpmPala1_3::App1RigBodyDyn a("pala1");
   a.setTitle("Rigid Body Dynamics");
   a.run();
#else
   cpmerror(" 3D - representation not defined");
#endif
   return 1;
}

Z pala2c(Word const& iniFile)
{
   Word loc("pala2c with ini-File "&iniFile);
   Z mL=1;
   CPM_MA
   using namespace CpmApplication;
   SysControl::reset();
      // former tasks may have changed the static variables
      // in a way that differes from the state that
      // the present task experienced in independent
      // testing
   setDimApp(iniFile);
       // needed to initialize dimApp from ini-file
   if (dimApp==2){
      cpmmessage("dimApp==2");
      #if defined(CPM_DIMDEF2)
      CpmPala2_2::Sys2Appl a(iniFile);
      a.setTitle("Particles and Walls");
      a.run();
      #else
      cpmerror(" 2D - representation not defined");
      #endif
   }
   else{
      cpmmessage("dimApp!=2");
      #if defined(CPM_DIMDEF3)
      CpmPala2_3::Sys2Appl a(iniFile);
      a.setTitle("Particles and Walls");
      a.run();
      #else
      cpmerror(" 3D - representation not defined");
      #endif
   }
   CPM_MZ
   return 2;
}

Z pala3(Word const& dummy)
{
   dummy;
   cpmmessage("pala3 started");
   cpmstatus("pala3 started",4);
   SysControl::reset();
   CpmPalaAux::SysViewerAppl a("pala3");
   a.setTitle("Movie Viewer");
   a.run();
   cpmmessage("pala3 done");
   cpmstatus("pala3 done",4);
   return 3;
}

Z pala4(Word const& dummy)
// Started 2006-06-22
{
   dummy;
   cpmmessage("pala4 started");
   cpmstatus("pala4 started",4);
   SysControl::reset();
   CpmPalaAction::ActionAppl a("pala4");
   a.setTitle("Action Principle");
   a.run();
   return 4;
}

}// namespace

Z CpmApplication::main_()
{
// we define tasks and combine them to a single one
   Z mL=0;
   Word loc("CpmApplication::main_()");
   CPM_MA
   CpmApplication::title_=Word("PaLa");
   Z nT=4;
   V< Task > vTask(nT);
   vTask[1]=Task("pala1",pala1);
   vTask[2]=Task::subMen("pala2",pala2c);
   vTask[3]=Task("pala3",pala3);
   vTask[4]=Task("pala4",pala4);
   Task rt=Task::men("pala",vTask);
// finally we execute them (by executing their combined version)
   Z res=rt.work();
   CPM_MZ
   return res;
}
