//? cpmdefinitions.h
//? C+- by Ulrich Mutze. Status of work 2018-11-24.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//?
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//?
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for
//? more details.


#ifndef CPM_DEFINITIONS_H_
#define CPM_DEFINITIONS_H_
/*************************************************************************
   cpmdefinitions.h
   Description: In this file one can place defines that are valid in all
   translation units working with Cpm classes
   Cpm = Classes for Physics and Mathematics || C+- (C plus minus)
         -           -           -                   - -    -
************************************************************************/
// This file will never be included by a C+- program since its name is
// cpmdefinitionswrc.h instead of cpmdefinitions.h. Upon renaming, it can
// function as cpmdefinitions.h.
// The name appendix 'wrc' stands for 'with reference comments'.
// We dont't use the more readable form cpmdefinitions_wrc.h in order
// not to disturb pdflatex-based Rugy script listing generation.
// The role of this file is to hold valid and complete explanations of the
// macros which can be set in any file cpmdefinitions.h.
// A file with similar scope is cpmsystemdependencieswrc.h.

#define CPM_NAMEOF
   // If this is defined, the C+- naming scheme is used instead of
   // C++ typeinfo. Recommended since for template types, native
   // typeinfo creates inconveniently long names.

//#define CPM_LONG
   // If this is defined, the types Z and N are long int and unsigned
   // long int, otherwise simply int and unsigned int.
   // If this is defined and CPM_MP is not defined the type R is
   // long double (which, unfortunately, is not always longer than
   // double).
   // Was not found to make much of a difference.

#define CPM_USECOUNT
   // If this is defined, all arrays use reference counting and
   // copy on write.
   // Recommended, increases efficiency.

#define CPM_RANGE_CHECK
   // If this is defined, all indexed access to arrays is index-
   // checked.
   // Recommended, increases safty.

//#define CPM_USE_MPI
   // MPI = message passing interface (for parallel computing).
   // See cpmmpi.h and cpmmpi.cpp.
   // This functionality was virtually unused for 7 years and was
   // reanimated in 2009-05-27. After dislocation of a single
   // b(road)cast statement in the main constructor of RecordHandler
   // a 2D example with gravitating and elastic repelling particles
   // worked fine and created a movie on which all particles were
   // to be seen in a common field of view. The command for
   // running was:
   //    mpirun -localonly -np 4 palaMS.exe
   // However, the C+- system favors a style of programming which
   // abstracts from technical aspects of the computing environment.
   // So questions of how to distribute a computational task over
   // various processors or various computers are not considered
   // of interest here. They  w e r e  of interest in the industrial
   // toning modeling project which contributed significantly to
   // the experiences on which the C+- system is based.

//#define CPM_MP 32
   // This macro enables usage of multiple precision
   // mathematics and assumes that the libraries gmp and
   // mpfr are available for linking (and, of course, their headers
   // for inclusion). Also the files mpreal.cpp, mpreal.h (by Pavel
   // Holoborodko) need to be available for compilation and for
   // inclusion repectively. There is, however also a mode of installation
   // of mpfr and mpreal in which it is sufficient to have a
   // '#include <mpreal.h> once in the user code.
   // The number of decimal places may also be set in cpmconfig.ini.
   // This line has only than an effect if CPM_MP is already defined.

//#define CPM_FIGURE_FOR_PUBLICATION_1
   // Selects particular code in cpmpardynsys.h, which is easily found
   // by searching for 'CPM_FIGURE_FOR_PUBLICATION_1'.
   // Ad hoc solution.
   // No effect if file cpmpardynsys.h is not used.

//#include <cpmcompdef.h>
   // This allows the defines given to the compiler via options of type
   // -D to override the definitions of CPM_LONG and CPM_MP given
   // previously in this file. Although, for instance, a compiler option
   // -DCPM_R_MP_E asks for real numbers of 256 decimal places, a line
   // CPM_MP 100
   // in file config.ini will override this to 100 decimal places.

#endif
