//? tut1.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*************************************************************************
   Comparing std::vector and CpmArays::V with respect to speed
   of copy-construction and assignment. Notice that this drastic
   advantage of CpmArays::V over std::vector still exists in C++11, where
   moving instead of copying is said to be done where appropriate.
*********************************************************************/
#include <cpmbas.h>
/*************************************************************************
1. C+- filenames start with 'cpm' and contain neither underscores nor
   blanks. Therefore, even projects employing files from many sources are
   not likely to run into problems with non-unique file names.

This directive includes all C+- header files needed here;
'bas' stands for 'basics'. Notice that we have not written
#include "cpmbas.h"
and thus have assumed that our C++ project - let it be named tut1 -
has a list of include directories defined.
Who wants to use C+- for more than a single project, should hold
C+- files in directories separate from his project files.
The general purpose C+- files to be used in the project tut1 are on my
computer in xxx/cpm/cpm0/include and xxx/cpm/cpm0/source.
The files that are specific for the tut1 application are in
yyy/tut1/include and yyy/tut1/source.
In yyy/tut1/include there are the project specific headers
among which there need to be two C+- related configuration files:
cpmdefinitions.h and cpmsystemdependencies.h. Customizing these files
allows us to control the behavior of the C+- classes in our project, as
will be explained later.
In yyy/tut1/source the present project-defining main source tut1.cpp is to
be placed.
Now we are in a position to define the file content of project
tut1: Its include directories have to be set as
yyy/tut1/include and xxx/cpm/cpm0/include
and the files to be compiled (translation units) have to be chosen as
yyy/tut1/source/tut1.cpp and xxx/cpm/cpm0/source/cpmbas.cpp.

1.1 Details (skip on first reading ?):
The file cpmbas.cpp is actually a collection of all other files
in xxx/cpm/cpm0/source. Its content is

#include "cpmc.cpp"
#include "cpmangle.cpp"
#include "cpmv.cpp"
#include "cpmgreg.cpp"
#include "cpmsystem.cpp"
#include "cpmtypes.cpp"
#include "cpmuc.cpp"
#include "cpmzinterval.cpp"
#include "cpmnumbers.cpp"
#include "cpmmpi.cpp"
#include "cpmword.cpp"
#include "cpmviewport.cpp"
End of 1.1
*********************************************************************/
using namespace CpmRoot;
/*************************************************************************
2. C+- namespace names start with 'Cpm' followed by an identifier
   which starts with a capital letters.

The namespace CpmRoot is rather small, so that the present using directive
should be applicable also in projects in which classes from various
sources are being used.

3. The C+- names for integer, real, and complex numbers are
   CpmRoot::Z, CpmRoot::R, Cpmoot::C. CpmRoot::Word wraps std::string and
   adds functionality to it.

The main effect of this using directive is that we may use these names
simply as Z, R, C, Word.

3.1 Details (skip on first reading ?):
There are the less ubiquitous types L, and N in CpmRoot:
L for unsigned characters ('L' for 'letter', after 'integer promotion'
the values range from 0 to 255), N for 'natural numbers', i.e. unsigned
integers. Types L,Z, and N are typedefs for unsigned char, int,
unsigned int. If we add in file cpmdefinition.h the macro
#define CPM_LONG, we change Z to long int, and N to unsigned long int.
If in cpmdefinitions.h the macro CPM_MP is defined, R has the meaning of
mpreal as defined by the wrapper library MPFRC++ to the multiple precision
library mpfr. See comments to CPM_MP in file cpmdefinitionswrc.h. If
CPM_MP is not defined R is also a typedef, long double if CPM_LONG and
double else. The types bool and string from standard C++ are useful as
they stand. Nevertheless they will be wrapped into classes - CpmRootX::B
in cpmtypes.h and class CpmRoot::Word in file cpmword.h.
End of 3.1

There is a namespace CpmRootX for the 'less essential essentials':

4. There are classes CpmRootX::B, CpmRootX::Z1, CpmRootX::R1,  for
   Boolean values, integer, and real numbers, which may replace the
   non-class types bool, Z, and R in situations where class func-
   tionality, such as automatic initialization, is indispensable.
   Note 2017-02-18: With C++11 automatic initialization can be achieved
   also for the built-in types. For instance a data member of a class X
         bool b {false};
   can be handeled as if we had declared it as
         B b;

4.1 Details (skip on first reading ?):
Namespace CpmRoot in mainly layed out in file cpmnumbers.h
and CpmRootX in cpmtypes.h. The vigilant reader may argue that we
have not available the declarations of CpmRoot, since we have no
   #include <cpmnumbers.h>
seen so far. Actually, it is there since file cpmbas.h is the following
collection of include directives
#ifndef CPM_BAS_H_
#define CPM_BAS_H_
   #include <cpmtypes.h>
   #include <cpmfr.h>
   #include <cpmvr.h>
   #include <cpmsr.h>
   #include <cpmm.h>
   #include <cpmp.h>
   #include <cpmc.h>
   #include <cpmangle.h>
   #include <cpmgreg.h>
#endif
Also here we don't see
   #include <cpmnumbers.h>
but inspecting cpmtypes.h, we find the commented directive
   #include <cpmsystem.h> // includes cpmwords.h and
      // thus also cpmnumbers.h
which shows the steps which finally include cpmnumbers.h. Acually the
directory cpm/cpm0/include contains 32 header files. This set of files
(as any set of C++ header files) is a directed graph in a natural manner:
There is an edge leading from file1 to file2 iff file2 contains the
directive  #include <file1> .
The files that can be reached from file1 along a path of that graph depend
on file1 and every translation unit which includes such a file1-dependent
header has to be re-compiled upon some change in file1. So, any system
capable of generating proper make files has to do this dependency
analysis and thus has to work with this  d e p e n d e n c y  g r a p h.
I made a tool consisting of a Ruby program for file analysis
and a C+- program for graph analysis and representation which
creates a  pictorial representation of this dependency graph of an
arbitrary collecton of C/C++ header files. This tool turned out to be very
useful.
It is an obvious logical requirement that the dependency graph is free of
cycles. It is a non-trivial task to organize the header dependencies in a
way that each translation unit gets the declarations which it needs to
know by inclusion of only a few header files. For instance, cpmword.cpp
and cpmtypes.cpp need only cpmtypes.h; cpmc.cpp needs cpmc.h and
cpmtypes.h. The present header file hierarchy is the result of many
simplifying re-organizations. However, many attempts of a simplification
failed since they entailed unacceptable complications elsewhere. I'm not
aware of tools which would automatize relevant parts of this organization
process.
End of 4.1
***********************************************************************/
using namespace CpmArrays;
/*************************************************************************
5. The general-purpose array in C+- is the template class CpmArrays::V
   Most constructors set the first valid index of a non-void V is 1 and
   not 0 as for std::vector. There is, however, a member function which
   shifts all indexes and so allows indexing to start with 0. This allows
   identical code to be used for these two types of arrays.

5.1 Details on C+- namespaces (skip on first reading ?):
The C+- class system introduces many namespaces. Their belonging
togeher under the Cpm banner is not expressed by making them
sub-namespaces of a single Cpm-namespace. Instead, their names all
start with 'Cpm' and continue with a capital letter.
If the name contains a digit, this is 2 or 3 and refers to the space
dimension under consideration. So the namespace CpmDim2 contains the
2D ('flatland') analogs of the geometric classes defined in
namespace CpmDim3.
The namespaces which are declared in the present scope by including
cpmbas.h are
CpmRoot, CpmRootX, CpmSystem, CpmArrays, CpmFunctions, CpmGeo,
CpmGraphics, CpmMPI, CpmStd, CpmTests, CpmTime.
End of 5.1
*********************************************************************/

void info();
Z tutorial1(Z,Z);
   // Declaration of two functions, the definition of which will
   // determine the functionality of the program.

int main(int argc, char* argv[])
   // Traditional main function with types of arguments and return value
   // according to C/C++. Not all compilers accept using Z instead of
   // int here.
{
 //  Z m=200, n=20000; // default values to be used if no input from the
      //  command line can be found
   Z m=5000, n=5000;

   V<Word> arg=comLine(argc,argv); // CpmRoot::Word is similar to
      // std::string. CpmArrays::V is the C+- array type. There are two
      // 'schools' about indexing arrays. The 'C school' lets valid
      // indexes start with 0 and the 'Fortran school' which lets valid
      // valid indexes start with 1. Till September 2010 C+- had a
      // 'C school'-array Vl (l for 'lean') and a 'Fortran school'-array
      // V , where the implementation of V was such that any instance of
      // V had a data member of type Vl. The type Vl was used mostly
      // internally in situation where efficiency was considered to be
      // of utmost importance. Finally it became clear that the decision
      // to have these two types of arrays was a mistake. It creates a
      // permanent uncertainty for the programmer whether he should use
      // the default type or should strive for utmost efficiency by
      // using the lean version. Now we have only a single array type V
      // for which the valid indexes may form any contiguous set of
      // integers. This index range is set to start with 1 in most
      // of the available constructors but can be very conveniently reset
      // to any other start index such as 0.
      // CpmArrays::comLine is a function which transforms the
      // traditional argument of main into the more functional data type
      // of C+-. The function name 'comLine' results from
      // the  d e s c r i p t i v e  f u l l  n a m e (DFN)
      // 'command line' by a simple deterministic rule: The rule is
      // to leave unchanged all words with up to four letters and shorten
      // all longer words to 3 by taking the first two letters as they are
      // and take the first of the following consonants as the third
      // letter. Thus 'ooops' gets converted to 'oop'. If there is no
      // such consonant, the vowel at place 3 has to be taken: 'hooiii'
      // gets converted to 'hoo'. In a chain of words, the contracted
      // components get concateneted in a style which is evident from
      // the example that 'descriptive full name' goes to
      // 'desFullName'. In most cases the names come out quite nice, but
      // there are exceptions (not to the rules!): I cosider it ugly to
      // see 'field' abbreviated to 'fil' (notice that 'file', as having
      // only four letters, remains 'file')
      // The simplicity of the rule allows us to use the the function name
      // fluently whenever we remember the full name correctly. Of course,
      // the descriptive full name of a function has to be stated in the
      // declaration of this function. In our case the declaration is in
      // file cpmv.h:
      // inline V<Word> comLine(int argc, char* argv[])
      //    //: command line
      // The colon hints at the fact that here a descriptive full name
      // is being communicated.
   if (arg.valInd(2)){ // valInd's DFN is 'valid index'
      // C+- arrays judge the validity of an index directly without
      // a need to ask for 'size' and making a comparison.
      Word w2=arg[2];
      if (w2=="?"){ // Thus a question mark as the second entry of the
            // command line triggers a call to function info.
         info();
         return 0;
      }
      else{
         m=w2.toZ(); // Converting Word to Z, for getting a control
            // parameter. There are also conversions toR(), toBool(),
            // and toStr()
         if (arg.valInd(3)) n=arg[3].toZ();
      }
   }
   return tutorial1(m,n);
  }

using namespace std;

C cz(Z i) // an ad-hoc function Z --> C
// CpmRoot::C is the class of complex numbers
{
   R ir=i;
   return C(ir/(1+ir*ir),ir*cpmpi); // there is a CpmRoot::Pi = 3.14159...
      // calling constructor C(R,R)
}

// Now we define the performance measuring functions perf_v and perf_V.
// These functions have to execute identical code for two different
// vector templates: std::vector in perf_v and Cpmrrays::V in perf_V.
// Since we want to iterate these templates, it would not be easy to
// achieve this code doublication by defining a template function.
// Such a template would need at least two template arguments and would
// enforce unnatural expressions.

// This function is designed as to maximize the benefit from reference
// counting and to report the time spent on copy-construction and
// assignment (not of the tedious verification part of the function).

#define Vec vector
V<R> perf_v(Z m, Z n)
{
   Z mL=1;
/*
   This line, together with the following
   Word loc(...);
   CPM_MA
   CPM_MZ
   consitute a convenient idiom for signaling
   entry to and exit from a function block to the log file
   cpmcerr.txt. This writing to the log file takes place only if
   mL is larger or equal to the static data member
   CpmSystem::Message::verbose. This bulky quantity has a
   macro alias 'cpmverbose' (i.e. we have, in file cpmsystem.h
       #define cpmverbose      CpmSystem::Message::verbose
   ) and it can be set by a statement like
       cpmverbose=10;
   Its initial value is 2.
   Soon we will encounter macros cpmtime, cpmwait, cpmdebug.
   See the remarks on the macro namespace in function perf_V.
*/
   Word loc("perf_v(Z,Z)"); // messages use this as name of the function
   CPM_MA // defined in cpmmacros.h, assumes that 'mL' and 'loc' are
      // defined
   R t1=cpmtime(); /* time of function call in seconds from some
       system-defined 'point-zero in time'.
       cpmtime is a short name for function CpmSystem::time defined
       in file cpmsystem.h as follows:
           #define cpmtime         CpmSystem::time
   */
   Z i,j;
   Vec< Vec<C> > vi,vf; // we use arrays, the components of which
      // are potentially large objects. This is the strong point of
      // C+- arrays and a weak one of std::vector and of std::valarray.
   {
      Vec<C> v1(m);
      for (i=0;i<m;++i) v1[i]=cz(i);
         // The preferred form of this statement in C+- is
         // for (i=v1.b();i<=v1.e();++i) v1[i]=cz(i);
         // Written in this form, it is alo correct if v1 is of type V<C>,
         // the C+- array for which indexing starts with 1 instead of 0.

      Vec< Vec<C> > w(n,v1);
         // If we use valarray for Vec, one has to write w(v1,n) instead.
         // This is a inconvenience of the C++ standard library which is
         // due to their multi-source origin.
      Vec< Vec<C> > temp1=w; // copy constructor
      Vec< Vec<C> > temp2=temp1; // copy constructor
      Vec< Vec<C> > temp3=temp2; // copy constructor
      temp2=Vec< Vec<C> >(0); // assignment
      vi=w; // assignment
      vf=temp3; // assignment
   }
   R t2=cpmtime();
   R t12=t2-t1; // time needed for copy and assignment
   R err=0,sum=0;
   for (i=0;i<n;++i){
      for (j=0;j<m;++j){ // sic!
         err+=(vi[i][j]-vf[i][j]).abs();
         sum+=(vf[i][j]).abs();
      }
   } // making sure that copying and assignment worked correctly
   R t3=cpmtime();
   R t23=t3-t2;
   V<R> res{t12,t23,err,sum};
/*
   Convenient  l i s t  c o n s t r u c t o r, requires C++11
*/
  CPM_MZ
  return res;
}
#undef Vec
#define Vec V
V<R> perf_V(Z m, Z n)
{
   Z mL=1;
   Word loc("perf_V(Z,Z)");
      // Here goes code which is identical to the one of the
      // previous function perf_v. In actual C+- code this
      // is never implemented by copying some piece of code
      // but always by defining the macro CPM_SC (SC for 'shared
      // code' (and undefining it after usage). C+- claims a
      //  m a c r o  n a m e s p a c e :  all macros beginning
      // with the four characters 'CPM_' or the three characters
      // 'cpm'. The upper case macros stand in most cases for a few
      // lines of code and should be recognized by the user as not
      // being a simple identifier. They are thus written without a
      // semi-colon at the end in code since the semi-colon is part
      // of the macro's definition.
      // The lower-case macros, by contrast, are in most cases
      // #define-implemented abbreviations of quantities for which
      // a normal name is available (e.g. CpmSystem::Message::verbose
      // for cpmverbose). Whether these are variables or functions, they
      // are used as normal identifiers in code and the user needs not to
      // be aware of their macro nature. Hence, these macros are thus
      // written in code with a semi-colon at the end. (The semi-colon is
      // not part of the macro's definition.)
      //
   CPM_MA  // macro not followed by a semi-colon
   R t1=cpmtime(); // macro followed by a semi-colon
   Z i,j;
   Vec< Vec<C> > vi, vf;
   {
      Vec<C> v1(m);
      v1.b_(0); // lets indexing of v1 start with 0
      for (i=0;i<m;++i) v1[i]=cz(i);
      Vec< Vec<C> > w(n,v1);
      w.b_(0); // lets indexing of w start with 0
      Vec< Vec<C> > temp1=w;
      Vec< Vec<C> > temp2=temp1;
      Vec< Vec<C> > temp3=temp2;
      temp2=Vec< Vec<C> >(0);
      vi=w;
      vf=temp3;
   }
   R t2=cpmtime();
   R t12=t2-t1;
   R err=0,sum=0;
   for (i=0;i<n;++i){
      for (j=0;j<m;++j){
         err+=(vi[i][j]-vf[i][j]).abs();
         sum+=(vf[i][j]).abs();
      }
   }
   R t3=cpmtime();
   R t23=t3-t2;
   CPM_MZ
   return V<R>{t12,t23,err,sum};
}
#undef Vec

Z tutorial1(Z m, Z n)
{
   Z mL=1;
   Word loc("tutorial1(Z,Z)");

   CPM_MA
   V<R> res_V=perf_V(m,n), res_v=perf_v(m,n);
   R t12_V=res_V[1], t12_v=res_v[1];
   R t23_V=res_V[2], t23_v=res_v[2];
   R t_V=t12_V+t23_V, t_v=t12_v+t23_v;
   R err_V=res_V[3], err_v=res_v[3];
   R s_V=res_V[4], s_v=res_v[4];
   R fac12=t12_v*cpminv(t12_V);
   R fac=t_v*cpminv(t_V);
   cout<<"m="<<m<<", n="<<n<<endl;
   cout<<"t12_v="<<t12_v<<", t12_V="<<t12_V<<endl;
   cout<<"t23_v="<<t23_v<<", t23_V="<<t23_V<<endl;
   cout<<"fac12="<<fac12<<endl;
   cout<<" fac="<<fac<<endl;
   cout<<" err_v="<<err_v<<", err_V="<<err_V<<endl;
      // output of the result to the console
   cpmdebug(m);
   cpmdebug(n);
   cpmdebug(t12_v);
   cpmdebug(t12_V);
   cpmdebug(t23_v);
   cpmdebug(t23_V);
   cpmdebug(fac12);
   cpmdebug(fac);
   cpmdebug(err_v);
   cpmdebug(err_V);
      // convenient documentation of the main result on
      // the auto-generated log file cpmcerr.txt.
      // Defined in file cpmtypes.h as
   // #define cpmdebug(X) cpmcerr<<endl<<"C+- debug: "<<#X "="<< X <<endl
      // Here are the corresponding lines of this file:

//2017-08-01 17:47:16.0 Z

//C+- debug: m=5000
//
//C+- debug: n=5000
//
//C+- debug: t12_v=0.680035
//
//C+- debug: t12_V=0.000561
//
//C+- debug: t23_v=20.7179
//
//C+- debug: t23_V=20.9999
//
//C+- debug: fac12=1212.18
//
//C+- debug: fac=1.01893
//
//C+- debug: err_v=0
//
//C+- debug: err_V=0
//
//counter = 7
// Message: tutorial1(Z,Z) done
//  42.4427 s after program start
// Thus, as expected, the copy and assignment is much faster (in our
// case times 1212 for C+- vectors as it is for STL vectors.

   CPM_MZ
   return 0;
}

void info(){
   cout<<endl<<"takes zero, one, or two integer argument";
   cout<<endl<<"compares the performance or std::vector and CpmArays::V";
   cout<<endl<<"in copy construction and assignment operations"<<endl;
}
// end of tut1.cpp
