//? vqm2app3.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//////////////////////////////////////////////////////////////////////////
//   vqm2app3.cpp
//////////////////////////////////////////////////////////////////////////

#include <vqm2app3.h>
#include <vqm2.h>
#include <cpmqm.h>
#include <cpmqmx.h>
#include <cpmfieldstate.h>
#include <cpmvectormore.h>
#include <cpmmultiimage.h>
#include <cpmfile.h>

using namespace CpmSystem;
using namespace CpmRoot;
using namespace CpmRootX;
using namespace CpmArrays;
using namespace CpmFunctions;
using namespace CpmGraphics;
using namespace CpmImaging;
using CpmAlgorithms::CyclicAlarm;

//////////////////////////////// App17 ///////////////////////////////////

void CpmVQM2::App17::doTheWork()
// Experimenting with two and three separated, weakly interacting
// particles. Also tool for creating the publication data.
// All publication data made with sel==8.
{
   Z mL=1;
   Word loc("CpmVQM2::App17::doTheWork()");
   CPM_MA
   using namespace CpmQM;
   V<OFile> vout(outFiles.dim());
   for (Z it=vout.b();it<=vout.e();++it){ // open OFiles
      vout[it]=OFile(outFiles[it]);
   }
   Z d1=50,d2=50,d3=50,d4=3;
   Z n1=0,n2=0,n3=0;
   R gap=0;

   Word sec="system data";
   cpmrh(gap);
   cpmrh(d1);
   cpmrh(d2);
   cpmrh(d3);
   cpmrhf(d4);

   cpmrh(n1);
   cpmrh(n2);
   cpmrh(n3);

   Iv iv1(0,1);
   Iv iv2(1,2);
   Iv iv3(2,3);
   iv2+=gap;
   iv3+=gap*2;

   Lat1 lat1(iv1,d1);
   Lat1 lat2(iv2,d2);
   Lat1 lat3(iv3,d3);

   if (vout.valInd(1)){
       vout[1]()<<"lat1="<<lat1<<endl;
       vout[1]()<<"lat2="<<lat2<<endl;
       vout[1]()<<"lat3="<<lat3<<endl;
   }

   R m1=lat1.mMax();
   R m2=lat2.mMax();
   R m3=lat3.mMax();

   R q1=1,q2=1,q3=1;

   R lambda1=lat1.lamOsc(m1);
   R lambda2=lat2.lamOsc(m2);
   R lambda3=lat3.lamOsc(m3);

   R p1=2,p2=2,p3=2;
   R x1Rel=0.5, x2Rel=0.5, x3Rel=0.5;
   R lambda1Rel=1, lambda2Rel=1, lambda3Rel=1;
   R m1Rel=0.5, m2Rel=0.5, m3Rel=0.5;
   R fac1=1, fac2=1, fac3=1;
   R beta=0;
   Z out1=0,out2=0,out3=0;
   bool homIniOff=true;
   bool simplifyFun3=false;
   // method parameter in Fun3::step_
   R epsilonFun3=0;
   // method parameter in Fun3::step_

   cpmrh(x1Rel);
   cpmrh(x2Rel);
   cpmrh(x3Rel);

   cpmrh(m1Rel);
   cpmrh(m2Rel);
   cpmrh(m3Rel);

   cpmrh(q1);
   cpmrh(q2);
   cpmrh(q3);

   cpmrh(lambda1Rel);
   cpmrh(lambda2Rel);
   cpmrh(lambda3Rel);

   cpmrh(p1);
   cpmrh(p2);
   cpmrh(p3);

   cpmrhf(out1); // needs not to be there
      // since 0 is OK
   cpmrhf(out2);
   cpmrhf(out3);

   cpmrh(fac1);
   cpmrh(fac2);
   cpmrh(fac3);

   cpmrh(beta);
   cpmrhf(homIniOff);

   cpmrhf(simplifyFun3);
   cpmrhf(epsilonFun3);

   Fun3::eps()=epsilonFun3;
   Fun3::sim()=simplifyFun3;


   R x1=iv1.x(x1Rel);
   R x2=iv2.x(x2Rel);
   R x3=iv3.x(x3Rel);

   m1*=m1Rel;
   m2*=m2Rel;
   m3*=m3Rel;

   lambda1*=lambda1Rel;
   lambda2*=lambda2Rel;
   lambda3*=lambda3Rel;

   DynDat dd1(m1,q1,lambda1,x1,p1,out1);
   DynDat dd2(m2,q2,lambda2,x2,p2,out2);
   DynDat dd3(m3,q3,lambda3,x3,p3,out3);

   sec="interaction data";
   IntActDat dd12(rch,sec);

   if (vout.valInd(1)){
       vout[1]()<<"dd1="<<dd1<<endl;
       vout[1]()<<"dd2="<<dd2<<endl;
       vout[1]()<<"dd3="<<dd3<<endl;
       vout[1]()<<"dd12="<<dd12<<endl;
   }

   R alpha=dd12.alpha();

   if (homIniOff)  DynDat::beta()=0;
   else DynDat::beta()=beta;

   SysLat1 s1(lat1,dd1,n1);
  // SysLat1 s1e(lat1,dd1,n1+1);
   SysLat1 s2(lat2,dd2,n2);
 //  SysLat1 s2e(lat2,dd2,n2+1);
   SysLat1 s3(lat3,dd3,n3);
//   SysLat1 s3e(lat3,dd3,n3+1);
      // eigenstates of one-particle hamiltonian
      // here no homogeneous fiels should be in action,
      // since particle shift (polarization) due to the
      // field should only occur during a dynamical run.

   DynDat::beta()=beta;
      // now DynDat::pot includes a homogeneous field

   SysLat2 s12(s1,s2);

   sec="run control";
   Z nAlpha=1,nBeta=1;
   bool fun3=false;
   cpmrhf(nAlpha);
   cpmrhf(nBeta);
   cpmrhf(fun3);
   RunCon rc(rch,sec);
   R tWaitFinal=6;
   R tWaitFinal2=6;
   cpmrh(tWaitFinal);
   cpmrhf(tWaitFinal2); // needs not to be there
   if (tWaitFinal>0 && rc.tWaitRes<=0) rc.tWaitRes=0.5;
   if (tWaitFinal>0 && rc.tWaitDocu<=0) rc.tWaitDocu=0.5;
      // otherwise res will not be created

   sec="display control";
   DisCon dc(rch,sec);

   dd1.mulLam_(fac1);
   dd2.mulLam_(fac2);
   dd3.mulLam_(fac3);
      // changing the coupling constant after
      // having formed the states

   Z sel=rc.sel;

   Graph g;
   if (sel==0){
       TrjLat1(dd1,s1).run_(g,rc,dc);
   }
   else if (sel==1){
      if (dc.ipolMet<0) dc.ipolMet=0;
      TrjLat2(dd1,dd2,dd12,s12).run_(g,rc,dc);
   }
   else if (sel==2){
      ; // free
   }
   else if (sel==3){
      Sep2(dd1,dd2,dd12,s1,s2).run_(g,rc,dc);
   }
   else if (sel==4){
      using namespace CpmImaging;
      DisCon dcMem=dc;
      if (dc.ipolMet<0) dc.ipolMet=0;
      V< V<R2> > res1=TrjLat2(dd1,dd2,dd12,s12).run_(g,rc,dc);
      dc.ipolMet=dcMem.ipolMet;
      V< V<R2> > res2=Sep2(dd1,dd2,dd12,s1,s2).run_(g,rc,dc);
      V< V<R2> > res=res1&res2;
      res=homDataRan(res);
      g.clr_();
      g.setAutoScl_(true);
      V<Color> vc(3);
      vc[1]=BLUE;
      vc[2]=GREEN;
      vc[3]=RED;

      g.mark(res,vc);
      g.addText("correlations from TrjLat2 and Sep2");
      g.vis();
      cpmwait(tWaitFinal,2);
   }
   else if (sel==5){
      using namespace CpmImaging;
      DisCon dcMem=dc;
      if (dc.ipolMet<0) dc.ipolMet=0;
      x2vv x1=Sep3(dd1,dd2,dd3,dd12,s1,s2,s3).run_(g,rc,dc);
      V<V<R2> > res1=x1.c1();
      dc.ipolMet=dcMem.ipolMet;
      x2vv x2=Fun3(dd1,dd2,dd3,dd12,s1,s2,s3).run_(g,rc,dc);
      V<V<R2> > res2=x2.c1();
      V< V<R2> > res=res1&res2;
      res=homDataRan(res);
      g.clr_();
      g.setAutoScl_(true);
      V<Color> vc(3);
      vc[1]=BLUE;
      vc[2]=GREEN;
      vc[3]=RED;

      g.mark(res,vc);
      g.addText("correlations from TrjLat2 and Sep2");
      g.vis();
      cpmwait(tWaitFinal,2);
   }
   else if (sel==6){
      if (dc.ipolMet<0) dc.ipolMet=0;
      SysLat3 s123(s1,s2,s3);
      TrjLat3(dd1,dd2,dd3,dd12,s123).run_(g,rc,dc);
   }
   else if (sel==7){
      if (fun3)  Fun3(dd1,dd2,dd3,dd12,s1,s2,s3).run_(g,rc,dc);
      else Sep3(dd1,dd2,dd3,dd12,s1,s2,s3).run_(g,rc,dc);
   }
   else if (sel==8){ // exact and separated dynamics for
      // three particles. Comparison
      using namespace CpmImaging;
      DisCon dcMem=dc;
      Z i;
      if (dc.ipolMet<0) dc.ipolMet=0;
  //    SysLat1 s2a=s2+s2e;
  //    s2a.nor_();
      SysLat3 s123(s1,s2,s3);
      s123.nor_();
      x2vv x1=TrjLat3(dd1,dd2,dd3,dd12,s123).run_(g,rc,dc);
         // exact dynamics, implemented by the direct midpoint integrator
         // may change rc by automatic time step adjustment, so that the
         // adjusted time step will be used by Sep3::run_ without further
         // adjustment
      if (vout.valInd(1)) vout[1]()<<"rc.dt="<<rc.dt<<endl;
      V<V<R2> > res1=x1.c1();
      V<R2> stExc=res1[8];
      dc.ipolMet=dcMem.ipolMet;
      x2vv x2=fun3 ? Fun3(dd1,dd2,dd3,dd12,s1,s2,s3).run_(g,rc,dc)
               : Sep3(dd1,dd2,dd3,dd12,s1,s2,s3).run_(g,rc,dc);
            // separated dynamics
      V<V<R2> > res2=x2.c1();
// comparing polarization (i.e. total particle shift)
// as a function of time
      V<R2> stSep=res2[8];
      V<V<R2> > res{stExc,stSep};
      if (dc.showCls){
          V<R2> stCls=res2[16];
          res&=stCls;
      }
      if (vout.valInd(2)) toGnuPlot(vout[2](),res);
      g.clr_();
      g.setAutoScl_(true);
      g.setSclMem_(false);
      g.mark(res);
      Word mes1("TrjLat3, ");
      Word mes2=fun3 ? "Fun3" : "Sep3";
      Word mes3=": polarization from exact and separated ";
      Word mes4=dc.showCls ? "and classical dynamics" : "dynamics";
      Word mes=mes1&mes2&mes3&mes4;
      g.addText(mes);
      g.vis();
      cpmwait(tWaitFinal,2);
// comparing also quantum correlations from both methods
      V<V<R2> > resCor=res1.resize(7)&res2.resize(7);
      resCor=homDataRan(resCor);
      if (vout.valInd(3)) toGnuPlot(vout[3](),resCor);
      bool homFirst=false; // data already homogenized
      resCor=difOfCorData(resCor,homFirst);
      if (vout.valInd(4)) toGnuPlot(vout[4](),resCor);
      V<Color> vc1{RED,GREEN,BLUE,ORANGE,MAGENTA,CYAN,DARKGRAY};
      V<Color> vc2{LIGHTRED,LIGHTGREEN,LIGHTBLUE,YELLOW,
         LIGHTMAGENTA,LIGHTCYAN,LIGHTGRAY};
      g.clr_();
      g.setAutoScl_(true);
      g.setSclMem_(false);
      g.mark(resCor,vc1);
      mes=mes1&mes2&
      ": difference of correlations for exact and separated dynamics";
      g.addText(mes);
      g.vis();
      cpmmessage("talata: came here");
      cpmwait(tWaitFinal2,2);
      V<Word> vw1=x1.c2();
      V<Word> vw2=x2.c2();
      Z dw1=vw1.dim();
      Z dw2=vw2.dim();
      cpmassert(dw1==dw2,loc);
      Word fn("twins_");
      if (dw2>0){ // only then there are images to be processed
         Z nf=dw1;
         for (i=1;i<=nf;++i){
            Word fn1=vw1[i];
            Word fn2=vw2[i];
            Img24 img_i=Img24(fn1).pair(Img24(fn2));
            img_i.write(fn&cpm(i,3));
         }
      }
   }
   else if (sel==9){
      // should create data which show the dependence of polarization
      // on alpha and beta. For this purpose we interprete the input
      // values as maximum value which, in a linear manner get finally
      // reduced to 0 in nAlpha and nBeta steps.
      using namespace CpmImaging;
      OFileStream ofs1(outFiles[1]);
      DisCon dcMem=dc;
      if (dc.ipolMet<0) dc.ipolMet=0;
      SysLat3 s123(s1,s2,s3);
      V< V<R2> > dat;
      V< Color > colors;
      R_Vector alphas(R2(alpha,0),nAlpha);
      Z nA=alphas.dim();
      R_Vector betas(R2(beta,0),nBeta);
      betas=betas.cut(1); // we don't beta=0, since polarizability is
         // shift/beta
      Z nB=betas.dim();
      R_Vector levels(R2(0,180),nA); // one level for each
         // alpha value
      // n equidistant components starting with p[1] and ending
      for (Z i=1;i<=nA;++i){
         R alphai=alphas[i];
         dd12.alpha()=alphai;
    //     betaj=beta;
         V<R2> dat1,dat2,dat3;
         for (Z j=1;j<=nB;++j){
            R betaj=betas[j];
            if (!homIniOff){ // then the initial states depend on
               // beta and have to be constructed anew
               DynDat::beta()=betaj;
               s1=SysLat1(lat1,dd1,n1);
               s2=SysLat1(lat2,dd2,n2);
               s3=SysLat1(lat3,dd3,n3);
               s123=SysLat3(s1,s2,s3);
            }
            DynDat::beta()=betaj;
            cpmvalues2("alpha,beta",alphai,betaj,3);
            x2vv x1=TrjLat3(dd1,dd2,dd3,dd12,s123).run_(g,rc,dc);
            V< V<R2> > res1=x1.c1();
         // exact dynamics, implemented by the direct midpoint integrator
         // may change rc by automatic time step adjustment, so that the
         // adjusted time step will be used by Sep3::run_ without further
         // adjustment. Note that the time step adjustment depends on
         // alpha and beta
            Z k;
            Z r11d=res1[1].dim();
            Z r12d=res1[2].dim();
            Z r13d=res1[3].dim();
            cpmassert(r11d==r12d,loc);
            cpmassert(r12d==r13d,loc);
            R_Vector pol1(r11d);
            for (k=1;k<=r11d;++k){
               R2 ri=res1[1][k]+res1[2][k]+res1[3][k];
               pol1[k]=ri[2];
            }
            dc.ipolMet=dcMem.ipolMet;
            x2vv x2=fun3 ? Fun3(dd1,dd2,dd3,dd12,s1,s2,s3).
               run_(g,rc,dc)
               : Sep3(dd1,dd2,dd3,dd12,s1,s2,s3).run_(g,rc,dc);
            V< V<R2> > res2=x2.c1();
            // separated dynamics
            Z r21d=res2[1].dim();
            Z r22d=res2[2].dim();
            Z r23d=res2[3].dim();

            Z rc21d=res2[8].dim();
            Z rc22d=res2[9].dim();
            Z rc23d=res2[10].dim();

            cpmassert(r21d==r22d,loc);
            cpmassert(r22d==r23d,loc);
            cpmassert(r23d==rc21d,loc);
            cpmassert(rc21d==rc22d,loc);
            cpmassert(rc22d==rc23d,loc);

            R_Vector pol2(r21d);
            for (k=1;k<=r21d;++k){
               R2 ri=res2[1][k]+res2[2][k]+res2[3][k];
               pol2[k]=ri[2];
            }
            R_Vector pol2c(r21d);
            for (k=1;k<=r21d;++k){
               R2 ri=res2[8][k]+res2[9][k]+res2[10][k];
               pol2c[k]=ri[2];
            }
            R betaInv=cpminv(betaj);
            R p1=extMean(pol1)*betaInv;
            R p2=extMean(pol2)*betaInv;
            R p3=extMean(pol2c)*betaInv;
            dat1<<R2(betaj,p1);
            dat2<<R2(betaj,p2);
            dat3<<R2(betaj,p3);
            cpmnote(alphai);
            cpmnote(betaj);
            cpmnote(p1);
            cpmnote(p2);
            cpmnote(p3);
         }// j-loop
         dat<<dat1;
         R level=levels[i];
         R high=255;
         colors<<Color(high,level,level);
         dat<<dat2;
         colors<<Color(level,high,level);
         dat<<dat3;
         colors<<Color(level,level,high);
      }//i-loop
      g.clr_();
      g.setAutoScl_(true);
      g.mark(dat,colors);
      g.addText("polarization curves, homIniOff="
         &cpm(homIniOff)&", alphaMax="&cpm(alpha));
      g.vis(true);
      ofs1()<<dat;
      cpmwait(tWaitFinal,2);
   }// sel==9
   else if (sel==10){
      testEC(d1,d2,d3);
   }
   else if (sel==11){
      testEC4(d1,d2,d3,d4);
   }
   else if (sel==12){  // exact and separated dynamics for
      // three particles. Comparison. Form usin Record input
      SysLat3 s123(s1,s2,s3);
      s123.nor_();
      x2vv x1=TrjLat3(dd1,dd2,dd3,dd12,s123).run_(g,rch);
      x2vv x2=Sep3(dd1,dd2,dd3,dd12,s1,s2,s3).run_(g,rch);
   }
   else{
      cpmmessage("unvalid selection");
   }
   CPM_MZ
}

//////////////////////////////// App18 ///////////////////////////////////

void CpmVQM2::App18::doTheWork()
// Experimenting with a chain of spin 1/2 particles with Heisenberg model
// like interaction

{
   Z mL=1;
   Word loc("CpmVQM2::App18::doTheWork()");
   CPM_MA
   using namespace CpmQM;

   Word sec="system data";
   SysConx sc(rch,sec);

   sec="run control";
   RunConx rc(rch,sec);

   sec="display control";
   DisConx dc(rch,sec);

   V<C2> vx=SpinNx::iniState(sc);

   Graph g;

   SpinNx sn(vx,sc.h,sc.qXY,sc.qZ,sc.dotFac,rc);

   if (rc.twoRuns) sn.runX_(g,dc);
   else sn.run_(g,dc);

   CPM_MZ
}

void CpmVQM2::App19::doTheWork()
{
   Z mL=1;
   Word loc("CpmVQM2::App19::doTheWork()");
   CPM_MA
   using namespace CpmQuantumMechanics;
   Z comp=16;
   R err=FieldState::test(comp);
   cout<<"FieldState::test(comp) = "<<err<<endl;
   CPM_MZ
}


//////////////// void CpmVQM2::function(void) //////////////////////////

// functions of signature void *f(void) for final association with a menu
// selection

void CpmVQM2::app17()
{
   cpmmessage("app17() started");
   App17 a("app17");
   a.run();
}

void CpmVQM2::app18()
{
   cpmmessage("app18() started");
   App18 a("app18");
   a.run();
}

void CpmVQM2::app19()
{
   cpmmessage("app19() started");
   App19 a("app19");
   a.run();
}
